/* EM_C_ARITH.C 910415  */

#include "stdarg.h"
#include "stdio.h"
#include "ctype.h"
#include "string.h"
#include "p2_symtab.h"
#include "em_common.h"
#include "p2_common.h"
#include "p2_header.h"
extern quad_cnt;


/* IN ALL OF THE ROUTINES IN THIS MODULE:
      arg1,arg2,.. are the numeric addresses of the variable
      carg1,carg2,.. are the symbolic names of the variables */

/* ===========================================================================

   function name: em_opcode
   input               : optype - integer for type of operation
			    mode                - integer for mode of operation
		  arg1          - input operand
		  arg2          - input operand
   output       : arg3          - output operand

   return value : emulated operation

   global vars  : ---

   description   : This routine calls the appropriate emulator subroutine
		  for the arithmetic, boolean and logical operators.

=========================================================================== */

/* all hex and binary combinations were added 880311 jlp */

em_opcode(optype, mode)
int optype, mode;

{  int status,arg1, arg2, arg3, size, size2, smallest, largest;
   int sarg;
	int mode1, mode2;

if(em_debug == 13)printf("In em_opcode, noex_bl=%d, optype=%x, mode=%x\n",
			  noex_bl,optype,mode);
   mode = fixmode(mode);
   mode1 = mode / SIZE_MODE2;
	mode2 = mode % SIZE_MODE2;
if(em_debug == 13) printf("fixedmode = %x, mode1=%x mode2=%x\n",mode,mode1,mode2);
   status = SUCCESS;
   em_scano_3();
   em_carg_3(&arg1,&arg2,&arg3);
   em_stack_buf();
if(em_debug == 13) printf(" noex_bl=%d\n",noex_bl);

   if(noex_bl > 0) return(status);

   size = get_size(carg1);
   size2 = get_size(carg2);
   smallest = size;
	largest = size2;
   if(smallest > size2) {smallest = size2; largest = size;}
if(em_debug == 13)printf("optype=%x, size,2,smallest=%d, %d %d\n",optype,size,size2,smallest);
   switch(optype)
     {case T_ADD  :
      case T_SUB  :
	  switch(mode)
	    {case CHPA:
	     case CAPA:
	     case OCPA:
	     case BNPA:
	     case HXPA:
	     case LIPA:
	     case FXPA: put_size(carg3,largest); /* set size to largest */
			fxpaadsb(optype,arg1+size-smallest, arg2+size2-smallest,
			       arg3+largest-smallest,smallest);
			/* sign extend to clear left garbage */
			pasnexmk(arg3+largest-smallest,arg3,MREG,largest-smallest);
			break;
	     case FLPA: flpaad(optype,arg1,arg2,arg3);
			break;
	     case CHCL: if(size <= SC_TEMP_SIZE && (sarg=get_add("SC_TEMP")) != FAIL)
			  {em_chrcnv(arg1,sarg,size);
			   arg1 = sarg;}
			else
			  printf("%d ERROR - no SC_TEMP\n",linenum);
	     case CACL:
	     case OCCL:
	     case BNCL:
	     case HXCL:
	     case LICL:
	     case FXCL: smallest = get_size(carg2);
			put_size(carg3,smallest);
			fxcladsb2(optype,arg2,arg1,arg3,size2,size);
			break;
	     case FLCL: flcmadd(optype,arg1,arg2,arg3); break;
	     case CHCR: if(size <= SC_TEMP_SIZE && (sarg=get_add("SC_TEMP")) != FAIL)
			 {em_chrcnv(arg2,sarg,size);
			 arg2 = sarg;}
		 else
		printf("%d ERROR - no SC_TEMP\n",linenum);
	      case OCCR:
	      case HXCR:
	      case BNCR:
	      case CACR:
	      case FXCR: smallest = get_size(carg1);
			 put_size(carg3,smallest);
			 fxcradsb2(optype,arg1,arg2,arg3,size,size2);
			 break;
	      case FLCR: if(optype == T_ADD)
			    flcmadd(optype,arg2,arg1,arg3);
			 else if (optype == T_SUB)
			    flcmaddcr(optype,arg1,arg2,arg3);
			 break;
	   case FXCN:
	   case CHCN:
	   case CACN:
	   case OCCN:
	   case HXCN:
	   case BNCN:
	   case LGCN:
	   case HXSC:
	   case OCSC:
			  case BNSC:
			  case CASC:
			  case CHSC:
		     case FXSC: put_size(carg3,size);
		      fxscadsb(optype, arg1, arg2, arg3, size);
					  break;
	   case FLCN:
		     case FLSC: flscad(optype,arg1,arg2,arg3);
					  break;
			  case LISC: put_size(carg3,size);
// 020505     liscadsb(optype, arg1, arg2, arg3, size);
		      liscadsb(optype, arg1, arg2, arg3);
					  break;
		     default  : printf("%d ERROR - unrecognized mode, opcode = %x\n",linenum,optype);
				    break;
	   }
		    break;
      case T_MULT :
		  switch(mode)
		    {case CHPA:
		     case OCPA:
		     case HXPA:
		     case BNPA:
	   case LIPA:
		     case FXPA: size2 = get_size(carg2);
				     put_size(carg3,size+size2);
				     fxpaml(arg1, arg2, arg3, size);
					  break;
		     case FLPA: flpaad(T_MULT,arg1,arg2,arg3);
					  break;
		     case CHCL: if(size <= SC_TEMP_SIZE && (sarg=get_add("SC_TEMP")) != FAIL)
						 {em_chrcnv(arg1,sarg,size);
			 arg1 = sarg;}
							 else
								printf("%d ERROR - no SC_TEMP\n",linenum);
		     case OCCL:
		     case HXCL:
		     case BNCL:
		     case FXCL: size = get_size(carg2);
				     put_size(carg3,2*size);
				     fxcrml2(arg2, arg1, arg3, size2, size);
				       break;
		     case FLCL: flcmadd(T_MULT,arg1,arg2,arg3); break;
			  case CHCR: if(size <= SC_TEMP_SIZE && (sarg=get_add("SC_TEMP")) != FAIL)
						 {em_chrcnv(arg2,sarg,size);
			 arg2 = sarg;}
							 else
								printf("%d ERROR - no SC_TEMP\n",linenum);
			  case OCCR:
		     case HXCR:
		     case BNCR:
		     case FXCR: put_size(carg3,2*size);
				     fxcrml2(arg1, arg2, arg3, size, size2);
				       break;
		     case FLCR: flcmadd(T_MULT,arg2,arg1,arg3); break;
	   case FXCN:
	   case CHCN:
	   case CACN:
	   case OCCN:
	   case HXCN:
	   case BNCN:
	   case LGCN:
	   case HXSC:
	   case OCSC:
			  case BNSC:
			  case CASC:
			  case LGSC:
			  case CHSC:
		     case FXSC: put_size(carg3,size);
		      fxscml(arg1, arg2, arg3, size);
					  break;
	   case FLCN:
		     case FLSC: flscad(T_MULT,arg1,arg2,arg3);
							 break;
			  case LISC: put_size(carg3,size);
// 020505      liscml(arg1, arg2, arg3, size);
		      liscml(arg1, arg2, arg3);
					  break;
		     default  : printf("%d ERROR - unrecognized mode, opcode = %x\n",linenum,optype);
				    break;
		    }
		    break;
      case T_MOD  :
/*
printf("line=%d, op_code=%d, mode=%x, quad_cnt=%d, size=%d, size2=%d\n",
linenum,optype,mode,quad_cnt,size,size2);
*/
		  switch(mode)
		    {case CHPA:
		     case OCPA:
		     case HXPA:
		     case BNPA:
		     case LIPA:
		     case FXPA: put_size(carg3,smallest);
				if(checksize_3() == SUCCESS)
//020505		   fxpamd(arg1+size-smallest,arg2+size2-smallest,arg3,smallest);
				   fxpamd(arg1+size-smallest,arg2+size2-smallest,arg3);
				else
printf("a line=%d, op_code=%d, quad_cnt=%d, size=%d, size2=%d\n",
linenum,optype,quad_cnt,size,size2);
				break;
		     case FLPA: flpaad(T_MOD,arg1,arg2,arg3);
				break;
		     case CHCL: if(size <= SC_TEMP_SIZE && (sarg=get_add("SC_TEMP")) != FAIL)
				  {em_chrcnv(arg1,sarg,size);
				   arg1 = sarg;}
				else
				  printf("%d ERROR - no SC_TEMP\n",linenum);
		     case OCCL:
		     case HXCL:
		     case BNCL:
		     case FXCL: put_size(carg1,smallest);
				put_size(carg3,smallest);
				if(checksize_3() == SUCCESS)
				   fxclmd2(arg1,arg2,arg3,size,size2);
				else
printf("b line=%d, op_code=%d, quad_cnt=%d, size=%d, size2=%d\n",
linenum,optype,quad_cnt,size,size2);
				break;
		     case FLCL: flcmadd(T_MOD,arg1,arg2,arg3); break;
		     case CHCR: if(size <=  SC_MEM_SIZE && (sarg=get_add("SC_TEMP")) != FAIL)
				  {em_chrcnv(arg2,sarg,size);
				   arg2 = sarg;}
				else
				  printf("%d ERROR - no SC_TEMP\n",linenum);
		     case OCCR:
		     case LICR:
		     case HXCR:
		     case BNCR:
		     case FXCR: put_size(carg2,smallest);
				put_size(carg3,smallest);
				if(checksize_3() == SUCCESS)
//020505		  fxcrmd2(arg1,arg2,arg3,size,size2);
				  fxcrmd2(arg1,arg2,arg3);
				else
				  {printf("c line=%d, quad_cnt=%d, op_code=%d, size=%d, size2=%d\n",
					   linenum,quad_cnt,optype,size,size2);
//020505		   fxcrmd2(arg1,arg2,arg3,size,size2);
				   fxcrmd2(arg1,arg2,arg3);
				   }
				break;
		     case FLCR: flcmaddcr(T_MOD,arg1,arg2,arg3); break;
		     case FXCN:
		     case CHCN:
		     case CACN:
		     case OCCN:
		     case BNCN:
		     case HXCN:
		     case LGCN:
		     case HXSC:
		     case OCSC:
		     case BNSC:
		     case CASC:
		     case LGSC:
		     case CHSC:
		     case FXSC:
			       put_size(carg3,size);
//020505	       fxscmd(arg1, arg2, arg3, size);
			       fxscmd(arg1, arg2, arg3);
			       break;
		     case FLCN:
		     case FLSC: flscad(T_MOD,arg1,arg2,arg3);
				break;
		     case LISC: put_size(carg3,size);
				liscmd(arg1, arg2, arg3);
				break;
		     default  : printf("%d ERROR - unrecognized mode, opcode = %x\n",linenum,optype);
				    break;
	  }
			 break;
      case T_DIV  :
		  switch(mode)
		    {case CHPA:
		     case OCPA:
		     case HXPA:
		     case BNPA:
                     case LIPA:
		     case FXPA: put_size(carg3,smallest);
                                if(checksize_3() == SUCCESS)
                                   fxpadv(arg1+size-smallest,arg2+size2-smallest,arg3,smallest);
                                else
printf("a line=%d, op_code=%d, quad_cnt=%d, size=%d, size2=%d\n",
linenum,optype,quad_cnt,size,size2);
                                break;
		     case FLPA: flpaad(T_DIV,arg1,arg2,arg3);
                                break;
                     case CHCL: if(size <= SC_TEMP_SIZE && (sarg=get_add("SC_TEMP")) != FAIL)
                                  {em_chrcnv(arg1,sarg,size);
                                   arg1 = sarg;}
                                else
                                  printf("%d ERROR - no SC_TEMP\n",linenum);
		     case OCCL:
		     case HXCL:
		     case BNCL:
		     case FXCL: put_size(carg1,smallest);
                                put_size(carg3,smallest);
                                if(checksize_3() == SUCCESS)
                                   fxcldv2(arg1,arg2,arg3,size,size2);
/*                                 fxcldv(arg1,arg2+size2-smallest,arg3,smallest);*/
                                else
printf("b line=%d, op_code=%d, quad_cnt=%d, size=%d, size2=%d\n",
linenum,optype,quad_cnt,size,size2);
                                break;
		     case FLCL: flcmadd(T_DIV,arg1,arg2,arg3); break;
		     case CHCR: if(size <=  SC_MEM_SIZE && (sarg=get_add("SC_TEMP")) != FAIL)
                                  {em_chrcnv(arg2,sarg,size);
                                   arg2 = sarg;}
                                else
                                  printf("%d ERROR - no SC_TEMP\n",linenum);
                     case OCCR:
		     case LICR:
		     case HXCR:
		     case BNCR:
		     case FXCR: put_size(carg2,smallest);
                                put_size(carg3,smallest);
                                if(checksize_3() == SUCCESS)
                                  fxcrdv2(arg1,arg2,arg3,size,size2);
                                else
                                  {printf("c line=%d, quad_cnt=%d, op_code=%d, size=%d, size2=%d\n",
                                           linenum,quad_cnt,optype,size,size2);
                                   fxcrdv2(arg1,arg2,arg3,size,size2);
                                   }
                                break;
		     case FLCR: flcmaddcr(T_DIV,arg1,arg2,arg3); break;
                     case FXCN:
                     case CHCN:
                     case CACN:
                     case OCCN:
                     case BNCN:
                     case HXCN:
                     case LGCN:
                     case HXSC:
                     case OCSC:
                     case BNSC:
                     case CASC:
                     case LGSC:
                     case CHSC:
		     case FXSC:
                               put_size(carg3,size);
                               fxscdv(arg1, arg2, arg3, size);
                               break;
                     case FLCN:
		     case FLSC: flscad(T_DIV,arg1,arg2,arg3);
                                break;
                     case LISC: put_size(carg3,size);
                                liscdv(arg1, arg2, arg3);
                                break;
		     default  : printf("%d ERROR - unrecognized mode, opcode = %x\n",linenum,optype);
				    break;
	  }
			 break;
      case T_LT   :
		  switch(mode)
		    {case CHPA:
		     case OCPA:
		     case BNPA:
		     case HXPA:
	   case LIPA:
		     case FXPA: fxpalt(arg1, arg2, arg3, size);
		      lgpaand(arg3,MREG,arg3,1);
  			  inc_pa_perform(-4);
			  break;
		     case CHCR: if(size <= SC_TEMP_SIZE && (sarg=get_add("SC_TEMP")) != FAIL)
						 {em_chrcnv(arg2,sarg,size);
			 arg2 = sarg;}
							 else
								printf("%d ERROR - no SC_TEMP\n",linenum);
	   case CACR:
		     case OCCR:
		     case HXCR:
		     case BNCR:
		     case FXCR: fxcrlt2(arg1, arg2, arg3, size, size2);
		      lgpaand(arg3,MREG,arg3,1);
							 inc_pa_perform(-4);
				       break;
	   case FLCR: flcmcmp(T_GT,arg2,arg1,arg3); break;
		     case CHCL: if(size <= SC_TEMP_SIZE && (sarg=get_add("SC_TEMP")) != FAIL)
						 {em_chrcnv(arg1,sarg,size);
			 arg1 = sarg;}
							 else
								printf("%d ERROR - no SC_TEMP\n",linenum);
		     case OCCL:
	   case CACL:
		     case HXCL:
		     case BNCL:
		     case FXCL: size = get_size(carg2);
				     fxcrgt2(arg2,arg1,arg3,size2,size);
		      lgpaand(arg3,MREG,arg3,1);
							 inc_pa_perform(-4);
				       break;
	   case FLCL: flcmcmp(T_LT,arg1,arg2,arg3); break;
	   case FXCN:
	   case CHCN:
	   case CACN:
	   case OCCN:
	   case BNCN:
	   case HXCN:
	   case LGCN:
	   case HXSC:
	   case OCSC:
			  case BNSC:
			  case CASC:
			  case LGSC:
			  case CHSC:
		     case FXSC: fxsclt(arg1, arg2, arg3, size, size2);
					  break;
		     case FLSC: flsccmp(T_LT,arg1,arg2,arg3);
							 break;
	   case FLPA: flpacmp(T_LT,arg1,arg2,arg3);  break;
	   case FLCN:
		     default:   em_print_err(FLOAT,5);
					  printf("Line number %d, opcode = %x\n",linenum,optype);
				    break;
		    }
		    break;
      case T_LE   :
		  switch(mode)
		    {case CHPA:
		     case OCPA:
		     case HXPA:
		     case BNPA:
	   case LIPA:
		     case FXPA: fxpale(arg1, arg2, arg3, size);
		      lgpaand(arg3,MREG,arg3,1);
							 inc_pa_perform(-4);
					  break;
		     case CHCR: if(size <= SC_TEMP_SIZE && (sarg=get_add("SC_TEMP")) != FAIL)
						 {em_chrcnv(arg2,sarg,size);
			 arg2 = sarg;}
							 else
								printf("%d ERROR - no SC_TEMP\n",linenum);
	   case CACR:
		     case OCCR:
		     case BNCR:
		     case HXCR:
		     case FXCR: fxcrle2(arg1, arg2, arg3, size, size2);
		      lgpaand(arg3,MREG,arg3,1);
							 inc_pa_perform(-4);
				       break;
	   case FLCR: /* switch args, switch cmp */
		      flcmcmp(T_GE,arg2,arg1,arg3); break;
		     case CHCL: if(size <= SC_TEMP_SIZE && (sarg=get_add("SC_TEMP")) != FAIL)
						 {em_chrcnv(arg1,sarg,size);
			 arg1 = sarg;}
							 else
								printf("%d ERROR - no SC_TEMP\n",linenum);
			  case OCCL:
	   case CACL:
		     case BNCL:
		     case HXCL:
		     case FXCL: fxcrge2(arg2,arg1,arg3,size2,size);
		      lgpaand(arg3,MREG,arg3,1);
							 inc_pa_perform(-4);
				       break;
	   case FLCL: flcmcmp(T_LE,arg1,arg2,arg3); break;
	   case FXCN:
	   case CHCN:
	   case CACN:
	   case OCCN:
	   case HXCN:
	   case BNCN:
	   case LGCN:
	   case HXSC:
	   case OCSC:
			  case BNSC:
			  case CASC:
			  case LGSC:
			  case CHSC:
		     case FXSC: fxscle(arg1, arg2, arg3, size, size2);
					  break;
		     case FLSC: flsccmp(T_LE,arg1,arg2,arg3);
							 break;
	   case FLPA: flpacmp(T_LE,arg1,arg2,arg3);  break;
	   case FLCN:
		     default  : em_print_err(FLOAT,5);
					  printf("Line number %d, opcode = %x\n",linenum,command_bl);
				    break;
		    }
		    break;
      case T_EQ   :
		  switch(mode)
		    {case CHPA:
		     case OCPA:
		     case BNPA:
		     case HXPA:
		     case LIPA:
		     case FXPA: fxpaeq(arg1, arg2, arg3, size);
		      lgpaand(arg3,MREG,arg3,1);
							 inc_pa_perform(-4);
					  break;
			  case CHCR: if(size2 <= SC_TEMP_SIZE && (sarg=get_add("SC_TEMP")) != FAIL)
				       {nw_chrcnv(arg2,sarg,size2);
					arg2 = sarg;}
				       else
					 printf("%d ERROR - no SC_TEMP\n",linenum);
/* Don't think this is necessary, acesuite works ok without it.
				       {em_chrcnv(arg2,sarg,size);
printf("In %d em_c_ari, at CHCR, size=%d, size2=%d, sc_mem[sarg]=%x, sc_mem[arg2]=%x\n",
linenum,size,size2,sc_mem[sarg],sc_mem[arg2]);
				      if(size < size2)
					     movchr(arg2,sarg,size,size2);
					arg2 = sarg;
					}
				      else
				      printf("%d ERROR - no SC_TEMP\n",linenum);
*/
/*                                   break; - no break, must fall through */
	   case CACR:
		     case OCCR:
		     case BNCR:
		     case HXCR:
		     case FXCR: fxcreq2(arg1, arg2, arg3, size, size2);
		      lgpaand(arg3,MREG,arg3,1);
							 inc_pa_perform(-4);
				  break;
	   case FLCR: flcmcmp(T_EQ,arg2,arg1,arg3); break;
	   case CHCL: if(size <= SC_TEMP_SIZE && (sarg=get_add("SC_TEMP")) != FAIL)
			 {nw_chrcnv(arg1,sarg,size);
			  arg1 = sarg;}
		      else
			printf("%d ERROR - no SC_TEMP\n",linenum);
/*                    break; - no break must fall through */
	   case OCCL:
	   case CACL:
	   case BNCL:
	   case HXCL:
	   case FXCL: fxcreq2(arg2,arg1,arg3,size2,size);
		      lgpaand(arg3,MREG,arg3,1);
		      inc_pa_perform(-4);
		      break;
	   case FLCL: flcmcmp(T_EQ,arg1,arg2,arg3); break;
	   case FXCN:
	   case CHCN:
	   case CACN:
	   case OCCN:
	   case BNCN:
	   case HXCN:
	   case LGCN:
	   case HXSC:
	   case OCSC:
	   case BNSC:
	   case CASC:
	   case LGSC:
	   case CHSC:
	   case FXSC:
/*
printf("line %d, arg1=%x,%x,%x, arg2=%x,%x,%x, arg3=%x, size=%d, size2=%d\n",
linenum,sc_mem[arg1+2],sc_mem[arg1+1],sc_mem[arg1],sc_mem[arg2+2],sc_mem[arg2+1],
sc_mem[arg2],sc_mem[arg3],size, size2);
*/
		      fxsceq(arg1, arg2, arg3, size, size2);
					  break;
	   case FLSC: flsccmp(T_EQ,arg1,arg2,arg3);
							 break;
	   case FLPA: flpacmp(T_EQ,arg1,arg2,arg3);  break;
	   case FLCN:
		     default  : em_print_err(FLOAT,5);
					  printf("Line number %d, opcode = %x\n",linenum,command_bl);
				    break;
		    }
		    break;
      case T_NE   :
		  switch(mode)
		    {case CHPA:
		     case OCPA:
		     case HXPA:
		     case BNPA:
	   case LIPA:
		     case FXPA: fxpane(arg1, arg2, arg3, size);
		      lgpaand(arg3,MREG,arg3,1);
							 inc_pa_perform(-4);
					  break;
		     case CHCR: if(size <= SC_TEMP_SIZE && (sarg=get_add("SC_TEMP")) != FAIL)
						 {em_chrcnv(arg2,sarg,size);
			 arg2 = sarg;}
							 else
								printf("%d ERROR - no SC_TEMP\n",linenum);
	   case CACR:
		     case OCCR:
		     case HXCR:
		     case BNCR:
		     case FXCR: fxcrne2(arg1, arg2, arg3, size, size2);
		      lgpaand(arg3,MREG,arg3,1);
							 inc_pa_perform(-4);
				       break;
	   case FLCR: flcmcmp(T_NE,arg2,arg1,arg3); break;
		     case CHCL: if(size <= SC_TEMP_SIZE && (sarg=get_add("SC_TEMP")) != FAIL)
						 {em_chrcnv(arg1,sarg,size);
			 arg1 = sarg;}
							 else
								printf("%d ERROR - no SC_TEMP\n",linenum);
		     case OCCL:
		     case HXCL:
		     case BNCL:
			  case CACL:
		     case FXCL: fxcrne2(arg2,arg1,arg3,size2,size);
		      lgpaand(arg3,MREG,arg3,1);
							 inc_pa_perform(-4);
				       break;
	   case FLCL: flcmcmp(T_NE,arg1,arg2,arg3); break;
	   case FXCN:
	   case CHCN:
	   case CACN:
	   case OCCN:
	   case HXCN:
	   case BNCN:
	   case LGCN:
	   case HXSC:
	   case OCSC:
			  case BNSC:
			  case CASC:
			  case LGSC:
			  case CHSC:
                     case FXSC: fxscne(arg1, arg2, arg3, size, size2);
					  break;
		     case FLSC: flsccmp(T_NE,arg1,arg2,arg3);
							 break;
	   case FLCN:
	   case FLPA: flpacmp(T_NE,arg1,arg2,arg3);  break;
		     default  : em_print_err(FLOAT,5);
					  printf("Line number %d, opcode = %x\n",linenum,command_bl);
				    break;
		    }
		    break;
      case T_GT   :
		  switch(mode)
		    {case CHPA:
		     case HXPA:
		     case BNPA:
		     case OCPA:
	   case LIPA:
		     case FXPA: fxpagt(arg1, arg2, arg3, size);
		      lgpaand(arg3,MREG,arg3,1);
							 inc_pa_perform(-4);
					  break;
			  case CHCR: if(size <= SC_TEMP_SIZE && (sarg=get_add("SC_TEMP")) != FAIL)
						 {em_chrcnv(arg2,sarg,size);
			 arg2 = sarg;}
							 else
								printf("%d ERROR - no SC_TEMP\n",linenum);
		     case OCCR:
	   case CACR:
		     case HXCR:
		     case BNCR:
		     case FXCR: fxcrgt2(arg1, arg2, arg3, size, size2);
		      lgpaand(arg3,MREG,arg3,1);
							 inc_pa_perform(-4);
				       break;
	   case FLCR: /* switch args, switch cmp */
		      flcmcmp(T_LT,arg2,arg1,arg3); break;
			  case CHCL: if(size <= SC_TEMP_SIZE && (sarg=get_add("SC_TEMP")) != FAIL)
						 {em_chrcnv(arg1,sarg,size);
			 arg1 = sarg;}
							 else
								printf("%d ERROR - no SC_TEMP\n",linenum);
		     case OCCL:
	   case CACL:
		     case HXCL:
		     case BNCL:
		     case FXCL: fxcrlt2(arg2,arg1,arg3,size2,size);
		      lgpaand(arg3,MREG,arg3,1);
							 inc_pa_perform(-4);
				       break;
	   case FLCL: flcmcmp(T_GT,arg1,arg2,arg3); break;
	   case FXCN:
	   case CHCN:
	   case CACN:
	   case OCCN:
	   case HXCN:
	   case BNCN:
	   case LGCN:
	   case HXSC:
	   case OCSC:
	   case BNSC:
	   case CASC:
	   case LGSC:
	   case CHSC:
	   case FXSC: fxscgt(arg1, arg2, arg3, size, size2); break;
	   case FLSC: flsccmp(T_GT,arg1,arg2,arg3);  break;
	   case FLPA: flpacmp(T_GT,arg1,arg2,arg3);  break;
	   case FLCN:
		     default  : em_print_err(FLOAT,5);
					  printf("Line number %d, opcode = %x\n",linenum,command_bl);
				    break;
		    }
		    break;
      case T_GE   :
		  switch(mode)
		    {case CHPA:
		     case HXPA:
		     case BNPA:
		     case OCPA:
	   case LIPA:
		     case FXPA: fxpage(arg1, arg2, arg3, size);
				lgpaand(arg3,MREG,arg3,1);
				inc_pa_perform(-4);
				break;
		     case CHCR: if(size <= SC_TEMP_SIZE && (sarg=get_add("SC_TEMP")) != FAIL)
				  {em_chrcnv(arg2,sarg,size);
				   arg2 = sarg;}
				 else
				   printf("%d ERROR - no SC_TEMP\n",linenum);
		     case OCCR:
		     case CACR:
		     case HXCR:
		     case BNCR:
		     case FXCR: fxcrge2(arg1, arg2, arg3, size, size2);
				lgpaand(arg3,MREG,arg3,1);
				inc_pa_perform(-4);
				break;
	   case FLCR: /* switch args, switch cmp */
		      flcmcmp(T_LE,arg2,arg1,arg3); break;
		     case CHCL: if(size <= SC_TEMP_SIZE && (sarg=get_add("SC_TEMP")) != FAIL)
						 {em_chrcnv(arg1,sarg,size);
			 arg1 = sarg;}
							 else
								printf("%d ERROR - no SC_TEMP\n",linenum);
		     case OCCL:
		     case HXCL:
	   case CACL:
		     case BNCL:
		     case FXCL: fxcrle2(arg2,arg1,arg3,size2,size);
		      lgpaand(arg3,MREG,arg3,1);
							 inc_pa_perform(-4);
				       break;
	   case FLCL: flcmcmp(T_GE,arg1,arg2,arg3); break;
	   case FXCN:
	   case CHCN:
	   case CACN:
	   case OCCN:
	   case HXCN:
	   case BNCN:
	   case LGCN:
	   case HXSC:
	   case OCSC:
			  case BNSC:
			  case CASC:
			  case LGSC:
			  case CHSC:
		     case FXSC: fxscge(arg1, arg2, arg3, size, size2);
					  break;
		     case FLSC: flsccmp(T_GE,arg1,arg2,arg3);
							 break;
	   case FLCN:
	   case FLPA: flpacmp(T_GE,arg1,arg2,arg3);  break;
		     default  : em_print_err(FLOAT,5);
					  printf("Line number %d, opcode = %x\n",linenum,command_bl);
				    break;
		    }
		    break;
      case T_OR   :
		  switch(mode)
		    {case LGPA: lgpaor(arg1, arg2, arg3, size);
					  break;
		     case LGCR: lgcror(arg1, arg2, arg3, size);
					  break;
		     case LGCL: size = get_size(carg2);
					  lgcror(arg2, arg1, arg3, size);
			     break;
		     case LGCN:
		     case LGSC: lgscor(arg1, arg2, arg3, size);
					  break;
		     default  : em_print_err(FLOAT,6);
				    break;
		    }
		    break;
      case T_AND  :
		  switch(mode)
		    {case LGPA: lgpaand(arg1, arg2, arg3, size);
					  break;
		     case LGCR: lgcrand(arg1, arg2, arg3, size);
					  break;
		     case LGCL: size = get_size(carg2);
					  lgcrand(arg2, arg1, arg3, size);
			     break;
		     case LGCN:
		     case LGSC: lgscand(arg1, arg2, arg3, size);
					  break;
		     default  : em_print_err(FLOAT,6);
				    break;
		    }
		    break;
      case T_XOR  :
		  switch(mode)
		    {case LGPA: lgpaxor(arg1, arg2, arg3, size);
					  break;
		     case LGCR: lgcrxor(arg1, arg2, arg3, size);
					  break;
		     case LGCL: size = get_size(carg2);
					  lgcrxor(arg2, arg1, arg3, size);
			     break;
		     case LGCN:
		     case LGSC: lgscxor(arg1, arg2, arg3, size);
					  break;
		     default  : em_print_err(FLOAT,6);
				    break;
		    }
		    break;
     default: printf("%d ERROR - unrecognized opcode, %x\n",linenum,command_bl);
   }

   return(status);
}


em_minusplus(optype, mode)
int optype, mode;
{
   int status,arg1,arg2,size;     /* program error status -- SUCCESS or FAIL */
   status = SUCCESS;
   em_scano_2();
   em_carg_2(&arg1,&arg2);
   em_stack_buf();
   if(noex_bl) return(status);

   mode = fixmode(mode);
   size = get_size(carg1);
   if (optype == T_MINUS)
     {switch(mode)
       {case CHPA:
	     case OCPA:
	     case BNPA:
	case HXPA:
	case LIPA:
	case FXPA: put_size(carg2,size);
		   fxpamin(arg1,arg2,size);
			     break;
	case FLPA: /*put_size(carg2,size);*/
		   flpaad(optype,arg1,arg2,arg2);  break;
	case FXCN:
	case CHCN:
	case CACN:
	case OCCN:
	case BNCN:
	case HXCN:
	case LGCN:
	case HXSC:
	case OCSC:
	  case BNSC:
		  case CASC:
		  case LGSC:
		  case CHSC:
	case LISC:
	case FXSC:put_size(carg2,size);
		  fxscmin(arg1,arg2,size);
		  break;
		  case FLCN:
	     case FLSC: flscad(optype,arg1,arg2,arg2); break;
	default  : printf("ERROR - illegal mode %x in minusplus\n",mode);
	 }
      return(status);
    }

   if (optype == T_PLUS)
     {printf("ERROR - T_PLUS quadruple not implemented.\n");
      return(FAIL);}
 return(SUCCESS);
}



em_not(mode)
int mode;
{
   int status, arg1, arg2, size; /* program error status -- SUCCESS or FAIL */

   mode = fixmode(mode);
   status = SUCCESS;
   em_scano_2();
   em_carg_2(&arg1,&arg2);
   em_stack_buf();
   if(noex_bl) return(status);

   size = get_size(carg1);
   switch(mode)
     {case LGPA : put_size(carg2,size);
		  lgpanot(arg1,emptys,size);
		  lgpaand(emptys,fillds,arg2,size);
			 break;
      case LGCN :
      case LGSC : lgscnot(arg1,arg2,size);
		       break;
      default   : printf("%d ERROR - Illegal mode %x combination for not\n",
								 linenum,mode);
     }
   return(status);
}


em_short(mode)
/* routine to convert double precision (i.e. result of mult) to single */
/* magnitude is modulo, sign is copied */
int mode;
{
   int arg1,arg2,size,size1,size2;
   int status;
	int mode2;

   mode = fixmode(mode);
   mode2 = mode % SIZE_MODE2;
   status=SUCCESS;
   switch(mode2)
     {case T_PARALLEL : em_scano_2();
			     em_carg_2(&arg1,&arg2);
			     em_stack_buf();
			if(noex_bl) break;
			     size1 = get_size(carg1);
			     size2 = get_size(carg2);
			if(size1 >= size2)
			       {papamvmk(arg1+size1-size2,arg2,MREG,size2);
				papamvmk(arg1,arg2,MREG,1);}
			     else
					 {papamvmk(arg1,arg2,MREG,size1);
					  put_size(carg2,size1);}
				       break;
      case T_CNST:printf("WARNING - constant treated as scalar in em_short\n");
      case T_SCALAR   : em_scano_2();
			     em_carg_2(&arg1,&arg2);
			em_stack_buf();
			if(noex_bl) break;
			     size = get_size(carg1);
			scscmv(arg1,arg2,size);
				       break;
      default: print_err("Unrecognized intermediate code for SHORT quadruple",0,0);
			      status=FAIL;
			      break;
    }
    return(status);
}



em_float(mode)
int mode;
{
   int arg1,arg2;
	int mode1, mode2;

   mode = fixmode(mode);
   mode1 = mode / SIZE_MODE2;
   mode2 = mode % SIZE_MODE2;
   em_scano_2();
   em_carg_2(&arg1,&arg2);
   em_stack_buf();
   if(noex_bl != 0) return(SUCCESS);

   put_size(carg2,REAL_SIZE);
   if (mode2 == T_PARALLEL)
     pa_float(mode1,arg1,arg2);
   else if (mode2 == T_SCALAR || mode2 == T_CNST)
     sc_float(mode1,arg1,arg2);
   else
     printf("%d ERROR - Unrecognized mode %x for FLOAT.",linenum,mode);

   return(SUCCESS);
}


em_fix(mode)
int mode;
{
   int arg1,arg2,size1,size2;
   int mode2, result;

   mode = fixmode(mode);
   mode2 = mode % SIZE_MODE2;
   em_scano_2();
   em_carg_2(&arg1,&arg2);
   em_stack_buf();
   if(noex_bl != 0) return(SUCCESS);
   size1 = get_size(carg1);
   size2 = INT_SIZE;
   put_size(carg2,size2);
   if (mode2 == T_PARALLEL)
     {result = pa_iterate(T_FIX,arg1,arg2,arg2,size1,size2);
      return result;}
   else if (mode2 == T_SCALAR || mode2 == T_CNST)
     {sc_fix(arg1,arg2);
      return SUCCESS;}
   else
     {print_err("%d ERROR - Unrecognized mode %x for FIX.",linenum,mode);
      return FAIL;}
}

fixmode(mode)
int mode;
{ int mode2,mode1;

   mode1 = mode / SIZE_MODE2;
	mode2 = mode % SIZE_MODE2;
   switch(mode2)
	  {case T_REG_SCALAR  : mode2 = T_SCALAR;       break;
		case T_REG_LEFT    : mode2 = T_COMMON_LEFT;  break;
		case T_REG_RIGHT   : mode2 = T_COMMON_RIGHT; break;
		case T_SCALAR_LEFT : mode2 = T_SCALAR;       break;
		case T_SCALAR_RIGHT: mode2 = T_SCALAR;       break;
		case T_CNST        : mode2 = T_SCALAR;       break;
	  }
	mode = mode1*SIZE_MODE2 + mode2;
return(mode);
}
