/* EM_CMDS_IO.C  910415 */
/* ASC maintains a table called filetab for OPENed files.  The table
   contains a pointer to the inputfile name, filein, a pointer to the
   output file name, fileot, and the output stream number, streamot.
   The entires in the table are offset by -6 to allow for system
   handle numbers.  I.e. handle #7 which is the first user # available
   is in slot #1.  There are also flages to record that there was input and
   or output on this stream.
*/

#include "errno.h"
#include "stdarg.h"
#include "stdio.h"
#include "ctype.h"
#include "string.h"
#include "p2_symtab.h"
#include "em_common.h"
#include "p2_common.h"
#include "p2_header.h"
#include "p2_parser.h"

#define TOK_SIZ        (TKBF_SZ*8)/SIZE_HOST
#define MODE_SIZE      256   /* from p1_common.h  */
#define MAX_NO_LEN     256
#define IO_SIZE_BIN      8   /* The number of bits neededd to make an */
#define IO_SIZE_HEX     16   /* 4 digit output                        */
#define IO_SIZE_OCT     15
#define IO_SIZE_CHAR    16
#define IO_SIZE_LOGICAL  8
#define W_INT           16   /* bit field width for integer */
#define W_LONGINT       32   /* bit field width for long integer */
#define W_CARD          16
#define W_REAL          32   /* bit field width for real */
#define C_RSQB          93
#define C_LSQB          91
#define C_DOL           36
#define C_WS            32
#define C_ADD           43
#define C_SUB           45
#define C_MULT          42
#define C_DIV           47
#define C_ASG           61
#define C_LPRN          40
#define C_RPRN          41
#define C_COMMA         44
#define C_COLON         58
#define C_BAR          124
#define C_DOT           46
#define C_SLASH         47
#define C_AT            64
#define C_SEMI          59
#define L_COMMA      11264  /* 2c00 */
#define L_RSQB       23808  /* 5d00 */
#define L_LSQB       23296  /* 5b00 */
#define L_AT         16384  /* 4000 */
#define L_DOT        11776  /* 2e00 */


char *index();
char *findex();
long int longtemp;
int lwflag;
long fstcd();
long leftadj();
int nl_mode;
int sv_perform;
/* defined in p2_main.c - em_common.m/mh
int token_length;
char token[MAX_NO_LEN], *pnt, data[MAX_NO_LEN], *datapnt; */

em_open()
{int arg1,arg2,arg3,arg4,arg5;
 char temp[50], how[6], *pnt, *pntot, buf[6];
 int handin;
 FILE *lw, *lw2;
 char tpath[70];

 em_scano_5();
 em_carg_5(&arg1,&arg2,&arg3,&arg4,&arg5);
 em_stack_buf();
 if(noex_bl>0) return SUCCESS;
 put_size(carg4,SIZE_PACK); /* update sizes for temps */
 put_size(carg5,SIZE_PACK);

 pnt = em_catstr(temp,arg1,get_size(carg1));
 *pnt++ = C_DOT;
 pnt = em_catstr(pnt,arg2,get_size(carg2));
 if(sc_mem[arg3] == 1)strcpy(how,"r+");
 else if (sc_mem[arg3] == 2)strcpy(how,"w");
 else if (sc_mem[arg3] == 4)strcpy(how,"r+");
 else {printf("em_open - unrecognized how open code %d\n",sc_mem[arg3]);
       return FAIL;}
 //strcpy(tpath,"c:\\asc\\asc\\suites\\asc\\");
 strcpy(tpath,"");
 strcat(tpath,temp);
 lw =   fopen(tpath,how);
 if(lw == NULL)
	 {printf("%d ERROR - em_open - can't open %s for %s\n",linenum,tpath,how);
     return FAIL ;}
 handin = fileno((FILE*) lw);
 handin -= NOOFSY; /* adjust for system files*/
 if(handin >= FILETABSIZE || handin < 0)
   {printf("%d ERROR - filetab overflow in em_open, handin=%d\n",
	    linenum,handin);
    return FAIL;}
 strcpy(filetab[handin].filein,temp);
/* open output file */
 if((pntot = index_ss(temp,".")) != NULL)
	{pntot++;
	*pntot=0;}
 strcat(temp,"OT");
/* em_itoa(toascii(handin),buf,10); */
 em_itoa( toascii( handin ), buf );
 strcat(temp,buf);
 lw2 = fopen(temp,"w");
/* set output lw2 to "input" file if type is "w" */
 if(strcmp(how,"w") == 0)
	 lw2 = lw; 
 if(lw2 == NULL)
	 {printf("%d ERROR - em_open - can't open %s for w\n",linenum,temp);
     return FAIL ;}
 filetab[handin].streamot = lw2;
 strcpy(filetab[handin].fileot,temp);
 filetab[handin].didout = FALSE;
 filetab[handin].didin = FALSE;
 sc_mem[arg4] = (long unsigned int) lw >> 16;
 sc_mem[arg5] = (long unsigned int) lw ;
//printf("End of open, lw=%x, sc_mem[%d]=%x, sc_mem[%d]=%x\n",
//	   lw,arg4,sc_mem[arg4],arg5,sc_mem[arg5]);
 return SUCCESS;
}

dump_filetab()
{ int i;
  printf(" Filetab\n   filein     fileot   didin didot\n");
  for(i=0; i<FILETABSIZE; i++)
    printf("%s %s  %d  %d %p %p\n",
	 filetab[i].filein, filetab[i].fileot,
	 filetab[i].didin, filetab[i].didout,
	 filetab[i].filein, filetab[i].fileot);
}

int em_close()
{int arg1;
 long unsigned int lw;
 FILE *lw2;
 int handin;
 char temp[50];
 char *pntot;

 em_scano_1();
 em_carg_1(&arg1);
 em_stack_buf();
 if(noex_bl>0) return FAIL;
//printf("In em_close\n");
 lw = (unsigned) sc_mem[arg1];
 lw <<= 16;
 lw += (unsigned) sc_mem[arg1+1] & 0xFFFF;
 handin = fileno((FILE *) lw);
 handin -= NOOFSY; /* adjust for system files*/
 if(handin >= FILETABSIZE)
   {printf("%d ERROR - filetab overflow in em_close, handin=%d\n",
	   linenum,handin);
    return FAIL;}
/* fprintf((FILE *)lw, "\n");*/
 fclose((FILE *)lw);
 lw2 = filetab[handin].streamot;
 fprintf(lw2, "\n");
 fclose(lw2);
 /* switch files if any output */
 if(filetab[handin].didout == TRUE)
  if(filetab[handin].didin  == TRUE)
  /* didout = 1 && didin = 1 */
   {strcpy(temp,filetab[handin].filein);
    if((pntot = index_ss(temp,".")) != NULL)
	{pntot++;
	*pntot=0;}
    strcat(temp,"BAK");
    if(rename(filetab[handin].filein,temp) != 0)
       {if(errno == EACCES)
	   if(unlink(temp) == 0)
	      {if(rename(filetab[handin].filein,temp) != 0)
		 {printf("%d ERROR - can not rename %s to %s because",linenum,
			    filetab[handin].filein,temp);
		  perror(" ????");}}
	   else
	    {printf("%d ERROR - can not rename %s to %s because",linenum,
		    filetab[handin].filein,temp);
	     perror(" unlink failure");}
	else
	  {printf("%d ERROR - can not rename %s to %s because",linenum,
		    filetab[handin].filein,temp);
	   perror(" bak");}
      }
   if(rename(filetab[handin].fileot,filetab[handin].filein) != 0)
       {printf("%d ERROR - can not rename %s to %s because",linenum,
		filetab[handin].fileot,filetab[handin].filein);
	perror(" out");}
   }
  else /* didout == TRUE, didin == FALSE - print only file */
    {if(lw == (long unsigned int) lw2) /* store output file only */
	   return SUCCESS;
	   if(unlink(filetab[handin].filein) != 0)
       {printf("%d - ERROR - print only file %s",linenum,filetab[handin].filein);
	perror(" unlink failure");}
     if(rename(filetab[handin].fileot, filetab[handin].filein) != 0)
	 {printf("%d ERROR - renaming print only file %s to %s because",linenum,
		    filetab[handin].fileot,filetab[handin].filein);
	     perror(" ");}
    }
  else /* didout == FALSE */
    if(filetab[handin].didin == TRUE)
      /* didout == FALSE,  didin = TRUE  - read only file */
      {if(unlink(filetab[handin].fileot) != 0)
       {printf("%d - ERROR - read only file %s",linenum,filetab[handin].filein);
		perror(" unlink failure");}
       }
 return SUCCESS;
}

/* fstout is the first output quadruple, it initializes the IOMASK */
void em_fstout()
{
int arg1;

 em_scano_1();
 em_carg_1(&arg1);
 em_stack_buf();
 sv_perform = sc_mem[get_add("PERFORM")];
 sc_mem[get_add("PERFORM")] = 0;
 if(noex_bl > 0) return;
 papamvun(arg1,IOMASK,1);
}

/* em_outpt moves data in parallel from the specified field plus offset
   (arg1) to the IOSTACK.  The IOMASK (set by em_fstout) determines the
   entries in the IOSTACK used. The masked stack is cleared prior to any
   data movements. The result is LEFT justified in IOSTACK. MREG is saved.
   */
void em_outpt(mode)
int mode;
{
int arg1,size;
int mode1, mode2;

 mode1 = mode / SIZE_MODE2;
 mode2 = mode % SIZE_MODE2;

 em_scano_2();
 em_carg_1(&arg1);
 em_stack_buf();
 size = (int) get_value(carg2);
 if(noex_bl > 0) return;

 if(mode2 == T_PARALLEL)
   {/* use sign as fill for sign extend */
    /* do not sign extend for bin, oct and hex */
    if(mode1 == T_BIN  || mode1 == T_OCT || mode1 == T_HEX || mode1 == T_CHAR)
      paclrmk(IOSTACK,IOMASK,IOSTACK_SIZE);
    else
      pasnexmk(arg1,IOSTACK,IOMASK,IOSTACK_SIZE);
    papamvmk(arg1,IOSTACK+IOSTACK_SIZE-size,IOMASK,size);

//printf("In em_outpt, arg1=%d\n",arg1);
//minipdump(arg1, size, T_HEX);
//printf("In em_outpt, IOSTACK=%d\n",IOSTACK+IOSTACK_SIZE-size);
//minipdump(IOSTACK+IOSTACK_SIZE-size, size, T_HEX);

 }
 else
   print_err("Scalar outpt not implemented",0,0);
}


/* em_colapse takes the entries in the IOSTACK flagged by IOMASK and moves
   them to the scalar memory I/O buffer  plus offset (arg1).
   The number of items moved is returned in the scalar variable (arg2).
   arg3 is the address of the original variable field,
   arg4 is the size of the variable being colapsed.  This value is
   necessary in to know how many bits to pack into the scalar
   iobuffer.  */

void em_colapse(mode)
int mode;
{
int arg1,arg2,arg3,argt;
int size,count;
int mode1;

 mode1 = mode / SIZE_MODE2;
 em_scano_4();
 em_carg_3(&arg1,&arg2,&arg3);
 size = (int) get_value(carg4);
 em_stack_buf();
 if(noex_bl > 0) return;
 count = 0;
 put_size(carg2,SIZE_PACK);
 argt = get_add("TEMP2");
 papamvun(IOMASK,argt,1); /* save IOMASK */
 while(em_responders(IOMASK))
  {switch(mode1)
    {case T_BIN:     /* treat as unsigned for columnar output */
     case T_OCT:
     case T_HEX:
/* to32   case T_LONGINT:       */
     case T_LOGICAL: pascmv(IOSTACK+IOSTACK_SIZE-size,arg1,IOMASK,size);
		     break;
     case T_LONGINT:
     case T_CARD:
     case T_INT:     /* must use sign extend move to get negative numbers on otput */
		     pascmvsx(IOSTACK+IOSTACK_SIZE-size,arg1,IOMASK,size);
			       break;
     case T_CHAR:    /* data must be left adjusted in word for char(string) */
		     pascmvsx(IOSTACK+IOSTACK_SIZE-size,arg1,IOMASK,size);
		     sc_mem[arg1] = sc_mem[arg1] << (SIZE_PACK-size);
			       break;
     case T_REAL:    pascmv(IOSTACK+IOSTACK_SIZE-size,arg1,IOMASK,size);
//		 printf("In em_colapse, at %d, size=%d, sc_mem[%d]=%x\n",
//				IOSTACK+IOSTACK_SIZE-size,size,arg1,sc_mem[arg1]);
				break;
     default :       printf("ERROR in em_colapse. Unrecognized data type %d.\n",mode1);
		     break;
    }
   arg1++;
   step(IOMASK,IOMASK);
   count++;}
 sc_mem[arg2] = count;
 papamvun(argt,IOMASK,1); /* restore IOMASK */
}


/* em_print outputs a single varable from the buffer (arg1).  The buffer is
   divided into sub-bufferers specified as an offset.  The sub-buffers are
    indexed by a scalar variable (arg2).  em_print does internal to ascii
   conversion.  */

void em_print(int mode)
{
int arg1,arg2,arg4;
int mode1,mode2;
char strbuf[IOSTACK_SIZE], *pnt, temp[20];
int dat_dx;
FILE *lw2;
int handin;
FILE *fpnt;
long int lw;

  union forx {
		long int fhex;
		float flt_dat;}
	fltint;

 mode1 = mode / SIZE_MODE2;
 mode2 = mode % SIZE_MODE2;
 em_scano_4();
 em_stack_buf();
 if(noex_bl > 0) return;
 arg1 = get_caddo(carg2);
 arg2 = get_caddo(carg3);
 arg4 = get_caddo(carg4);
 dat_dx  = arg1 + sc_mem[arg2];  /* em_strcat requires sc_mem index */
/*
printf("\nIn emPprint, arg3=%d, ",sc_mem[arg2]);
*/
 pnt = em_catstr(temp,arg4,get_size(carg4));
 *pnt++ = 0;
 if(strcmp(carg1,"SYSOT") == 0)
   switch(mode1)
    {case T_LOGICAL:printf("%1d",sc_mem[dat_dx]);    break;
     case T_INT: if(strlen(temp) > 2) printf(temp,sc_mem[dat_dx]);
		 else printf("%4d ",sc_mem[dat_dx]);  break;
     case T_LONGINT:if(lwflag == 0)
						{longtemp = (unsigned) sc_mem[dat_dx];
					     longtemp &= 0xffff; /* mask out sign extend */
						 lwflag = 1;}
					 else
						{ lw = (unsigned) sc_mem[dat_dx];
					      lw &= 0xffff; /* mask out sign extend */
						  lw |= longtemp << SIZE_PACK;
						  if(strlen(temp) > 2) printf(temp,lw);
						  else printf("%ld ",lw);
						  lwflag = 0;}
						  break;
      case T_REAL:   if(lwflag == 0)
		      {longtemp = (unsigned) sc_mem[dat_dx] & 0xFFFF;  //020505
//printf("In em_print, at T_REAL,lwflag=0, longtemp=%x\n",longtemp);
		  lwflag = 1;}
		     else
		      {fltint.fhex = (unsigned) sc_mem[dat_dx] & 0xFFFF;  //020505
//printf("In em_print, at T_REAL,lwflag=1, before or fhex=%x\n",fltint.fhex);
		       fltint.fhex |= longtemp <<  SIZE_PACK;
//printf("In em_print, at T_REAL,lwflag=1, after or fhex=%x\n",fltint.fhex);
		       if(strlen(temp) > 2) printf(temp,fltint.flt_dat);
		       else printf("%11f",fltint.flt_dat);
		       lwflag = 0;}
		      break;
      case T_DOL:    if(lwflag == 0)
		      {longtemp = (unsigned) sc_mem[dat_dx];
		       lwflag = 1;}
		      else
		       {fltint.fhex = (unsigned) sc_mem[dat_dx];
			fltint.fhex |= longtemp <<  SIZE_PACK;
			printf("%11.2f",fltint.flt_dat);
			lwflag = 0;}
						  break;
     case T_BIN:    printbin(dat_dx);                break;
     case T_OCT:    printf("%05o",sc_mem[dat_dx]);   break;
     case T_HEX:    printf("%04x",sc_mem[dat_dx]);   break;
     case T_CHAR:   em_strcat(strbuf,dat_dx);
		    printf("%s",strbuf);             break;
     case T_CARD:   printf("%4u", (unsigned) sc_mem[dat_dx]);    break;
     default: printf("ERROR - unrecognized data type in em_print\n");
   }
 else
  {lw = (unsigned) sc_mem[get_caddo(carg1)];
   lw <<= 16;
   lw += (unsigned) sc_mem[get_caddo(carg1)+1] & 0xFFFF;
   handin = fileno((FILE *) lw)-NOOFSY; /* adjust for system files*/
   lw2 = filetab[handin].streamot;
   fpnt =(FILE *) lw2;
   filetab[handin].didout=TRUE;
   switch(mode1)
    {case T_LOGICAL:fprintf(fpnt,"%1d",sc_mem[dat_dx]);    break;
     case T_LONGINT:if(lwflag == 0)
		      {longtemp = (unsigned) sc_mem[dat_dx];
							  lwflag = 1;}
						  else
							 {lw = (unsigned) sc_mem[dat_dx];
							  lw |= longtemp << SIZE_PACK;
		       fprintf(fpnt,"%ld ",lw);
		       lwflag = 0;}
						  break;
     case T_INT:    fprintf(fpnt,"%4d",sc_mem[dat_dx]);    break;
     case T_REAL:    if(lwflag == 0)
		      {longtemp = (unsigned) sc_mem[dat_dx];
							  lwflag = 1;}
						  else
							 {fltint.fhex = (unsigned) sc_mem[dat_dx];
						     fltint.fhex |= longtemp <<  SIZE_PACK;
		       pnt = em_catstr(temp,arg4,get_size(carg4));
		       *pnt++ = 0;
						if(strlen(temp) > 2) fprintf(fpnt,temp,fltint.flt_dat);
						     else fprintf(fpnt,"%11f",fltint.flt_dat);
							  lwflag = 0;}
						  break;
      case T_DOL:    if(lwflag == 0)
		      {longtemp = (unsigned) sc_mem[dat_dx];
							  lwflag = 1;}
						  else
							 {fltint.fhex = (unsigned) sc_mem[dat_dx];
						     fltint.fhex |= longtemp <<  SIZE_PACK;
						     fprintf(fpnt,"%11.2f",fltint.flt_dat);
							  lwflag = 0;}
						  break;
     case T_BIN:    fprintbin(fpnt,dat_dx);                break;
     case T_OCT:    fprintf(fpnt,"%05o",sc_mem[dat_dx]);   break;
     case T_HEX:    fprintf(fpnt,"%04x",sc_mem[dat_dx]);   break;
     case T_CHAR:   em_strcat(strbuf,dat_dx);
			      fprintf(fpnt,"%s",strbuf);            break;
     case T_CARD:   fprintf(fpnt,"%4u",sc_mem[dat_dx]);    break;
     default: printf("ERROR - unrecognized data type in em_print\n");
   }
  }
}

void em_printnl()
{
 int arg1;
 long int lw;
 FILE *lw2;
 int handin;

 em_scano_1();
 em_carg_1(&arg1);
 em_stack_buf();
 if(noex_bl > 0) return;
 if(strcmp(carg1,"SYSOT") == 0)
   printf("%s\n",strbuf);  /* output current buffer full */
 else
  {lw = (unsigned) sc_mem[arg1];
   lw <<= 16;
   lw += (unsigned) sc_mem[arg1+1] & 0xFFFF;
   handin = fileno((FILE *) lw)-NOOFSY; /* adjust for system files*/
   lw2 = filetab[handin].streamot;
   filetab[handin].didout = TRUE;
   fprintf(lw2,"%s\n",strbuf);}
 strpnt = strbuf;  /* and reinit output buffer */
 *strpnt = 0;
 lwflag = 0;
}

void em_dash()
{
 int arg1,arg2;
 long int lw;
 FILE *lw2;
 int handin;
 int tstln;

 em_scano_2();
 em_carg_2(&arg1,&arg2);
 em_stack_buf();
 if(noex_bl > 0) return;
 if(strcmp(carg1,"SYSOT") == 0)
   tstln = 1;
 else
  {lw = (unsigned) sc_mem[arg1];
   lw <<= 16;
   lw += (unsigned) sc_mem[arg1+1] & 0xFFFF;
   handin = fileno((FILE *) lw)-NOOFSY; /* adjust for system files*/
   lw2 = filetab[handin].streamot;
   filetab[handin].didout = TRUE;
   tstln = 2;}

 gen_line_delim(lw2, tstln, arg2);

 strpnt = strbuf;  /*reinit output buffer */
 *strpnt = 0;
 lwflag = 0;
}

void em_print_del(int mode)
{long int lw;
 int arg1, mode1;
 FILE *lw2;
 int handin;

 em_scano_1();
 em_carg_1(&arg1);
 em_stack_buf();
 if(noex_bl > 0) return;
 mode1 = mode / SIZE_MODE2;
 if(strcmp(carg1,"SYSOT") == 0)
   if(mode1 == T_CSTR) /* ACE delimeter */
     printf("|");
   else if(mode1 == T_COMDEL)
     printf(", ");
   else
     printf(" "); /* put delimiter directly into output file */
 else
    {lw = (unsigned) sc_mem[arg1];
     lw <<= 16;
     lw += (unsigned) sc_mem[arg1+1] & 0xFFFF;
     handin = fileno((FILE *) lw)-NOOFSY; /* adjust for system files*/
     lw2 = filetab[handin].streamot;
     filetab[handin].didout = TRUE;
     if(mode1 == T_CSTR)  /* for ACE */
       fprintf(lw2,"|");
     else if(mode1 == T_COMDEL)
       fprintf(lw2,", ");
     else
       fprintf(lw2," ");

    }
}

/* em_begprint prints out the header for the output */
void em_begprint(int mode)
{
char arg1[100],arg2[100],arg3[(MAX_ID_SIZE+1)*MAX_NO_OUT];
long int lw;
 FILE *lw2;
 int handin, mode1;


 em_rscan_3(arg1,arg2,arg3);
 em_stack_buf();
 if(noex_bl > 0) return;
 mode1 = mode / SIZE_MODE2; /* no longer used */
 if(strcmp(arg1,"SYSOT") == 0)
     printf("DUMP OF ASSOCIATION %s FOLLOWS:\n %s\n",arg2,arg3);
 else
  {lw = (unsigned) sc_mem[get_caddo(arg1)];
   lw <<= 16;
   lw += (unsigned) sc_mem[get_caddo(arg1)+1] & 0xFFFF;
   handin = fileno((FILE *) lw)-NOOFSY; /* adjust for system files*/
   lw2 = filetab[handin].streamot;
   if(filetab[handin].didin != TRUE)
    {filetab[handin].didout = TRUE;
     if(mode1 == T_CSTR)   /* ACE print */
       fprintf(lw2,"%s\n",arg3);
     else
       fprintf(lw2,"DUMP OF ASSOCIATION %s FOLLOWS:\n %s\n",arg2,arg3);}
  }
  strpnt = strbuf;
 *strpnt = 0;
 lwflag = 0;
}

/* em_begread prints the header for the input operation */
void em_begread(int mode)
{
char arg1[100],arg2[100],arg3[(MAX_ID_SIZE+1)*MAX_NO_OUT];

 em_rscan_3(arg1,arg2,arg3);
 em_stack_buf();
 sv_perform = sc_mem[get_add("PERFORM")];
 sc_mem[get_add("PERFORM")] = 0;
 if(noex_bl > 0) return;

 if(strcmp(arg1,"SYSOT") == 0)
   if(mode != T_SILENT)
     printf("INPUT VALUES FOR %s:\nA blank line terminates input\n%s\n",arg2,arg3);
/* else
  {lw = (unsigned) sc_mem[get_caddo(arg1)];
   lw <<= 16;
   lw += (unsigned) sc_mem[get_caddo(arg1)+1] & 0xFFFF;
   fprintf((FILE *) lw,
			 "INPUT VALUES FOR %s:\nA blank line terminates input\n%s\n",arg2,arg3);
  }*/
}


/* em_readln inputs, (f)gets, the next line of data, data, and "parses it".  em_read()
   moves the parsed values from the line buffer to the INPUT buffer */
void em_readln(int mode)
{ int arg1,arg2,tstln;
  long unsigned int lw;
  int handin;

em_scano_3();
em_carg_1(&arg1);
em_carg_3p(&arg2);
em_stack_buf();
if(noex_bl <= 0)
 {nl_mode = mode;
  if(strcmp(carg1,"SYSIN") == 0)
    {if(gets(data) == NULL) *data = 0;
     tstln = 1;}
  else
    {lw = (unsigned) sc_mem[arg1];
     lw <<= 16;
     lw += (unsigned) sc_mem[arg1+1] & 0xFFFF;
     if(fgets(data,MAX_NO_LEN,(FILE *) lw) == NULL) *data = 0;
     handin = fileno((FILE *) lw) - NOOFSY; /* adjust for system files*/
     filetab[handin].didin = TRUE;
     tstln =2;}

  if(strlen(data)<(unsigned) tstln) sc_mem[arg2] = TRUE;
  else
   {strcat(data," ");
    datapnt = data;
    sc_mem[arg2] = FALSE;
    lwflag = 0;
    token_length = 0;}
  }
//printf("em_readln read %s#\n",data);
}


/* em_read() will input the next "token" into IOBUF if it will fit into
   the IO STACK.  Note IOBUF is assumed to be divided into IOSTACK size
   chunks.
 */
em_read_dum(mode)
int mode;
{ int arg1,arg2,threshold,i;
  char *pnt,save_ch;
  int sc_mem_dx;
  int mode1;

  mode1 = mode / SIZE_MODE2;
  em_scano_4();  /* modified to read size, arg3 880722 jlp */
  em_carg_2(&arg1,&arg2);
  em_stack_buf();
  if(noex_bl <= 0)
    {threshold = IOSTACK_SIZE/CHAR_SIZE;
     sc_mem_dx = arg1 + sc_mem[arg2]*(IOSTACK_SIZE/SCALAR_SIZE);
     if(T_REAL == mode1)
       sscanf(datapnt,"%f",sc_mem[sc_mem_dx]);
     else
       {/* clear the buffer as you use it */
	for(i=0; i<IOSTACK_SIZE/SCALAR_SIZE; i++)
			  sc_mem[sc_mem_dx+i] = 0;
	/* find first blank in input */
	if((pnt = index_ss(datapnt," ")) == NULL)
			  pnt=strlen(datapnt)+datapnt-1;

	if((pnt-datapnt) <= 0)
			 em_strcpy(sc_mem_dx,"  ");
	else if((pnt-datapnt) <= threshold)
	  {*pnt = 0; /* if data string is less than IOSTACK wide, move it */
	   em_strcpy(sc_mem_dx, datapnt);} /* to IOBUF in scalar memory */
	else                /* THIS DOES NOT WORK */
	  {pnt=datapnt+threshold; /* Otherwise, break it up into IOSTACK size */
	   save_ch = *pnt;        /* chunks and move them to IOBUF */
	   *pnt=0;
	   em_strcpy(sc_mem_dx,datapnt);
	   *pnt=save_ch;}
	while(*pnt == 0 || *pnt == ' ') pnt++; /* skip zeroes and spaces */
	   datapnt = pnt;       /* update beginning of data */
       }
  }
}

/* em_input moves the data in parallel from the IOSTACK to the
   array through the IOMASK */

void em_input(mode)
int mode;
{ int arg1,size2;
  int mode1,mode2;

  mode1 = mode / SIZE_MODE2;
  mode2 = mode % SIZE_MODE2;

  em_scano_2();
  em_carg_1(&arg1);
  em_stack_buf();
  if(noex_bl > 0) return;
//  printf("In em_input, calling papamvmk\n");
  size2 = (int) get_value(carg2);
  /* data is left adjusted in IOSTACK with zero fill */
  if(mode2 == T_PARALLEL)
    {if(mode1 == T_CHAR)
       papamvmk(IOSTACK+IOSTACK_SIZE-size2,arg1,IOMASK,size2);
     else
       papamvmk(IOSTACK+IOSTACK_SIZE-size2,arg1,IOMASK,size2);
    }
 else print_err("Scalar input not implemented.",0,0);

}

/* em_expand moves the data from the scalar memory to the IOSTACK.
   It is guided by the IOMASK */
/* -cm requires that the size argument for expand be the size of the
   data not the iostack(i.e. arg1), it is passed as arg3 of the
   expand quadruple but is ignorred here.
   Arg2 is the destination field id. It is supplied here to allow
   the emulator to convert data using the formated read capability. */
void em_expand(int mode)
{ int arg1,size,mode1;

  em_scano_3();  /* modified to read size, arg3 880722 jlp */
  em_carg_1(&arg1);
  em_stack_buf();
  mode1 = mode / SIZE_MODE2;
  size = (int) get_value(carg3);
  if(noex_bl > 0) return;
  expand(arg1,size,mode1);
}


/* em_ioalloc() is similar to allocate, however, it hardwires the
   output flag vector to IOMASK and claers the IOSTACK
*/

void em_ioalloc(int mode)
{ int arg1,arg2;

  em_scano_2();
  em_carg_2(&arg1,&arg2);
  em_stack_buf();
  if(noex_bl > 0) return;
  if(mode == T_READ/MODE_SIZE)
    {balloc(IOMASK,arg1,sc_mem[arg2]);
     lgpaor(IOMASK,arg1,arg1,1);}
  else if(mode == T_REREAD/MODE_SIZE) papamvun(arg1,IOMASK,1);
  else if(mode == T_COREAD/MODE_SIZE)
    {printf("WARNING - contiguous read not implemented - read used\n");
	  balloc(IOMASK,arg1,sc_mem[arg2]);
     lgpaor(IOMASK,arg1,arg1,1);}
  else printf("ERROR - unrecognized mode %x in em_ioalloc\n",mode);
  paclrmk(IOSTACK,IOMASK,IOSTACK_SIZE);
}

void em_endread()
/* entry to allow perform to be turned back on after print */
{
  em_scano_0();
  em_stack_buf();
  sc_mem[get_add("PERFORM")] = sv_perform;
};

void em_endprint()
/* entry to allow perform to be turned back on after print */
{
  em_scano_0();
  em_stack_buf();
  sc_mem[get_add("PERFORM")] = sv_perform;
};


em_read(int mode)
{ int arg1,arg2,arg3,i;
  int threshold; /* the number of bits in a column */
  int col_wid;   /* the number of char in a column */
  char save_ch, *token_pnt, extend_char, *end_pnt;
  int  size;
  int mode1, status;
  int *address;
  int sc_mem_dx;
  long int lw;
  //char *pnt;

  union forx {
		long int fhex;
		float flt_dat;}
	fltint;

  mode1 = mode / SIZE_MODE2;
  em_scano_4();
  em_carg_3(&arg1,&arg2,&arg3);
  em_stack_buf();
  status = SUCCESS;
  if(noex_bl <= 0)
  {sc_mem_dx = arg1 + sc_mem[arg2]; /* for em_strcpy routine */
   address = sc_mem + arg1 + sc_mem[arg2]; /* make address a pointer */
   size = atoi(carg3);
   put_size(carg1,size);
   /* threshold is the number of bits in an 8 character column of data */
   /* col_wid is the number of char in a column - normally 8 */
   /* extend_char is the right fill character */
   extend_char = ' ';  /* set to default values, change in switch */
   col_wid = 8;        /* below, if necessary */
   switch(mode1)
     {case T_LOGICAL: threshold = IO_SIZE_LOGICAL; break;
      case T_LONGINT: threshold = W_LONGINT; break;
      case T_INT:     threshold = W_INT; break;
      case T_BIN:     threshold = IO_SIZE_BIN; break;
      case T_REAL:    threshold = W_REAL; break;
      case T_OCT:     threshold = IO_SIZE_OCT; col_wid = 5;  break;
      case T_HEX:     threshold = IO_SIZE_HEX; col_wid = 4;  break;
      case T_CHAR:    threshold = IO_SIZE_CHAR; col_wid = 2; extend_char = 0; break;
      case T_CARD:    threshold = W_CARD; break;
      default : printf("ERROR in em_read. Unrecognized data type.\n");
		break;
      }
   /* clear scalar buffer */
//   printf("In em_read, datapnt=%s\n",datapnt);
//   020510 pnt made global for drop through to status != fail
  for(i=0; i<IOSTACK_SIZE/SCALAR_SIZE; i++) *(address+i) = 0;
  if((mode1 != T_LONGINT && mode1 != T_REAL && mode1 != T_DOL) || lwflag == 0)
   if(token_length <= 0)/* load a new token in token */
     {while((pnt = index_ss(datapnt," ")) == datapnt) datapnt++; /* strip blanks */
     if((pnt=index_ss(datapnt," ")) == NULL)
	{if(nl_mode != 1) printf("ERROR insufficient data=%s.\n",pnt);
	      status = FAIL; }
      else
	{*pnt = 0;
	 /* check for terminating nl char*/
	 if(*(pnt-1) == 10) *(pnt-1) = 0;
	 /* initialize token with extend_char */
	 for(i=0; i<MAX_NO_LEN; i++) ctoken[i] = extend_char;
	 /* rigth adjust token in token */
	 /* T_CHAR should be left adjusted */
	 /* token_length is the number of characters need for size bits, rounded up */
	 /* to nearest column.  col_wid, is number of characters in a column */
	    token_length = (((size-1)/threshold)+1)*col_wid;

	    if(mode1==T_CHAR)
	       token_pnt = &ctoken[0]+MAX_NO_LEN-1-token_length;
	    else
	       token_pnt = &ctoken[0]+MAX_NO_LEN-1-strlen(datapnt);
	    strcpy(token_pnt,datapnt);
	    datapnt = pnt+1;
	    /* init pnt to left most char in token */
	    if(mode1==T_CHAR)
	       pnt = &ctoken[0]+MAX_NO_LEN-1-token_length;
	    else
	       pnt = &ctoken[0]+MAX_NO_LEN-1-token_length;
    }
   }
    /* take next threshold characters from token */
   if(status != FAIL)
    if((mode1 != T_LONGINT && mode1 != T_REAL && mode1 != T_DOL) || lwflag == 0)
      {token_length -= col_wid;  /* reduce length by a column */
       end_pnt = col_wid+pnt;
       save_ch = *end_pnt; *end_pnt = 0;
       switch(mode1)
	 {case T_LOGICAL:
	  case T_INT:  sscanf(pnt,"%d ",address);
				 break;
	  case T_LONGINT:
		  sscanf(pnt,"%ld ",&longtemp);
		  lwflag=1;
		  lw = longtemp >> SIZE_PACK;
		  sc_mem[sc_mem_dx] = (int) lw;
		  break;
	  case T_BIN:  
//	  printf("in em_read, at T_BIN, inputing %s# into sc_mem at %p\n",
//					pnt,address);
		       scanbin(pnt,address);
		       break;
	  case T_DOL:
	  case T_REAL: sscanf(pnt,"%f ",&fltint.flt_dat);
		       lwflag=1;
		       longtemp = fltint.fhex;
		       lw = fltint.fhex >> SIZE_PACK;
		       sc_mem[sc_mem_dx] = (int) lw;
		       break;
	  case T_OCT:  sscanf(pnt,"%o ",address); break;
	  case T_HEX:  sscanf(pnt,"%x ",address); break;
	  case T_CHAR: em_strcpy(sc_mem_dx,pnt);  break; /* requires an index */
	  case T_CARD: sscanf(pnt,"%u ",address); break;
	  default :    printf("ERROR in em_read. Unrecognized data type.\n");
		       break;
      }
    *end_pnt = save_ch;
    pnt=end_pnt;
   }
  else
   {lwflag = 0;
    lw = longtemp << SIZE_PACK;
    lw = lw >> SIZE_PACK;
    sc_mem[sc_mem_dx] = (int) lw;
}

/*if(mode1 == T_LONGINT)
  printf("In T_LONGINT, pnt=%s, lwflag=%d, longtemp=%ld, lw=%ld, sc_mem[sc_mem_dx]=%d\n",
	   pnt, lwflag, longtemp, lw, sc_mem[sc_mem_dx]);
*/
if(em_debug_4) printf("At end of em_read, address = %x(hex), pnt=%x, sc_mem=%d\n",
  address,pnt,sc_mem[sc_mem_dx]);

 }
return status;
}

/* em_dump outputs a single value from sc_mem[arg1] */
/* quadruple is of form
      dump_ 8711 sc_mem_add  description
   description is not currently used  */

void em_dump(int mode)
{
  int arg1,arg2,arg3,mode1,mode2;
  char temp[50],*pnt;
  unsigned long accum, one;
  int i,k,j,size,mask,rm;
  unsigned int temp1,temp2,temp3;
  long lint, lw;
  int handin;
  FILE *fpnt;
  union forx {
		long int fhex;
		float flt_dat;}
 fltint;


  mode = fixmode(mode);
  mode1 = mode/SIZE_MODE2;
  mode2 = mode%SIZE_MODE2;

  em_scano_3();
  em_carg_2(&arg2,&arg1);
  arg3 = get_caddo(carg3);
  em_stack_buf();
  if(noex_bl > 0) return;
  pnt = em_catstr(temp,arg3,get_size(carg3));
  *pnt++ = 0;
  one = 1; /* 32 bit one */
  size = get_size(carg2);
  /* skip leading zeroes */
  for(i=(size-1)/SIZE_PACK; i>0; i--) if(sc_mem[arg1+i]!=0) break;
  if(mode2 == T_SCALAR || mode2 == T_CNST)
   if(strcmp(carg1,"SYSOT") == 0)
    switch(mode1)
	   {case T_INT: if(size<=SIZE_PACK)
			  {if(strlen(temp) > 2) printf(temp,sc_mem[arg1]);
			  else printf("%d",sc_mem[arg1]); break;}
		 /* drop through to longint if size > SIZE_HOST  */
	    case T_LONGINT:lint =  (unsigned) sc_mem[arg1+1];
		 lint =  lint<<SIZE_PACK;
		 lint |= (unsigned) sc_mem[arg1];
		 if(strlen(temp) > 2) printf(temp,lint);
		 else printf("%ld ",lint);  break;
	    case T_LOGICAL:
			  if(strlen(temp) > 2) printf(temp,sc_mem[arg1]);
			  else printf("%1d",sc_mem[arg1]);  break;
	    case T_BIN:  for(j=i; j>=0; j--)
			 {k=sc_mem[j+arg1]; temp1 = sc_mem[j+arg1];
			  sc_mem[j+arg1] = 0xff & temp1>>8;     printbin(j+arg1);
			  sc_mem[j+arg1]=k; printbin(j+arg1);}
			  printf(" ");  break;
	    case T_OCT:  for(j=i; j>=0; j--)
		      {temp1 = sc_mem[arg1+((j+1)*15)/SIZE_PACK]<<j;
		       temp2 = sc_mem[arg1+(j*15)/SIZE_PACK];
		       temp2 >>= (SIZE_PACK-j);
		       temp3 = (temp1 | temp2) & 0x7fff;
		       printf("%05o",temp3);}
		       printf(" ");  break;
       default: printf("%d ERROR - unrecognized mode1 %d in em_dump, hex default\n",
								linenum,mode1); break;
	    case T_HEX:  for(j=i; j>=0; j--) printf("%04x",sc_mem[j+arg1] & 0xFFFF);
						  printf(" ");  break;
	    case T_CARD: printf("%d ",sc_mem[arg1]);  break;
	    case T_REAL: lw = (unsigned) sc_mem[arg1+1];
			  lw <<= 16;
			  lw |= (unsigned) sc_mem[arg1];
		    fltint.fhex = lw;
		   if(strlen(temp) > 2) printf(temp,fltint.flt_dat);
		  else
		  printf("%f",fltint.flt_dat);  break;
	    case T_DOL: lw = (unsigned) sc_mem[arg1+1];
						  lw <<= 16;
						  lw |= (unsigned) sc_mem[arg1];
		    fltint.fhex = lw;
						  printf("%11.2f ",fltint.flt_dat);  break;
	    case T_CHAR:
		   for(j=i; j>=0; j--)
		     {if((sc_mem[arg1+j] & 65280) != 0)
		      printf("%c",sc_mem[arg1+j]>>8); else printf(" ");
		  /* grouping ()s necessary */
		  if((sc_mem[arg1+j] & 255) != 0)
			 printf("%c",sc_mem[arg1+j]); else printf(" ");}
		    printf(" ");
		    break;
      }
   else
    {lw = (unsigned) sc_mem[get_caddo(carg1)];
     lw <<= 16;
     lw += (unsigned) sc_mem[get_caddo(carg1)+1] & 0xFFFF;
     handin = fileno((FILE *) lw)-NOOFSY; /* adjust for system files*/
     fpnt = filetab[handin].streamot;
     switch(mode1)
      {case T_INT:  if(size<=SIZE_PACK)
		      if(strlen(temp) > 2) fprintf(fpnt,temp,sc_mem[arg1]);
		      else fprintf(fpnt,"%d",sc_mem[arg1]);
		     break;
	    case T_LONGINT:lint =  sc_mem[arg1+1];
		      lint =  lint<<SIZE_PACK;
		      lint |= sc_mem[arg1];
		 if(strlen(temp) > 2) fprintf(fpnt,temp,lint);
		 else fprintf(fpnt,"%ld ",lint);  break;
       default: fprintf(fpnt,"ERROR - unrecognized type in em_dump, hex default\n"); break;
	    case T_LOGICAL: fprintf(fpnt,"%1d ",sc_mem[arg1]);  break;
	    case T_BIN:  for(j=i; j>=0; j--)
							 {k=sc_mem[j+arg1]; temp1 = sc_mem[j+arg1];
							  sc_mem[j+arg1] = 0xff & temp1>>8;     printbin(j+arg1);
							  sc_mem[j+arg1]=k; printbin(j+arg1);}
						  fprintf(fpnt," ");  break;
	    case T_OCT:  for(j=i; j>=0; j--)
		      {temp1 = sc_mem[arg1+((j+1)*15)/SIZE_PACK]<<j;
		       temp2 = sc_mem[arg1+(j*15)/SIZE_PACK];
							  temp2 >>= (SIZE_PACK-j);
		       temp3 = (temp1 | temp2) & 0x7fff;
			  fprintf(fpnt,"%05o",temp3);}
						  fprintf(fpnt," ");  break;
	    case T_HEX:  for(j=i; j>=0; j--) fprintf(fpnt,"%04x",sc_mem[j+arg1]);fprintf(fpnt," ");  break;
	    case T_CARD: fprintf(fpnt,"%d ",sc_mem[arg1]);  break;
	    case T_REAL: lw = (unsigned) sc_mem[arg1+1];
						  lw <<= 16;
						  lw |= (unsigned) sc_mem[arg1];
		    fltint.fhex = lw;
		    if(strlen(temp) > 2) fprintf(fpnt,temp,fltint.flt_dat);
		    else  fprintf(fpnt,"%f ",fltint.flt_dat);
		    break;
	    case T_DOL:  lw = (unsigned) sc_mem[arg1+1];
						  lw <<= 16;
						  lw |= (unsigned) sc_mem[arg1];
		    fltint.fhex = lw;
						  fprintf(fpnt,"%11.2f ",fltint.flt_dat);  break;
	    case T_CHAR:
		    for(j=i; j>=0; j--)
		       {if((sc_mem[arg1+j] & 65280) != 0)
			 fprintf(fpnt,"%c",sc_mem[arg1+j]>>8); else fprintf(fpnt," ");
		       if((sc_mem[arg1+j] & 255) != 0)
			 fprintf(fpnt,"%c",sc_mem[arg1+j]); else fprintf(fpnt," ");}
		       break;
      }
    }
  else if(mode2 == T_PARALLEL)
   if(strcmp(carg1,"SYSOT") == 0)
    {size = get_size(carg2); /* carg2 is in arg1 */
     pdump(arg1,size,mode1);}
  else
    {lw = (unsigned) sc_mem[get_caddo(carg1)];
     lw <<= 16;
     lw += (unsigned) sc_mem[get_caddo(carg1)+1] & 0xFFFF;
     handin = fileno((FILE *) lw)-NOOFSY; /* adjust for system files*/
     fpnt = filetab[handin].streamot;
     size = get_size(carg2); /* carg2 is in arg1 */
     for(j=0;j<NOOFW;j++)
			for(k=SIZE_PACK-1;k>=0;k--)
			   {mask=1<<k;
			    if(((rm=pa_mem[j][MREG]) & mask)!=0)
	       {accum = 0;
					 for(i=0;i<size;i++)
					   if(rm & pa_mem[j][arg1+i]) accum += one<<(size-i-1);
					}
	     fprintf(fpnt,"%ld, ",accum);
	    }
     fprintf(fpnt,"\n");
    }
  else
	 printf("%d ERROR - illegal mode2 %d in dump\n",linenum,mode2);
}

#define NOOFCOL 16
#define COL_OFF 16   /* for 32 bit machines to32 */
void pdump(int arg1, int size, int mode1)
{
 int i,j,k,mask,colcnt;
 unsigned long accum,one;
 union forx {
		long int fhex;
		float flt_dat;}
 fltint;

/*
printf("In dpump at %d, arg1=%d, size=%d\n",linenum,arg1,size);
*/
/* if(size > 32) xdump(arg1,size,T_HEX);
 else
 {                                  931030    */
 colcnt = 1;
 one = 1;
 for(j=0;j<NOOFW;j++)
   {printf("PE%3d:",j*NOOFCOL);
	//printf("%d arg1=%d pa_mem=%p PE%2d:",
	//          linenum,arg1,&pa_mem[j][arg1],j*SIZE_PACK);
	for(k=SIZE_PACK-1;k>=0;k--)
	 {//020505 SIZE_PACK is defined as 16, and seems to work elsewhere, 
	  // but the mask is for 32 bit words, thus the +16
	  mask=1<<(k+16);
	  accum = 0;
	  /* arg1 + 0123456789... */
	  for(i=0;i<size;i++)
	   {//printf("mask=%x, pa_mem[%d][%d]=%x",mask,j,arg1+i,pa_mem[j][arg1+i]);
	    if(mask & pa_mem[j][arg1+i])
		  accum += one<<(size-i-1);
		//printf(", accum=%d\n",accum);
	   }

	  if(mode1 == T_HEX) printf("%8lx ",accum);
	  else if(mode1 == T_CHAR)
		{/*for(i=3; i>=0; i--){tl = accum>>i*8;printf("%1c",tl);}*/
		 printf("%8lx",accum);
		 printf(" ");}
	  else if(mode1 == T_REAL)
		{fltint.fhex = accum; printf("%3f ",fltint.flt_dat);}
	  else if(mode1 == T_DOL)
		{fltint.fhex = accum; printf("%5.2f ",fltint.flt_dat);}
	  else printf("%3ld ",accum);
     
	  if(colcnt++ >= NOOFCOL) {printf("\n"); colcnt = 1;}
     }
}
/* }     */
}

/* same as pdump, but dumps only 2 rows (32 PEs) */
void minipdump(int arg1, int size, int mode1)
{
 int i,j,k,mask,colcnt;
 unsigned long accum,one;
 union forx {
		long int fhex;
		float flt_dat;}
 fltint;

/*
printf("In dpump at %d, arg1=%d, size=%d\n",linenum,arg1,size);
*/
/* if(size > 32) xdump(arg1,size,T_HEX);
 else
 {                                  931030    */
 colcnt = 1;
 one = 1;
 for(j=0;j<2 /*NOOFW*/;j++)
   {printf("PE%3d:",j*NOOFCOL);
	//printf("%d arg1=%d pa_mem=%p PE%2d:",
	//          linenum,arg1,&pa_mem[j][arg1],j*SIZE_PACK);
	for(k=SIZE_PACK-1;k>=0;k--)
	 {//020505 SIZE_PACK is defined as 16, and seems to work elsewhere, 
	  // but the mask is for 32 bit words, thus the +16
	  mask=1<<(k+16);
	  accum = 0;
	  /* arg1 + 0123456789... */
	  for(i=0;i<size;i++)
	   {//printf("mask=%x, pa_mem[%d][%d]=%x",mask,j,arg1+i,pa_mem[j][arg1+i]);
	    if(mask & pa_mem[j][arg1+i])
		  accum += one<<(size-i-1);
		//printf(", accum=%d\n",accum);
	   }

	  if(mode1 == T_HEX) printf("%8lx ",accum);
	  else if(mode1 == T_CHAR)
		{/*for(i=3; i>=0; i--){tl = accum>>i*8;printf("%1c",tl);}*/
		 printf("%8lx",accum);
		 printf(" ");}
	  else if(mode1 == T_REAL)
		{fltint.fhex = accum; printf("%3f ",fltint.flt_dat);}
	  else if(mode1 == T_DOL)
		{fltint.fhex = accum; printf("%5.2f ",fltint.flt_dat);}
	  else printf("%3ld ",accum);
     
	  if(colcnt++ >= NOOFCOL) {printf("\n"); colcnt = 1;}
     }
}
/* }     */
}

xdump(arg1,size,mode1)
int arg1,size,mode1;
{
 int i,j,k,l,mask,isize,rsize;
 unsigned long accum,one;
/*
 union forx {
		long int fhex;
		float flt_dat;}
 fltint;
*/
/* printf("In dpump at %d, arg1=%x, size=%d\n",linenum,arg1,size);*/

 one = 1;
 for(j=0;j<NOOFW;j++)
   {printf("PE%2d:",j*SIZE_PACK);
/*printf("%d arg1=%d pa_mem=%p PE%2d:",linenum,arg1,&pa_mem[j][arg1],j*SIZE_PACK);*/
	 for(k=SIZE_PACK-1;k>=0;k--)
	   {mask=1<<k;
	     accum = 0;
	     isize = size / 32;
	     rsize = size % 32;
	     for(i=0;i<rsize;i++)
	       if(mask & pa_mem[j][arg1+i])
			accum += one<<(rsize-i-1);
	     printf("%8lx",accum);
	     accum = 0;
	     for(l=0;l<isize;l++)
	       for(i=0;i<32;i++)
		 if(mask & pa_mem[j][arg1+i+rsize+l*32])
		    accum += one<<(32-i-1);
	     printf("%8lx",accum);
	     printf(" ");
	    }
    printf("\n");}
}

/*pdump(arg1,size,mode1)    masked pdump
int arg1,size,mode1;
{
 int i,j,k,mask,rm;
 unsigned long accum,one;

 one = 1;
 for(j=0;j<NOOFW;j++)
   {printf("PE %2d:  ",j*SIZE_PACK);
	 for(k=SIZE_PACK-1;k>=0;k--)
			   {mask=1<<k;
			    if((rm=pa_mem[j][MREG] & mask)!=0)
	       {accum = 0;
					 for(i=0;i<size;i++)
					   if(rm & pa_mem[j][arg1+i])
							accum += one<<(size-i-1);
					 if(mode1 == T_HEX || mode1 == T_CHAR) printf("%8lx ",accum);
				    else printf("%3ld ",accum);
	       }
				 else
	       printf(" 00 ");
				}
    printf("\n");}
}
*/


#define MAX_LEVEL    6
#define SZSC         4
#define MORSTR       1
#define LEFT_PAREN  40
#define RIGHT_PAREN 41
#define RIGHT_CB   125
#define COMMA       44
#define BLANK       32
#define EOL          0
#define EOS          0
#define TKBF_SZ     32

int tkbf[TKBF_SZ];
int tktp;

char *strip(char *pnt)
{
	while (*pnt == BLANK) pnt++;
	return pnt;
}

int cm_cnt;
char tmpbuf[IOSTACK_SIZE];

char *inspace(char *strbuf)
/* insert white spaces and add structure- b[$] -> b ([ $ ]) */
/* leave += as +=, and b$ as b$ */
/* delete comments */
{
   char chr, pchr, *pnt;

   pchr = 0;
   pnt = tmpbuf;
   while((chr = *strbuf++) != 0)
      {if(chr == C_SLASH && *strbuf == C_MULT) {cm_cnt++; chr = *strbuf++;}
       if( cm_cnt>0)
	 {if(chr == C_MULT && *strbuf == C_SLASH) cm_cnt--;
	  while (cm_cnt > 0)
	    {if(*strbuf == C_SLASH && *(strbuf+1) == C_MULT ) cm_cnt++;
	     if(*strbuf == C_MULT  && *(strbuf+1) == C_SLASH) cm_cnt--;
	     if(*strbuf == 0) break;
	     strbuf++;}
	  if (*strbuf != 0) {strbuf++; cm_cnt=0;}
	  }
       else
			{if(chr == C_LSQB) *pnt++ = C_LPRN;
	  if(chr == C_RSQB || chr == C_LSQB || chr == C_SEMI || chr == C_COLON)
	     *pnt++ = C_WS;
	  if(chr == C_ASG )
	    if(pchr != C_ADD  && pchr != C_SUB &&
					pchr != C_MULT && pchr != C_DIV)     *pnt++ = C_WS;
	    if(chr == C_ADD  || chr == C_SUB   ||
					chr == C_MULT || chr == C_DIV)       *pnt++ = C_WS;
	    *pnt++ = chr;  /* above inserts before char, below after */
	    if(chr == C_RSQB) *pnt++ = C_RPRN;
	    if(chr == C_RSQB  || chr == C_LSQB ||
					chr == C_DOL   || chr == C_ASG  ||
	       chr == C_COLON || chr == C_SEMI)     *pnt++ = C_WS;
	    if(chr == C_ADD   || chr == C_SUB  ||
					chr == C_MULT  || chr == C_DIV)
	      if(*strbuf != C_ASG) *pnt++ = C_WS;
	      pchr = chr; }
	  }
   *pnt = 0;
   return(tmpbuf);
}



char *gtk(char * pnt)
{
   int i;

   if(*pnt == EOS) tktp = MORSTR;
   else
     {i=0;
      tktp = (int)*pnt;
      if(*pnt == COMMA       || *pnt == LEFT_PAREN ||
			*pnt == RIGHT_PAREN || *pnt == RIGHT_CB)
	{tkbf[i++] = *pnt++; tkbf[i] = 0;}
      else
	while (*pnt != BLANK && *pnt != COMMA      && *pnt != RIGHT_PAREN &&
	       *pnt != EOS   && *pnt != LEFT_PAREN && i <= TKBF_SZ        &&
	       *pnt != RIGHT_CB)
	   tkbf[i++] = *pnt++;
      tkbf[i] = 0;}
   return pnt;
}

#define C_RSQB 93
#define C_LSQB 91

char *gtkp(char *pnt) /* modified for prolog, [a|b] */
{
   int i;

   if(*pnt == EOS) tktp = MORSTR;
   else
     {i=0;
      tktp = (int)*pnt;
      if(*pnt == COMMA    || *pnt == LEFT_PAREN || *pnt == RIGHT_PAREN ||
	 *pnt == RIGHT_CB || *pnt == C_RSQB     || *pnt == C_LSQB      ||
	 *pnt == C_BAR)
	{tkbf[i++] = *pnt++; tkbf[i] = 0;}
      else
	while (*pnt != BLANK    && *pnt != COMMA      && *pnt != RIGHT_PAREN &&
	       *pnt != EOS      && *pnt != LEFT_PAREN && i <= TKBF_SZ        &&
	       *pnt != RIGHT_CB && *pnt != C_RSQB     && *pnt != C_LSQB      &&
	       *pnt != C_BAR)
	     tkbf[i++] = *pnt++;
      tkbf[i] = 0;}
   return pnt;
}

int em_scin (int type)
/* scin_  atom,sc,mask,iobuf,maskout
   type is 1 for normal scin,
   type is 2 for list scin (i.e. with explicit nils),
   type is 3 for prolog scin (i.e. predicate dominates arguments)
		 lower case are variables must be done externally
		 commas are ignorred
   type is 4 for 8 bit structure codes
   type is 5 for compiler input,
		 i) "," are atoms, not place holders
		ii) a preprocessor inserts blanks between "[", etc
	       iii) a "[" is an atom and a structure marker
		iv) line numbers saved in line number field
   atom is argument of atom field
   sc is argument of structure code field
   mask is address of bi field in which items are to be allocated
   iobuf is the addres of the scalar memory buffer */
{
   int atom,sc,mask,iobuf,mout,head,phead, psize, pred, cid, functor,
       cidsiz, ccnt, ln;
   int position, level, itoken[TOK_SIZ];
   int add, i, j, tkn, asize, aasize, tsize, scsize,sclevels;
   int asizwd,csizwd,psizwd; /* sizes in words */
   int kend,k;
   char *strpt, nstrbuf[200];
   int ZERO, SPACES, nil, dsize, hd, ONE, nl;
   int bar[8];

  if(type == 3)
     {em_scano_9();
      em_carg_9(&atom,&sc,&mask,&phead,&head,&cid,&functor,&iobuf,&mout);}
  else if(type == 5)
     {em_scano_6();
      em_carg_6(&atom,&sc,&mask,&nl,&iobuf,&mout);}
  else
     {em_scano_5();
      em_carg_5(&atom,&sc,&mask,&iobuf,&mout);}
  em_stack_buf();
  sv_perform = sc_mem[get_add("PERFORM")];
  sc_mem[get_add("PERFORM")] = 0;
  if(noex_bl<=0)
    {asize = get_size(carg1);
     scsize = get_size(carg2);
     cidsiz = 1;
     psize = 1;
     if(type == 3)
       {psize= get_size(carg4);
	cidsiz = get_size(carg6);}
     dsize = SZSC;
     asizwd = ((asize-1)/SIZE_PACK)+1;
     csizwd = ((cidsiz-1)/SIZE_PACK)+1;
     psizwd = ((psize-1)/SIZE_PACK)+1;
     position = iobuf;  /* use iobuffer for temp storage for scin */
     tkn = position+1;     /* the "size" of the item is reflected */
     ZERO = tkn+asizwd;    /* in the increment of the next item */
     SPACES = ZERO+1;
     nil = SPACES+asizwd;
     hd = nil+asizwd;
     ccnt = hd+psizwd;
     ONE = ccnt+csizwd;
     ln = ONE+1;
     for(i=0; i<asizwd; i++) sc_mem[tkn+i] = 0;
     sc_mem[ZERO] = 0;
     for(i=0; i<asizwd; i++) sc_mem[SPACES+i] = BLANK;
     for(i=0; i<asizwd; i++) sc_mem[nil+i] = 0;
	  sc_mem[nil+1] = 110;  /* 6e */
     sc_mem[nil] = 26988;  /* 696c */
     for(i=0; i<psizwd; i++) sc_mem[hd+i] = 0;
     for(i=0; i<csizwd; i++) sc_mem[ccnt+i] = 0;
	  sc_mem[ccnt] = 1;
     sc_mem[ONE] = 1;
     sc_mem[ln] = 1;
     pred = 0; /* set predicate flag to false */
     cm_cnt = 0; /* set comment count to 0 */
     if(type == 4 || type == 5) dsize=8;
     /* use IOSTACK for temp parallel storage */
     /* 0 for original mask, 1,2,3,4 for type == 3  */
     lgpanot(mask,IOSTACK,1); /* save complement of original mask */
     lgpanot(mask,IOMASK,1);
     paclrmk(atom,IOMASK,asize);
     paclrmk(sc,IOMASK,scsize);
     sclevels = scsize/dsize;
     level = -1;
     if(type == 1 || type == 2 ||  type == 4 || type == 5) add = sc - dsize;
     else if (type == 3) add = sc;
     if(type == 1 || type == 4) printf("\nEnter list:\n");
     else if (type == 2)  printf("\nEnter Lisp list:\n");
     else if (type == 3)  printf("\nEnter Prolog list:\n");
     else if (type == 5)  printf("\nEnter source program:\n");
     if(gets(nstrbuf) == NULL || strlen(nstrbuf) <= 0)
       {paclrun(mout,1);  /* set result to 0 */
	sc_mem[get_add("PERFORM")] =  sv_perform;
	return(FAIL);}
     if(type == 5) strcpy(nstrbuf,inspace(nstrbuf));
     strpt = strip(nstrbuf);
  loop:
     sc_mem[position] = 1; /* reset position to 1st digit */
     tktp = 100; /* must be set to not FAIL */
     while (tktp != FAIL && tktp != EOL && level <= MAX_LEVEL)
       {if(type != 3) strpt = gtk(strpt); else strpt = gtkp(strpt);
	strpt = strip(strpt);
	i = 0;
		  kend=(SIZE_PACK-1)/8;
	for(j=asizwd-1; j>=0; j--)
	  {itoken[j] = 0;
			  k=kend;
	   while(tkbf[i] != EOS && tktp != EOL && k>=0)
	     {itoken[j] |= (tkbf[i]&255)<<k*8;
	      i++; k--;
				  }
	   }
	lgpanot(mask,IOMASK,1);
	switch(tktp)
	  {case MORSTR:
		gets(nstrbuf);
		sc_mem[ln] = 1 + sc_mem[ln]; /* increment line number */
		if(type == 5) strcpy(nstrbuf,inspace(nstrbuf));
		strpt = strip(nstrbuf);
		break;
	   case LEFT_PAREN:
					 if(type == 1 || type == 2 || type == 4 || type == 5)
		  if(level>=0) scpamvmk(position,add,IOMASK,dsize,dsize);
		if(type == 3)
		  {for(i=0; i<NOOFW; i++)pa_mem[i][IOSTACK+4]=0;
		   i=first(IOSTACK+3);
		   pa_mem[i][IOSTACK+4]=pa_mem[i][IOSTACK+3];
		   scpamvmk(ONE,functor,IOSTACK+4,1,1);}
		em_stack_sc(position);
		sc_mem[position]=1;
		add += dsize;
		level++;
		if(level >= sclevels)
		   printf("ERROR, number of structure code digits exceeded\n");
		break;
	   case RIGHT_CB:
					 strpt--; /* exit trhu right paren */
	   case RIGHT_PAREN:
					 if(type == 2)
		  {scpamvmk(position,add,IOMASK,dsize,dsize);
		   scpamvmk(nil,atom,IOMASK,asize,asize);
		   balloc(emptys,mask,1);
		   lgpanot(IOMASK,IOMASK,1);
		   lgpaor(IOMASK,emptys,IOMASK,1);
		   lgpanot(IOMASK,IOMASK,1);
		   sc_mem[position] = sc_mem[position] + 1;}
		scpamvmk(ZERO,add,IOMASK,dsize,dsize);
		scpamvmk(SPACES,atom,IOMASK,asize,asize);
		if(level>=0) em_pop_sc(position);
		  add -= dsize;
		if(level <=0) tktp=EOL;
		if(type == 1 || type == 2 || type == 4 || type == 5)
		   sc_mem[position] = sc_mem[position] + 1;
		level--;
		break;
	   case C_LSQB:
		scpamvmk(position,add,IOMASK,dsize,dsize);
		aasize = 0; i=0;
		tsize=SIZE_PACK;
					 if(asize<tsize,tkn,atom) tsize = asize;
		if(type == 3)
		  {for(i=0;i<asizwd;i++) itoken[0]=0;
						 itoken[asizwd-1] = C_DOT<<(SIZE_PACK-8);}
		infname(asizwd,aasize,asize,itoken,type,tsize,tkn,atom);
		if(type == 5)
						{sc_mem[tkn] = level+1;
		   scpamvmk(tkn,sc+scsize-dsize,
		   IOMASK,dsize,dsize); }
		balloc(emptys,mask,1);
		lgpanot(IOMASK,IOMASK,1);
		lgpaor(IOMASK,emptys,IOMASK,1);
		lgpanot(IOMASK,IOMASK,1);
		sc_mem[position] = sc_mem[position] + 1;
		bar[position] = FALSE;
		if(type == 3)
		  {em_stack_sc(position);
		   sc_mem[position]=1;
		   add += dsize;
		   level++;
		   if(level >= sclevels)
		      printf("ERROR, number of structure code digits exceeded\n");}
		break;
	     case C_RSQB:
						if(type != 3 || bar[position] == TRUE)
		    {scpamvmk(position,add,IOMASK,dsize,dsize);
		     aasize = 0; i=0;
		     tsize=SIZE_PACK;
							if(asize<tsize,tkn,atom) tsize = asize;
		     if(type == 3 && bar[position] == TRUE)
		       {for(i=0;i<asizwd;i++) itoken[0]=0;
								itoken[asizwd] = C_AT<<(SIZE_PACK-8);}
		     infname(asizwd,aasize,asize,itoken,type,tsize,tkn,atom);
		     if(type == 5)
						     {sc_mem[tkn] = level+1;
			scpamvmk(tkn,sc+scsize-dsize,
			IOMASK,dsize,dsize); }
		     balloc(emptys,mask,1);
		     lgpanot(IOMASK,IOMASK,1);
		     lgpaor(IOMASK,emptys,IOMASK,1);
		     lgpanot(IOMASK,IOMASK,1);
		     sc_mem[position] = sc_mem[position] - 1;}
		  if(type == 3)
		    {scpamvmk(ZERO,add,IOMASK,dsize,dsize);
		     scpamvmk(SPACES,atom,IOMASK,asize,asize);
		     if(level>=0) em_pop_sc(position);
		       add -= dsize;
		     if(level <=0) tktp=EOL;
		       level--;}
		  break;
	     case C_BAR:
						if(type == 3) bar[position] = TRUE;
		  else
		    {scpamvmk(position,add,IOMASK,dsize,dsize);
		     aasize = 0; i=0;
		     tsize=SIZE_PACK;
							if (asize<tsize,tkn,atom) tsize = asize;
		     infname(asizwd,aasize,asize,itoken,type,tsize,tkn,atom);
		     if(type == 5)
							  {sc_mem[tkn] = level+1;
			scpamvmk(tkn,sc+scsize-dsize,
			IOMASK,dsize,dsize); }
		     balloc(emptys,mask,1);
		     lgpanot(IOMASK,IOMASK,1);
		     lgpaor(IOMASK,emptys,IOMASK,1);
		     lgpanot(IOMASK,IOMASK,1);
		     sc_mem[position] = sc_mem[position] + 1;}
		  break;
	     /* COMMA must immediately preceed default */
	     case COMMA:
						if(type != 5)
		    {if(type == 3) break;
		     sc_mem[position] = sc_mem[position] + 1;
		     break;}
	     default:
		  scpamvmk(position,add,IOMASK,dsize,dsize);
		  aasize = 0; i=0;
		  tsize=SIZE_PACK;
		  if(asize<tsize,tkn,atom) tsize = asize;
		  infname(asizwd,aasize,asize,itoken,type,tsize,tkn,atom);
		  if(type == 3 && !pred)
		    {/*scpamvmk(hd,phead+SIZE_PACK?,IOMASK,psize,psize);*/
		     for(i=0;i<asizwd;i++)
		       sc_mem[hd+i] = sc_mem[tkn+i];
		     pred = 1;}
		  if(type == 5)
					     {sc_mem[tkn] = level+1;
		     scpamvmk(tkn,sc+scsize-dsize,IOMASK,dsize,dsize);
		     scpamvmk(ln,nl,IOMASK,get_size(carg4),get_size(carg4)); }
		  balloc(emptys,mask,1);
		  papamvun(IOMASK,IOSTACK+3,1);/*save for functor*/
		  lgpanot(IOMASK,IOMASK,1);
		  lgpaor(IOMASK,emptys,IOMASK,1);
		  lgpanot(IOMASK,IOMASK,1);
		  sc_mem[position] = sc_mem[position] + 1;
		  break;
		}
   }/* endwhile */
  /* if prolog type, check for :- and , and . */
  if(type == 3)
    {strpt = gtkp(strpt);
     strpt = strip(strpt);
     if(tkbf[0]==':' && tkbf[1]=='-')
       {/*sc_mem[hd] = sc_mem[tkn];*/
	lgpanot(IOMASK,IOSTACK+1,1);  /* use IOSTACK+1 as temp storage */
	lgpaand(IOSTACK+1,IOSTACK,IOSTACK+2,1);
	lgpaor(IOSTACK+2,head,head,1);
	goto loop;}
     if(tkbf[0]==',')
       {lgpaor(IOMASK,IOSTACK+1,IOSTACK+1,1);
	lgpanot(IOSTACK+1,IOSTACK+1,1);
	scpamvmk(ccnt,cid,IOSTACK+1,cidsiz,cidsiz);
	lgpanot(IOMASK,IOSTACK+1,1);  /* use IOSTACK+1 as temp storage */
	sc_mem[ccnt]++;
	goto loop;}
     if(tkbf[0]=='.')
       {lgpaor(IOMASK,IOSTACK+1,IOSTACK+1,1);
	lgpanot(IOSTACK+1,IOSTACK+1,1);
	scpamvmk(ccnt,cid,IOSTACK+1,cidsiz,cidsiz);
	lgpanot(IOMASK,IOSTACK+1,1);
	lgpaand(IOSTACK,IOSTACK+1,IOSTACK+1,1);
	scpamvmk(hd,phead,IOSTACK+1,psize,psize);}
    }
  }
  lgpanot(IOMASK,IOMASK,1);
  lgpaand(IOMASK,IOSTACK,mout,1); /* return allocated memory */
  sc_mem[get_add("PERFORM")] =  sv_perform;
/*  farfree(nstrbuf);*/
  return(SUCCESS);
}

void em_ftcd(code1,code2,sib1,sib2)
long code1,code2,*sib1,*sib2;
/* fstcd returns the first sibling of the input structure */
/* code1 is left four digits, code2 is right three */
{
  int fstdgt;

  fstdgt = em_scsize(code1,code2);
  if(fstdgt > 7) printf("ERROR - size of sc exceeded.\n");
  else if(fstdgt > 3)
    {*sib1 = code1;
     *sib2 = code2 | 1<<((7-fstdgt)*8);}
  else
    {*sib1 = code1 | 1<<(4-fstdgt)*8;
     *sib2 = code2;}
}


void em_scot(int type)
/* em scot produces a list output given a structure code and mask input
   type = 1 produces a normal list from normal input
   type = 2 produces a normal list form explicit list input
   type = 3 produces a prolog list form prolog input
   type = 4 parsed io for scanner
   type = 5 source code, i.e. remove structure from [ ..]
*/
/* scot_  atom,sc,mask,iobuf */
{
   int atom,sc,mask,iobuf,length,linel,lensc;
   int tabcnt, oldtab;
   int i, asize, ssize, aasize;
   int  tnil;
   int oldl,dd,shrtl,frstsc,tsc,tvalue[TOK_SIZ],tsc2;
   long int foo1,foo12,foo2,foo22,oldsc,oldsc2,scode,scode2;
   int skip_lsqr,skip_rsqr,prnt_bar, prnt_rsqb, fnd_at, tmpdx;

  em_scano_5();
  em_carg_4(&atom,&sc,&mask,&iobuf);
  em_stack_buf();
  sv_perform = sc_mem[get_add("PERFORM")];
  sc_mem[get_add("PERFORM")] = 0;
  if(noex_bl<=0)
    {oldsc = 0;
     oldsc2 = 0;
     oldl = 0;
     for(i=0; i<TOK_SIZ; i++)
	tvalue[i] = iobuf+i;
     tsc =  iobuf+TOK_SIZ;
     tnil = iobuf+TOK_SIZ+1;
     tsc2 = iobuf+TOK_SIZ+2;
     asize = get_size(carg1);
     ssize = get_size(carg2);
     prnt_bar=0;
     prnt_rsqb=0;
     fnd_at=0;
     sc_mem[tnil] = L_AT;
     tmpdx = get_add("TEMP");
     fxcreq(atom,tnil,tmpdx,asize);
     if(em_responders(tmpdx) != 0)fnd_at=1;
     sc_mem[tnil] = 7235948;
     if(type == 1 || type == 2 || type == 4 || type == 5) length = 0;
     if(type == 3) length = 1;
     linel = 0;
	  tabcnt = 0;
	  oldtab = 0;
     papamvun(mask,IOSTACK,1);
     papamvun(mask,IOMASK,1);
     mindex(sc,IOMASK,IOSTACK,ssize);
     while( em_responders(IOMASK) != 0)
       {/* first(IOMASK); */
	lgpaxor(IOMASK, IOSTACK, IOSTACK, 1);
	if(type != 4 && type != 5)
	  {/*pascmv(sc,tsc,IOMASK,ssize);*/
	   pascmv(sc,tsc,IOMASK,ssize-SIZE_PACK);
			  scode = sc_mem[tsc];
			  scode <<= SIZE_PACK;
	   pascmv(sc+ssize-SIZE_PACK,tsc,IOMASK,SIZE_PACK);
	   scode |= sc_mem[tsc];
	   lensc = scsize(scode);
	   foo1 = fstcd(scode);}
	else
	  {pascmv(sc,tsc,IOMASK,ssize-3*SIZE_PACK);
	   scode = sc_mem[tsc];
			  scode <<= SIZE_PACK;
	   pascmv(sc+SIZE_PACK,tsc,IOMASK,SIZE_PACK);
	   scode |= sc_mem[tsc];
	   pascmv(sc+2*SIZE_PACK,tsc,IOMASK,SIZE_PACK);
	   scode2 = sc_mem[tsc];
			  scode2 <<= SIZE_PACK;
	   pascmv(sc+3*SIZE_PACK,tsc,IOMASK,SIZE_PACK);
	   scode2 |= sc_mem[tsc];
	   scode2 &= 0xffffff00;
	   lensc = em_scsize(scode,scode2);
	   em_ftcd(scode,scode2,&foo1,&foo12);}
	aasize=asize;
	if(aasize > SIZE_PACK) aasize = SIZE_PACK;
	for(i=0; i<TOK_SIZ; i++) sc_mem[tvalue[i]]=0;
	i=0;
	while(i<TOK_SIZ && aasize*i<asize)
	  {pascmv(atom+aasize*i,tvalue[i],IOMASK,aasize);
	   i++;}
	skip_lsqr = 0; skip_rsqr = 0;
	if(type == 5)
	  {if(sc_mem[*tvalue] == L_LSQB) skip_lsqr = 1;
	   if(sc_mem[*tvalue] == L_RSQB) skip_rsqr = 1;}
	if(type == 3)
	  {if(sc_mem[*tvalue] == L_AT )
	     {sc_mem[*tvalue]=L_RSQB; skip_rsqr = 1;}
	   if(sc_mem[*tvalue] == L_DOT)
	     {sc_mem[*tvalue]=L_LSQB; skip_lsqr = 1;
	      if(fnd_at == 1) prnt_bar=2; else prnt_rsqb=1;}
			 }
	if(em_responders(IOSTACK)!=0)
	  {mindex(sc,IOMASK,IOSTACK,ssize-8);
	   if(type != 4 && type != 5)
	     {pascmv(sc,tsc,IOMASK,ssize-SIZE_PACK-8);
			     foo2 = sc_mem[tsc];
			     foo2 <<= SIZE_PACK;
	      pascmv(sc+ssize-SIZE_PACK-8,tsc,IOMASK,SIZE_PACK);
	      foo2 |= sc_mem[tsc];}
	   else
	     {pascmv(sc,tsc,IOMASK,ssize-3*SIZE_PACK);
	      foo2 = sc_mem[tsc];
			     foo2 <<= SIZE_PACK;
	      pascmv(sc+SIZE_PACK,tsc,IOMASK,SIZE_PACK);
	      foo2 |= sc_mem[tsc];
	      pascmv(sc+2*SIZE_PACK,tsc,IOMASK,SIZE_PACK);
	      foo22 = sc_mem[tsc];
			     foo22 <<= SIZE_PACK;
	      pascmv(sc+3*SIZE_PACK,tsc,IOMASK,SIZE_PACK);
	      foo22 |= sc_mem[tsc];
	      foo22 &= 0xffffff00;}
			 }
	else
	  {foo2 = 0; foo22 = 0;}
	if(type == 3 && foo1 == foo2)
	  {while(length>lensc) {printf(")"); length--; linel++;}
	   linel += pvalue(*tvalue);
	   printf("(");
	   length++;
	   linel++;
	   oldl = scsize(foo1);}
	else
	  {shrtl = lensc;
	   if(shrtl > oldl) shrtl=oldl;
	   if(type != 4 && type != 5)
	     frstsc = firstsc(scode^oldsc);
	   else
	     frstsc = em_firstsc(scode^oldsc,scode2^oldsc2);
	   if(prnt_rsqb == 1 && length>lensc)
				  {printf("]"); linel++; prnt_rsqb = 0; length--;}
			  if(type == 5)
				  {tabcnt = lensc-length;}
			  else
			    while(length>lensc) {printf(")"); length--; linel++;}
	   dd = shrtl-frstsc;
	   if(type == 5) tabcnt = tabcnt+dd;
			  else
	     while(dd>00) {printf(")"); dd--; linel++;}
	   dd = shrtl-frstsc;
	   if(type == 5) tabcnt = tabcnt-dd;
			  else
	     while(dd>00) {printf("("); dd--; linel++;}
	   if(skip_lsqr){skip_lsqr=0; length++;}
	   if(type != 5)
				 while(length<lensc) {printf("("); length++; linel++;}
	   if(type == 5 && oldtab != tabcnt)
				 {printf("\n");
				  for(i=0; i<tabcnt; i++)
					 {printf("  "); oldtab=tabcnt;}
				 }
/* must be modified to output any size field */
	   if(sc_mem[tvalue[0]] != sc_mem[tnil] || type != 2)
				  linel += (pvalue(*tvalue)+1);
	   if(prnt_bar==1) {printf(" |"); prnt_bar=0;}/* print | for type = 3 */
	   if(prnt_bar > 1) prnt_bar--;
	   if(skip_rsqr){skip_rsqr=0; length--;}
	   oldl = lensc;
						   }
	oldsc = scode;
	oldsc2 = scode2;
	if(linel > 60) {printf("\n"); linel=0;}
	if(em_responders(IOSTACK)) mindex(sc,IOMASK,IOSTACK,ssize);
	else break;
		}
	  if(type != 5)
	while (length>1) {printf(")"); length--;}
     if(type == 1 || type == 2 || type == 4) printf(")");
     printf("\n");
     sc_mem[get_add("PERFORM")] =  sv_perform;
     return;
    }
}

pvalue(temp)
int temp;
{
  char chr;
  int count,i;

  count = 0;
  printf(" ");
  for(i=0; i<8; i++)
    {if(sc_mem[temp] != 0)
       {if(SIZE_PACK >= 32)
		    {chr = (sc_mem[temp] & 0xff000000)/0x1000000;
	   if(chr >= BLANK) {count++; printf("%c",chr);}}
	if(SIZE_PACK >= 24)
		    {chr = (sc_mem[temp] & 0xff0000)/0x10000;
	   if(chr >= BLANK) {count++; printf("%c",chr);}}
	if(SIZE_PACK >= 16)
		    {chr = (sc_mem[temp] & 0xff00)/0x100;
	   if(chr >= BLANK) {count++; printf("%c",chr);}}
	if(SIZE_PACK >= 8)
		    {chr = (sc_mem[temp] & 0xff);
	   if(chr >= BLANK) {count++; printf("%c",chr);}}
      }
   temp++;}
  return(count);
}


/* scsize returns the size of the structure, i.e. the
   location of the first (left most) zero digit.     */
int scsize(long code)
{
/*  if(!(code & 0x0fffffff)) return(1);
    if(!(code & 0x00ffffff)) return(2);  */
    if((code & 0x00ffffff) == 0) return(0);
    if((code & 0x000fffff) == 0) return(1);
    if((code & 0x0000ffff) == 0) return(2);
    if((code & 0x00000fff) == 0) return(3);
    if((code & 0x000000ff) == 0) return(4);
    if((code & 0x0000000f) == 0) return(5);
    if((code & 0x00000000) == 0) return(6);
}


int firstsc(long code)
/* firstsc returns the first (left  most) non-
   zero digit of the xor of two structure codes - the first
   digit where the two structures differ   */
{
    if(code == 0) return(0);
/*  if(code & 0xf0000000) return(1);
    if(code & 0x0f000000) return(2); */
    if((code & 0x00f00000) != 0) return(1);
    if((code & 0x000f0000) != 0) return(2);
    if((code & 0x0000f000) != 0) return(3);
    if((code & 0x00000f00) != 0) return(4);
    if((code & 0x000000f0) != 0) return(5);
    if((code & 0x0000000f) != 0) return(6);
    return(0);
}


long fstcd(long code)
/* fstcd returns the first sibling of the input structure */
{
  int fstdgt,i;
  long mask;

  fstdgt = scsize(code);
  mask = 1;
  for(i=5; i>fstdgt; i--) mask = mask*16;
  mask = mask | code;
  return(mask);
}


/* em-scsize returns the size of the structure, i.e. the
   location of the first (left most) zero digit.     */
/* code1 is left four digits, code2 is right three */
int em_scsize(long code1, long code2)
{
  if(((code1 & 0x00ffffff) == 0) && ((code2 & 0xffffffff) == 0)) return(1);
  if(((code1 & 0x0000ffff) == 0) && ((code2 & 0xffffffff) == 0)) return(2);
  if(((code1 & 0x000000ff) == 0) && ((code2 & 0xffffffff) == 0)) return(3);
  if(((code2 & 0xffffffff) == 0)) return(4);
  if(((code2 & 0x00ffffff) == 0)) return(5);
  if(((code2 & 0x0000ffff) == 0)) return(6);
  if(((code2 & 0x000000ff) == 0)) return(7);
 return(7);  /* error return */
}

int em_firstsc(long code1, long code2)
/* firstsc returns the position of the first (left  most) non-
   zero digit of the xor of two structure codes - the first
   digit where the two structures differ   */
{
  if(code1 == 0 && code2 == 0) return(0);
  if((code1 & 0xff000000) != 0) return(1);
  if((code1 & 0x00ff0000) != 0) return(2);
  if((code1 & 0x0000ff00) != 0) return(3);
  if((code1 & 0x000000ff) != 0) return(4);
  if((code2 & 0xff000000) != 0) return(5);
  if((code2 & 0x00ff0000) != 0) return(6);
  if((code2 & 0x0000ff00) != 0) return(7);
  return(8);
}

long int leftadj(word,size)
long int word;
int size;
/* leftadj left adjusts the token in word. That is, if there
   are leading 0s, they are shifted out. But shifting is
   restricted by the size of the field.
*/
{
   unsigned long mask;
   int i;

   i = 0;
   if(size > SIZE_PACK) size=SIZE_PACK;
   mask = 255<<(size-8);
   while((word&mask)==0 && i<(SIZE_PACK)/8)
     {word <<= 8;
      i++;}
   return(word);
}

void infname(int asizwd, int aasize, int asize,int itoken[], int type, int tsize, int tkn, int atom)
/* insert functor name */
{int j;

   while(aasize<asize)
     {for(j=0; j<asizwd;j++)
		  sc_mem[tkn+j] =  itoken[j];
      scpamvmk(tkn,atom+aasize,IOMASK,tsize,tsize);
      aasize +=  tsize;}
}

/* em_encode converts an integer in sc_mem[arg1] into a
   left adjusted ASCII char string in sc_mem[arg2]      */
int em_encode(int mode)
{
int arg1,arg2;
int i, j, k, iend, size1;

 em_scano_2();
 em_carg_2(&arg1,&arg2);
 em_stack_buf();
 if(noex_bl>0) return FAIL;

size1 = get_size(carg1);
iend = (size1 - 1)/SIZE_PACK;
i = k = iend;
while(sc_mem[arg1+i] == 0) i--;
for(j=i; j>=0; j--)
  {encode16(arg1+j, arg2+k); k--;}
for(j=k; j>=0; j--)
  sc_mem[arg2+j] = 0;
return SUCCESS;
}

int encode16(int arg1, int arg2)
{
 char c_buf[6];

 sprintf(c_buf, "%02d", sc_mem[arg1]);
 sc_mem[arg2] = c_buf[0] << 8 | c_buf[1];
 c_buf[2]=0;
 return SUCCESS;
}

/* convert sc_mem[arg1] - left adjusted char(ASCII) variable to
	   sc_mem[arg2] - right adjusted integer variable */
int em_decode(int mode)
{
  int size1, size2, c_size;
  char c_buf[30];
  int i,j,jbeg,iend;
  int arg1, arg2;

 em_scano_2();
 em_carg_2(&arg1,&arg2);
 em_stack_buf();
 if(noex_bl>0) return FAIL;
 size1 = get_size(carg1);
 size2 = size1;
 put_size(carg2,size1);
 iend = (size1-1)/SIZE_PACK;
 i=0;
 while(sc_mem[arg1+i] == 0) i++;
 c_size = i;
 jbeg = (size2-1)/SIZE_PACK;
 for(j= jbeg; j>jbeg-c_size; j--) sc_mem[arg2+j] = 0;
 i = iend;
 for(j= jbeg-c_size; j>=0; j--)
   {c_buf[0] = sc_mem[arg1+i]>>8;
    if(c_buf[0] == 0) {c_buf[0] = sc_mem[arg1+i] & 255; c_buf[1]=0;}
    else {c_buf[1] = sc_mem[arg1+i] & 255; c_buf[2] = 0; }
    decode16(c_buf, arg2+j);
    i--;
   }
return SUCCESS;
}

decode16(c_buf, arg2)
char *c_buf;
int arg2;
{
 int foo;

  sscanf(c_buf, "%2d", &foo);
  sc_mem[arg2] = foo;

 return SUCCESS;
}
