/* EM_CMDS_STMT.C 910415 */

#include "stdarg.h"
#include "stdio.h"
#include "ctype.h"
#include "string.h"
#include "p2_symtab.h"
#include "em_common.h"
#include "p2_common.h"
#include "p2_header.h"

/* quad_bl     - the next quadruple to be read from the quadruple stack.
   source_bl   - flag indicating read from input (INPUT) or stack (STACK).
   save_bl     - flag indicating the input read from input is to be saved
                 on the quadruple stack.
   noex_bl     - flag indicating that the source is inside a body of
                 code for which the responders are false.  Quadruples
                 should not be simulated when this flag is true.
   noex_bl_sk  - for recursive stack structure.
   noex_bl_ext - see below for more details
   quads_bl    - the quadruple stack array.
   loops_bl    - the loop stack array.
   tols_bl     - the top of the loop stack
   toqs_bl     - the top of the quadruple stack

NOTE: flags are not binary.  They are incremented and decremented on
every appropriate quadruple.  They are TRUE if nonzero. False is zero.
*/

/* the control statements execute with noex_bl and save_bl in
   the following way.

BEGIN routine - BWHILE, FOR and LOOP quadruples

	   before	    after
	noex	save	noex	save
	  0	 0	    0	    1+	normal top level
	  0	 1	    0	    1+	normal active loop
	  1	 0	    1+	 0	   top level false
	  1	 1	    1+	 1+	normal exit loop

UNTIL routine - exit condition - sc==TRUE or no_responders
	         continue condition - no change

	   before	    after
	noex	save   noex	save
	  0	 0		             illegal state
	  0	 1       1+	 1
	  1	 0
	  1	 1


END routine

	   before	    after
	noex	save	 noex	save
	  0	 0			           illegal state
	  0	 1	      0	  1	  loop
	  1	 0	      1-	  0	  fall through
	  1 	 1	      1-	  1-	  fall through

FOR statement

	start	 begin	until	  end
	0   0	 0  1	   1  1	  0  0
	0   1	 0  1+	1+ 1	  0- 1-
	1   0  1  0    1+ 0    1- 0
   1   1  1+ 1+   1  1    1- 1-

ANY, ALLOCATE statement
       responders              responders           responders
 any     0    1    elsenany      0    1    endany     0    1
     ------------            ------------         ------------
     0 | 1    0              0 | 1    1           0 | 0    0
noex 1 | 1+   1+             1 | 0    0           1 | 0    0
     1+| 1+   1+             1+| no change        1+| 1-   1-


IF  statement is analyzed differently

		      if		      else		  endif
------------------------------------------------------------
responders	stack(MREG)	xxx		  pop(MREG)
		      arg1->MREG	xxx
-------------------------------------------------------------
no         	stack(MREG)	xxx		  pop(MREG)
responders	arg1->MREG	xxx
	    noex++				     noex--
--------------------------------------------------------------
external    noex++		xxx		  noex--
noex                    xxx
------------------------------------------------------------
responders	stack(MREG)	____		  pop(MREG)
responders	arg1->MREG	MREG->MREG
-------------------------------------------------------------
noresponders stack(MREG)____		  pop(MREG)
responders	 arg1->MREG	MREG->MREG
	     noex++     noex--
-------------------------------------------------------------
responders	 stack(MREG)____		  noex--
noresponders arg1->MREG	MREG->MREG pop(MREG)
			noex++
--------------------------------------------------------------
external     noex++		noex--	  noex--
noex                    noex++
-------------------------------------------------------------
*/

/*
===========================================================================

function name: em_control
input	       : optype	- integer for type of operation
               mode		- integer for mode of operation
output	    : ---
return value : ---

global vars  : input_fp	- pointer to input file descriptor

description	 : This routine handles the intermediate opcodes for the
 		         ASPROL control statements.

 IN ALL OF THE ROUTINES IN THIS MODULE:
      arg1,arg2,.. are the numeric addresses of the variable
      carg1,carg2,.. are the symbolic names of the variables

===========================================================================
*/
/* noex_bl,_ext == 1,1 indicates external noex_bl.  i.e. entire stack
                       construct is inside a no execute block.
   noex_bl,_sk  == 1,1 indicates stackwhile exit (thru stack).  When
                       noex_bl_sk is set, noex_bl must be set to assure
	                    that the code between STACKWHILE and RECURSE in
			    not executed.
   noex_bl,_sk  == 0,0 indicates recurse to top of construct, noex_bl may or
	                1,0 may not be set by bracketed fors, whiles, etc
*/

void em_control(int optype, int mode)
{
	int arg1,arg2,arg3,argt;
	int mode1,mode2;

if(em_debug == 2)
  printf("In em_control, optype = %x, noex=%d, save=%d, brch=%d, noex_sk=%d\n",
       optype, noex_bl,save_bl,brch_bl,noex_bl_sk);

   mode1 = mode / SIZE_MODE2;
	mode2 = mode % SIZE_MODE2;
   switch(mode2)
	  {case T_REG_SCALAR   : mode2 = T_SCALAR;      break;
	   case T_REG_LEFT     : mode2 = T_COMMON_LEFT; break;
	   case T_REG_RIGHT    : mode2 = T_COMMON_RIGHT;break;
	   case T_SCALAR_LEFT  : mode2 = T_SCALAR;      break;
	   case T_SCALAR_RIGHT : mode2 = T_SCALAR;      break;
		}
	mode = mode1*SIZE_MODE2+mode2;
	switch(optype)
	  {case T_FOR	:
		  em_scano_4();
		  em_carg_2(&arg1,&arg2);
		  em_stack_buf();
		  em_begin();
	if(em_while(arg1)==SUCCESS)
		    {papamvun(arg1,arg2,1);
		     em_stack(arg1);
		     em_stack(MREG);
		     resvfst(arg2);}
		  break;
	   case T_ENDFOR	:
		  em_scano_3();
		  em_carg_1(&arg1);
		  em_stack_buf();
		  if(noex_bl<=0)
		    {step(fillds+1,fillds+1);
		     papamvun(fillds+1,arg1,1);
		     if(em_responders(arg1) == TRUE) resvfst(arg1);}
		  if(em_while(arg1) == TRUE)
		    {em_pop(MREG);
		     em_pop(MREG);
		     papamvun(fillds,MREG,1);}
		  em_end_lp();
		  break;
	   case T_NEXT	:
		  em_scano_3();
		  em_carg_2(&arg1,&arg2);
		  em_stack_buf();
		  if(brch_tst(arg1) >0)
		    {papamvun(arg1,arg2,1);
		     resvfst(arg2);
		     step(arg1,arg1);};
		  break;
	   case T_ENDNEXT	:
		  em_scano_1();
		  em_stack_buf();
		  if(brch_bl>0)
			 {noex_bl--;
			  brch_bl--;};
		  break;
	   case T_GET	:
		  em_scano_3();
		  em_carg_2(&arg1,&arg2);
		  em_stack_buf();
		  if(brch_tst(arg1)>0)
		    {papamvun(arg1,arg2,1);
		     resvfst(arg2);}
		  break;
	   case T_ENDGET	:
		  em_scano_1();
		  em_stack_buf();
		  if(brch_bl>0)
			 {noex_bl--;
			  brch_bl--;};
		  break;
/*   T_ALLOCATE goes to em_allocate in em_quads
	   case T_ALLOCATE	:
		  em_scano_0();
		  em_stack_buf();
		  noex_bl++;
		  brch_bl++;
		  break;
*/
	   case T_ENDALLOCATE	:
		  em_scano_1();
		  em_stack_buf();
		  if(noex_bl>0)
			 noex_bl--;

/*
		  if(brch_bl>0)
			 {noex_bl--;
			  brch_bl--;};
*/
		  break;
	   case T_ANY	:
		  em_scano_2();
		  em_carg_1(&arg1);
		  em_stack_buf();
		  if(noex_bl<=0)
		    {papamvun(fillds,MREG,1);
		     argt = get_add("TEMP2");
		     lgpaand(arg1,MREG,argt,1);
		     em_stack(MREG);
		     if(em_responders(argt) == FALSE) noex_bl++;
		    }
		  else noex_bl++;
		  break;
      case T_ELSENANY :
	em_scano_2();
	em_stack_buf();
	if(noex_bl>0)
	  {noex_bl--;
	     if(noex_bl>0) noex_bl++;
	     else if(mode1 == T_IFNANY )
		/* this code copies the previous scope to the top of the
		   stack, this restores the previous scope of the elsenany
		   part of if, and allows endif to pop the scope off */
		papamvun(fillds+1,MREG,1);}
	else noex_bl++;
	break;
	   case T_BEGWHILE :
		  em_scano_1();
		  em_stack_buf();
		  em_begin();
	  break;
	   case T_WHILE	:
		  em_scano_3();
		  em_carg_2(&arg1,&arg2);
		  em_stack_buf();
		  lgpaand(MREG,arg1,arg1,1); /* and with tos */
		  if(em_while(arg1)==SUCCESS)
		    {papamvun(arg1,arg2,1);
		     resvfst(arg2);}
		  break;
	   case T_ENDWHILE	:
		  em_scano_2();
		  em_stack_buf();
		  em_end_lp();
		  break;
      case T_BEGSTACK:       /* begstack  counter,savemreg,repeat_label */
	     em_scano_3();
	     em_carg_2(&arg1,&arg2);
	     em_stack_buf();
	     sc_mem[arg1] = STACK_STOP;  /* set counter to zero */
	put_size(carg1,SIZE_HOST);
	if(noex_bl<=0)   /* execute code */
		    {save_bl++; /* save input   */
		     papamvun(MREG,arg2,1);  /* save initial state */
		     recurse_noexbl=noex_bl; /* of MREG, noex_bl,  */
		     recurse_savebl=save_bl; /* save_bl and brch_bl*/
		     recurse_brchbl=brch_bl; /* and topofquadruple stack */
		     if(source_bl == INPUT) recurse_stack=toqs_bl;}
	     else
		    {if(save_bl>0) save_bl++;
		       noex_bl++; /* noex_bl and _ext both 1 indicate external */
		       noex_bl_ext++;} /* noex_bl.  i.e. entire stack construct */
	     break;             /* is inside a no execute block.         */
      case T_STACKWHILE:/*stackwhile count,result,index,recurse_label */
	em_scano_4();
		  em_carg_3(&arg1,&arg2,&arg3);
		  em_stack_buf();
	if(noex_bl_ext<=0)
		    if(noex_bl<=0) /* noex_bl=1, _sk=1 indicates stackwhile exit */
	    if(em_responders(arg2)==FALSE)
			     {noex_bl_sk++; noex_bl++; }
	    else {papamvun(arg2,arg3,1); resvfst(arg3); }
		  break;
      case T_RECURSE :   /* recurse count,savemreg,recurse_label,repeat_label */
	     em_scano_4();
	em_carg_2(&arg1,&arg2);
	     em_stack_buf();
/*
printf("At recurse, count[%s,%d]=%d, noex_bl_ext=%d, noex_bl_sk=%d, noex_bl=%d\n",
 carg1, arg1, sc_mem[arg1], noex_bl_ext, noex_bl_sk, noex_bl);
*/
	if(noex_bl_ext<=0)
	  if(noex_bl_sk<=0 && noex_bl<=0) /* if noex_bl_sk = 0 branch to top of stack */
	    {quad_bl=recurse_stack;
		  source_bl = STACK;
	     sc_mem[arg1]++;/* Must be before noex_bl is saved, reset */
/*
printf("     branch to top, count[%s,%d]=%d, source_bl=STACK, quad_bl=%d\n",
carg1,arg1,sc_mem[arg1],quad_bl);
*/
	     em_push_bl(noex_bl);
	     em_push_bl(save_bl);
	     em_push_bl(brch_bl);
	     noex_bl=recurse_noexbl;
	     save_bl=recurse_savebl;
	     brch_bl=recurse_brchbl;
		       em_stack(MREG);     /* perform a STACK_SCOPE */
	     papamvun(arg2,MREG,1);}
	   else if (noex_bl_sk>0 && noex_bl>0)  /* noex_bl=1, _sk=1 - drop thru execute */
	     {sc_mem[arg1]--;
				  noex_bl_sk--; /* bottom of stack construct */
/*              if(quad_bl >= toqs_bl) source_bl = INPUT;*/
/*printf("     drop trhough, count[%s,%d]=%d, source_bl=INPUT, quad_bl=%d\n",
  carg1,arg1,sc_mem[arg1],quad_bl);
*/
		   if(sc_mem[arg1]>=STACK_STOP)
			  {brch_bl = em_pop_bl(); /* noex_bl is not decremented */
		      save_bl = em_pop_bl(); /* because it is restored to  */
		 noex_bl = em_pop_bl(); /* its previous value */
			   em_pop(MREG);}       /* permorm a POP_SCOPE */
	     }
	 else if (noex_bl_sk>0 && noex_bl<=0)
				printf("ERROR in RECURSE, noex_bl=%d, _sk=%d\n",
						  noex_bl, noex_bl_sk);
		   break;
       case T_ENDSTACKWHILE:/*endstack result,savemreg,end_label,repeat_label*/
	      em_scano_4();
	      em_carg_2(&arg1,&arg2);
	      em_stack_buf();
/*
printf("endstackwhile, count[%s,%d]=%d,noex_bl_ext=%d, noex_bl_sk=%d, noex_bl=%d\n",
carg1,arg1,sc_mem[arg1],noex_bl_ext, noex_bl_sk, noex_bl);
*/
	      if(noex_bl_ext<=0)
		{papamvun(arg2,MREG,1); /* restore M no matter what */
		 if(noex_bl<=0) /* check for stack loop branch */
			{if(sc_mem[arg1]>=STACK_STOP)
			   { quad_bl=recurse_stack; source_bl=STACK;}}
			   else noex_bl--;
	    } /* if drop thru, no action since last quad in stack stmt.*/
	 else /* external noex_bl, update both noex_bl and _ext */
	   {noex_bl--;
		 noex_bl_ext--;
	    if(save_bl>0) save_bl--;
	    if(save_bl<=0 && source_bl == INPUT) em_garbage(); };
	    break;
    case T_LOOP :
		   em_scano_1();
		   em_stack_buf();
		   em_begin();
		break;
	    case T_UNTIL 	:
	   em_scano_2();
		   em_carg_1(&arg1);
		   em_stack_buf();
		   em_until(mode1,mode2,arg1);
		   break;
	    case T_ENDLOOP	:
		   em_scano_2();
		   em_stack_buf();
		   em_end_lp();
		break;
		 case T_BEGIF  :
			em_scano_1();
			em_stack_buf();
			if(noex_bl<=0) papamvun(fillds,MREG,1);
			break;
	    case T_IF		:
		   if (mode2 == T_PARALLEL)
		     {em_scano_2();
		 em_carg_1(&arg1);
			   em_stack_buf();
			   if(noex_bl<=0)
		  {lgpaand(arg1,MREG,MREG,1);
			em_stack(MREG);
			      if(em_responders(MREG)==FALSE) noex_bl++;}
		      else noex_bl++;
		     }
		   else
		{em_scano_2();
		      em_carg_1(&arg1);
		      em_stack_buf();
			   if(noex_bl<=0)
				  {if(sc_mem[arg1]<=0)
					  {noex_bl++;
					   brch_bl++;};
			     }
	    else if (brch_bl==noex_bl)
				  {noex_bl++;
				   brch_bl++;};
		     }
		   break;
	    case T_ELSE	:
		   if (mode2 == T_PARALLEL)
		     {em_scano_1();
			   em_stack_buf();
			   if(noex_bl>0) noex_bl--;

			   /* the if part */
			   if(noex_bl<=0)
			     {lgpanot(MREG,MREG,1);
	       lgpaand(fillds+1,MREG,MREG,1);
					papamvun(MREG,fillds,1);
			      if(em_responders(MREG)==FALSE) noex_bl++;}
		      else noex_bl++;}
		   else
			  {em_scano_2();
			   em_stack_buf();
			   if(brch_bl==noex_bl)
			     switch(brch_bl)
			       {case 0:
				       noex_bl++;
				       brch_bl++;
				       break;
				case 1:
				       noex_bl--;
				       brch_bl--;
				       break;
			       };
		     };
		   break;
	    case T_ENDANY   :     /* modified 880608a jlp */
	    case T_ENDIF	:
		   if (mode2 == T_PARALLEL) /* for T_ENDIF only? */
		     {em_scano_1();
			   em_stack_buf();
			   if(noex_bl>0) noex_bl--;
			     if(noex_bl<=0) {em_pop(MREG); papamvun(fillds,MREG,1);}
		     }
		   else
		     {em_scano_1();
			   em_stack_buf();
			   if(brch_bl==noex_bl)
			     switch(brch_bl)
 			        {case 0: break;
			         case 1:
			         default:
					       noex_bl--;
					       brch_bl--;
			        };
		     };
		   break;
	   }
if(em_debug == 2)
  printf("   On exit,  noex=%d, save=%d, brch=%d.\n",noex_bl,save_bl,brch_bl);
}

brch_tst(arg1)
int arg1;
{
	if(noex_bl<=0) if(em_responders(arg1)==FALSE)
	  {noex_bl++;
		brch_bl++;}
	else return TRUE;
	else if (brch_bl>0)
	  {noex_bl++;
		brch_bl++;};
	return FALSE;
}

/* em_begin() performs the following function

     noex_bl  save_bl       noex_bl  save_bl
        0        0             0       1+
        0        1             0       1+
        1        0             1+
        1        1             1+      1+
*/

void em_begin()
{
 if(em_debug == 2)printf("Entering em_begin() with noex_bl=%d, save_bl=%d, called=%d, linenum=%d\n",
		      noex_bl,save_bl,called,linenum);
 if(noex_bl<=0)
   {save_bl++;
    em_push_loop();}
 else /* assures balancing END turns noex_bl off */
   {noex_bl++;
    if(save_bl>0) save_bl++; };
 if(em_debug == 2)printf("Exiting em_begin() with noex_bl=%d, save_bl=%d\n",
		      noex_bl,save_bl);
}

/* em_end_lp() performs the following function

     noex_bl  save_bl       noex_bl  save_bl
        0        0                               error
        0        1             0       1         loop
        1        0             1-      0         exit
        1        1             1-      1-        exit */

void em_end_lp()
{
if(em_debug == 12) printf("Entering em_end_lp() with noex_bl=%d, save_bl=%d, called=%d, linenum=%d\n",
		      noex_bl,save_bl,called,linenum);


/* if(noex_bl<=0)
   if(save_bl > 0)
     {quad_bl=loops_bl[tols_bl-1][in_sub];
      if(in_sub <=0)
		  source_bl = INPUT;
      else
		  source_bl = STACK;}
   else printf("WARNING - illegal combo noex save in em_end_lp()\n");*/

 if(noex_bl<=0)
   if(save_bl>0)
     {quad_bl=loops_bl[tols_bl-1][in_sub];
      source_bl = STACK;}
   else printf("WARNING - illegal combo noex save in em_end_lp()\n");
 else
   {noex_bl--;
    if(brch_bl<=0 && noex_bl<=0) em_pop_loop();
    if(save_bl>0) save_bl--;
    if(save_bl <= 0 && in_sub <= 0) source_bl = INPUT;
/*    if(save_bl <= saves[in_sub] && called == calleds[in_sub]
		 && noex_bl <=0 && save_bl>0)  */
    if(noex_bl <=0)
			{if(quad_bl >= toqs_bl)
			  {source_bl = INPUT;
/*			 printf("END_LP at %d in %s (in_sub=%d) setting source_bl to INPUT\n",
						linenum, &sub_file_stack[in_subs[in_sub]][0], in_sub);
	  printf("with noex_bl=%d, save_bl=%d, quad_bl=%d, noqs_bl=%d\n",
		            noex_bl,save_bl,quad_bl, toqs_bl);*/
			}
			 /*source_bl = INPUT;*/}
    if(save_bl<=0 && source_bl == INPUT) em_garbage(); };

if(em_debug == 12)
printf("Exiting em_end_lp() with noex_bl=%d, save_bl=%d, quad_bl=%d, saves[in_sub]=%d, called=%d, calleds[in_sub]=%d\n",
		      noex_bl,save_bl,quad_bl,saves[in_sub],called,calleds[in_sub]);
}


/* em_until() performs the following function if no responders or true.
                                                  ^
     noex_bl  save_bl       noex_bl  save_bl      |
        0        0                                modified 880309 jlp
        0        1             1+       1
        1        0
        1        1
*/
void em_until(mode1,mode2,arg1)
int mode1,mode2,arg1;
{
 if(em_debug == 2) printf("Entering em_until() at line number %d with noex_bl=%d, save_bl=%d\n",
		      linenum,noex_bl,save_bl);
  if(noex_bl<=0)
    { if(save_bl<=0) printf("WARNING - illegal combo noex save in em_until\n");
      if(mode2==T_PARALLEL)
			if(mode1==T_IFNANY)
			  {if(em_responders(arg1)==FALSE) noex_bl++;}
			else
			  {if(em_responders(arg1)==TRUE) noex_bl++;}
      if(mode2==T_SCALAR) if(sc_mem[arg1]) noex_bl++; };
}



/* em_while() performs the following function if no responders.
    while is parallel only.

     noex_bl  save_bl       noex_bl  save_bl
        0        0
        0        1             1+       1
        1        0
        1        1

em_while() returens TRUE when execute is on and there are no responders
(i.e. when time to exit the for loop )
It returns SUCCESS when execute is on and ther are responders.
It returns FAIL when no execute is on.
*/
em_while(arg1)
int arg1;
{ int status;

 if(em_debug == 2)printf("Entering em_while() with noex_bl=%d, save_bl=%d\n",
		      noex_bl,save_bl);
  status = FAIL;
  if(noex_bl<=0)
     if(em_responders(arg1)==FALSE) {noex_bl++; status = TRUE;} else status =SUCCESS;
  return(status);
}

em_stackwhile(arg1)
int arg1;
{ int status;

  if(em_debug == 2)printf("In em_stackwhile() with noex_bl_sk=%d, save_bl=%d\n",
		      noex_bl,save_bl);
  status = FAIL;
  if(noex_bl<=0)
    if(em_responders(arg1)==FALSE)
	   {noex_bl_sk++;
       noex_bl++;}
    else status =SUCCESS;
  if(em_debug == 2)printf("Exiting with  status=%d.\n",status);
  return(status);
}

