/* EM_COMMANDS.C 910415  */           

#include "stdarg.h"
#include "stdio.h"
#include "ctype.h"
#include "string.h"
#include "p2_symtab.h"
#include "em_common.h"
#include "p2_common.h"
#include "p2_header.h"
int quad_cnt;
/* ===========================================================================

   function name: em_commands
   input         : ---
   output           : ---
   return value : em_status     - SUCCESS, if no errors
						  - FAIL, otherwise

   global vars  : input_fp      - pointer to input file

   description   : This routine reads the integer representation of a command
			    from the input file and calls an appropriate routine to
			    handle the code generation for the command. If no errors
			    are detected before EOF is reached, SUCCESS is returned;
			    otherwise, the return value is FAIL.

=========================================================================== */

em_commands()
{
   int done;                   /* flag for loop termination */
   unsigned int command; /* integer value for command read from input file */
   unsigned int optype;                /* type of operation */
   unsigned int mode;    /* mode of oeration -- (mode + optype = command) */
   int em_status;              /* program error em_status -- SUCCESS or FAIL*/

   done = FALSE;
   em_status = SUCCESS;
   source_bl = INPUT;
   while (done == FALSE && em_status == SUCCESS)
     {if ((command=em_fscanf()) == EOF)
	     done = TRUE;
      else
	{mode = command % SIZE_MODE;
	 optype = (command / SIZE_MODE) * SIZE_MODE;
	 quad_cnt++;
//printf("in em_commonds, optype=%x, emptys=%d, fillds=%d\n",optype,emptys,fillds);

	 switch(optype)
		{case T_IDENT      : em_ident();
						    break;
	    case T_INCLUDE    : em_include();
				     break;
	    case T_CLRARRAY     : em_clrarray();
						    break;
	    case T_INITSTACK    : em_initstack();
						    break;
	    case T_LABEL        : em_label();
						    break;
	    case T_STOP            : em_stop();
						break;
	    case T_END             : em_end();
					       break;
	    case T_SETSCOPE     : em_setscope();
						break;
	    case T_STACK_SCOPE: em_stack_scope();
						break;
	    case T_POP_SCOPE  : em_pop_scope();
						break;
	    case T_ASSEMBLE     : em_assemble();
						break;
	    case T_DECLARE         : em_status = em_declare(mode);
						break;
	    case T_TEMP            : em_status = em_declare(mode);
						    break;
	    case T_SHORT      : em_status=em_short(mode);
						    break;
	    case T_ENTRY      : em_entry();
						    break;
	    case T_EXTERN     : em_extrn();
						    break;
	    case T_FIX        : em_status=em_fix(mode);
					       break;
	    case T_FLOAT      : em_status=em_float(mode);
						    break;
	    case T_INTCHR     : em_status=em_intchr(mode);
						    break;
	    case T_CHRINT     : em_status=em_chrint(mode);
						    break;
	    case T_ADD             :
		 case T_SUB                :
		 case T_MULT       :
		 case T_DIV                :
                 case T_MOD        :
	    case T_LT              :
		 case T_LE                 :
		 case T_EQ                 :
		 case T_NE                 :
		 case T_GT                 :
		 case T_GE                 :
	    case T_OR              :
		 case T_AND                :
		 case T_XOR                :
			   em_status = em_opcode(optype,mode);
						    break;
		 case T_NOT                : em_status = em_not(mode);
					   break;
	    case T_MINUS           :
	    case T_PLUS       : em_status = em_minusplus(optype, mode);
				break;
	    case T_ANY             :
	    case T_ELSENANY     :
		 case T_ENDANY     :
		 case T_FOR                :
		 case T_ENDFOR     :
		 case T_NEXT               :
		 case T_ENDNEXT    :
		 case T_GET                :
		 case T_ENDGET     :
		 case T_BEGWHILE   :
		 case T_WHILE      :
		 case T_ENDWHILE        :
		 case T_BEGSTACK   :
	    case T_STACKWHILE :
	    case T_ENDSTACKWHILE:
		 case T_RECURSE    :
	    case T_ENDRECURSE :
		 case T_LOOP       :
	    case T_UNTIL           :
	    case T_BEGIF      :
		 case T_ENDLOOP    :
		 case T_IF                 :
		 case T_ELSE               :
		 case T_ENDALLOCATE        :
		 case T_ENDIF      :em_control(optype, mode);
						   break;
	    case T_ANDIF      :
		 case T_ATHEN      :
		 case T_ENDANDIF   :
		 case T_ANDFOR     :
		 case T_ENDANDFOR  :
		 case T_ANDCONS    :
	    case T_ENDANDCONS : printf("ERROR - AND constructs not supported in -a option");
				em_status=FAIL;
				done=TRUE;
				break;
	    case T_ALLOCATE   : em_allocate(mode);
					     break;
		 case T_RELEASE    : em_release();
				     break;
		 case T_RESFST     : em_resfst();
				     break;
	    case T_MAIN       : em_main(mode);
						     break;
	    case T_SUBROUTINE : em_subroutine();
										  if(in_sub <= 0) done = TRUE;
						     break;
	    case T_ENDMAIN    : em_endmain();
				     done=TRUE;
						     break;
	    case T_ENDSUB     : em_endsub();
						     break;
	    case T_DEFLOG     : em_status=em_deflog();
						     break;
	    case T_DEFVAR     : em_status=em_defvar();
						     break;
	    case T_CALL       : em_call();
						     break;
	    case T_GETADD     : em_getadd();
						     break;
	    case T_DECLAR     : em_declar();
						     break;
	    case T_RETURN     : em_return();
						     break;
	    case T_MVSC       :
	    case T_MVPA       : em_status=em_move(optype,mode);
				     break;
	    case T_MVPAID     : em_mvpaid();
				     break;
	    case T_MVID       : em_mvid(mode);
						     break;
	    case T_MVCB       : em_mvcb();
						    break;
	    case T_MINDEX     : em_mindex();
					     break;
	    case T_MINVAL     : em_minval();
					     break;
	    case T_MAXDEX     : em_maxdex();
						     break;
	    case T_MAXVAL     : em_maxval();
						     break;
	    case T_NTHDEX     : em_nthdex();
						     break;
	    case T_NTHVAL     : em_nthval();
						     break;
		 case T_BOS        : em_bos();
						     break;
	    case T_PREFIX     :
		 case T_FSTCD      :
				case T_NXTCD      :
		 case T_PREVCD     :
		 case T_TRNCD      :
		 case T_TRNACD     : em_stcd(optype);
				     break;
	    case T_PREFIX8    :
		 case T_FSTCD8     :
		 case T_NXTCD8     :
		 case T_PREVCD8    :
		 case T_TRNCD8     :
		 case T_TRNACD8    : em_stcd8(optype);
				     break;
	    case T_MSG        : em_msg();
				     break;
            case T_SIBDEX     : em_sibdex(); break;
            case T_COUNT      : em_count();break;
            case T_SUM        : em_sum(mode); break;
            case T_ENCODE     : em_encode(mode); break;
            case T_DECODE     : em_decode(mode); break;
            case T_SQRT       : em_sqrt(mode); break;
            case T_EXP        : em_exp(mode); break;
            case T_POWER      : em_power(mode); break;
            case T_SIN        : em_sin(mode); break;
            case T_COS        : em_cos(mode); break;
	    case T_RAND       : em_rand(mode); break;
	    case T_TKNCMP     : em_tkncmp(mode); break;
            case T_HIT        : em_hit(mode); break;
/* IO ROUTINES */
/* print statment */
	    case T_OPEN       : em_status = em_open();  break;
	    case T_CLOSE      : em_close();  break;
	    case T_COPY       : em_copy(); break;
	    case T_CLEAR      : em_clear(mode);  break;
	    case T_FSTOUT     : em_fstout();  break;
	    case T_OUTPT      : em_outpt(mode);
						     break;
	    case T_COLAPSE    : em_colapse(mode);
						     break;
	    case T_BEGPRINT   : em_begprint(mode);
					     break;
	    case T_PRINT      : em_print(mode);
						     break;
	    case T_PRINT_DEL  : em_print_del(mode);
						     break;
	    case T_NL         : em_printnl();
					     break;
	    case T_ENDPRINT   : em_endprint();
				     break;
/* read statment */
		 case T_BEGREAD    : em_begread(mode);
				     break;
		 case T_READLN     : em_readln(mode);
				     break;
		 case T_READ       : em_status=em_read(mode);
				     break;
		 case T_ENDREAD    : em_endread(); break;
		 case T_IOALLOC    : em_ioalloc(mode);
				     break;
		 case T_EXPAND     : em_expand(mode);
				     break;
		 case T_INPUT      : em_input(mode);
				     break;
		 case T_STACK      : em_stack_gen(mode);
				     break;
		 case T_POP        : em_pop_gen(mode);
				     break;
		 case T_POPRC      : em_poprc(mode);
				     break;
		 case T_DUMP       : em_dump(mode);
				     break;
		 case T_DIMP       : em_dimp(mode);
				     break;
		 case T_LEX        : em_lex(mode);
				     break;
				case T_SCIN       : em_scin(1);
										  break;
				case T_SCINL      : em_scin(2);
										  break;
				case T_SCINP      : em_scin(3);
										  break;
				case T_SCIN8      : em_scin(5);
										  break;
				case T_SCOT       : em_scot(1);
										  break;
				case T_SCOTL      : em_scot(2);
										  break;
				case T_SCOTP      : em_scot(3);
										  break;
				case T_SCOT8      : em_scot(5);
										  break;
				case T_SHAPE      : em_shape();
										  break;
				case T_SCAREG     : em_scareg();
										  break;
				case T_REGSCA     : em_regsca();
										  break;
				case T_DASH       : em_dash();  break;
				case T_ADJ        : em_adj();   break;
				case T_NOUN       : em_noun();  break;
                case T_LOAD       : em_load(mode);  break;
                case T_STORE      : em_store(mode);  break;
	    default                   : printf("\nSYSTEM ERROR - line no %d quadrule no %d: Invalid",
					       linenum,quad_cnt);
				       printf(" intermediate code : %x  ", command);
				     printf("optyp=%x \n",optype);
				printf("  command_bl=%s\n",command_bl);
						    em_status = FAIL;
						    done = TRUE;
						    break;
	 }
      }
   }
   return(em_status);
}
