#ifndef EM_COMMON
#define EM_COMMON

#define NOOFP  320      /* number of processors in array memory */
#define SIZE_HOST 16    /* number of bits in host word */
#define SIZE_HOST32 32
#define SIZE_PACK 16    /* number of bits used in the host word for scalars */
#define SIZE_PACK32 32    /* number of bits used in the host word for scalars */
                         /* for 32 bit machines, not all SIZE_PACKs are affected */
#define SCALAR_SIZE  16   /* number of bits in the scalar portion of the machine*/
#define CHAR_SIZE     8   /* number of bits in an ascii character */
#define NOOFW NOOFP/SIZE_HOST /* number of words in C needed to emulate NOOFP*/
#define NOOFW32 NOOFP/SIZE_HOST32 /* number of words in C needed to emulate NOOFP on a 32 bit machine*/
#define INT_SIZE  32      /* the size of a scalar integer */
#define REAL_SIZE 32      /* the size of a real number */
#define L_STACK_SIZE   32  /* the size of the loop stack */
#define NOOFSTK         3  /* number of quadruple stacks */
#define Q_STACK_SIZE  3750  /* the size of the quadruple stack */
#define STACK  0          /* input is from the quadruple stack */
#define INPUT  1          /* input is from the source input file */
#define BOS    100           /* the beginning of the array stack */
#define TOP_SC_MEM  2560  /* non i/o scalar memory */
#define IO_BUF_SIZE 6144  /* area resereved for IO commands */
#define SC_MEM_SIZE TOP_SC_MEM+IO_BUF_SIZE /* total sc_mem size */
/* layout of scalar memory
	    ----
      STMP$|1
	   |:
	   |64
	    ----
     OUTBF$|65
	   |:
	   |6208
	    ----
     STACK$|6209
	   |:
	   |7232
	    ----
     C10D$ |7233
     C8D$  |7234
     C48D$ |7235
     C45D$ |7236
     C32D$ |7237
	    ----
	   |7238
	   |user declared scalars
	   |:
	   |tosm_bl
	    ----
	   |
	    ----
	   |sc_mem_size
	   |:
	   |scalar stack
	   |:
SC_MEM_SIZE|
	    ----
*/
/*
<---------PA_MEM_SIZE(4096)------------------------------------------------>
M| TEMP ... TEMP5|       |IOSTACK |I|T
R| <32>     <32 >|       |        |O|O--->STACK  SSSFTL LL LR RI I
E| <---160------>|       | 96 bits|M|M           YYYARI IP PP PP P
G|                       |(IOSTACK|A|            SSSLUT TT TT TT T
<-------RESERVE_SPACE--->| _SIZE) |S|            321SEM MM MM MM M
	(256 bits)       |        |K|            SSSESP PP PP PP P
						    S 4 01 04 01 0
						      S S5 SS S5 S
							 S     S

44 44 44 33 33 33       3 3      3 3 3          11 1111 66 66 55 0
00 00 00 99 99 99       8 8      7 7 7          11 0000 88 77 11
99 66 33 99 66 33       4 3      4 4 4          00 1000 87 21 21
54 32 10 98 76 54       09       4 3 2          65 0987
*/

/*#define PA_MEM_SIZE 4096*/  /* the size in bits of the parallel array memory */
#define PA_MEM_SIZE 5000  /* the size in bits of the parallel array memory */
#define MREG PA_MEM_SIZE-1 /* the address of the M register */
#define TEMP PA_MEM_SIZE-INT_SIZE-1
#define TEMP2 PA_MEM_SIZE-2*INT_SIZE-1
#define TEMP3 PA_MEM_SIZE-3*INT_SIZE-1
#define TEMP4 PA_MEM_SIZE-4*INT_SIZE-1
#define TEMP5 PA_MEM_SIZE-5*INT_SIZE-1
#define RESERVE_SPACE 256 /* the number of bits in the parallel array reserved
			   for the compiler and operating system */
#define IOSTACK_SIZE  96 /* the width of the IOSTACK */
#define IOSTACK PA_MEM_SIZE-RESERVE_SPACE-IOSTACK_SIZE /* left most bit of stack*/
#define IOMASK IOSTACK-1 /* the address of the iostack mask */
#define GLOBAL_BI IOSTACK-2 /* universal memory space */

#define TOM IOMASK-2 /* the top of (useable) memory */

#define SC_TEMP_SIZE 128   /* the size of SC_TEMP, coordinate with p1_parse.h */
/* em_print_err #defines  */
#define FLOAT  1
#define NOT_IMPLEMENTED 2
#define SIZE 3
#define LOGICAL 4

 int code_type; /* STARAN, ASPROL, EMULATE, or CM */
 int table_parse;
 int loops_bl[L_STACK_SIZE/NOOFSTK][NOOFSTK], tols_bl;
typedef struct {char  *qpnt[Q_STACK_SIZE/NOOFSTK];} qspnt;
 qspnt *quads_stk[NOOFSTK];
 int toqs_bl, quad_bl, quad_dx;
 int quads_dx[NOOFSTK], quads_bl[NOOFSTK], quads_sc[NOOFSTK];
 int toqss_bl[NOOFSTK], tolss_bl[NOOFSTK];
 int recurse_stack;
 int brch_bl;
/* char buffer_bl[(MAX_ID_SIZE+1)*MAX_NO_OUT+50]; defined in p2_common.h */
 char buffer_bl[(6+1)*24+50];
 int source_bl, save_bl, noex_bl, saves[NOOFSTK];
 int  sc_mem[SC_MEM_SIZE];
 int  pa_mem[NOOFW][PA_MEM_SIZE];
 int command_bl, fillds, emptys;
 int tosm_bl;    /* top of scalar memory - incremented as memory is allocated */
 int sc_mem_size; /* adjusted as scalar stack grows down */
 int em_debug_1,em_debug_2,em_debug_3,em_debug_4,em_debug_5;
 int em_debug;
/* this outbuf is for debugging dumps and is NOT OUTBF$ used for io */
 int  outbf[24][256]; /* added 880305 jlp */

/* symbol table definitions */

/* to allow STACK emulation - only one statement */
#define GBL_STACKSIZE 20
#define STACK_STOP 0
 int gbl_bl;
 int global[GBL_STACKSIZE];
 int recurse_noexbl,recurse_savebl,recurse_brchbl;
 int noex_bl_sk;   /* same as noex_bl but for STACKs only */
 int noex_bl_ext;   /*use to flag noex external to recurse struct */


/* for subroutine emulation */
#include "stdio.h"
#define SUB_DEPTH 20
 FILE *sub_stack[SUB_DEPTH];
 int in_sub, stack_cnt, in_subs[SUB_DEPTH];
 char sub_file[20];
 char  sub_file_stack[SUB_DEPTH][20];
 char code_bl[16];
 int sub_stk[SUB_DEPTH], sub_toqs[SUB_DEPTH], sav_toqs[SUB_DEPTH];
 int sub_src[SUB_DEPTH];
 int called, calleds[NOOFSTK];

#define MAX_NO_LEN 256  /* should be in common */
 int token_length;
 char ctoken[MAX_NO_LEN];
 char *pnt;

 char data[MAX_NO_LEN],*datapnt;
 char strbuf[100],*strpnt; /* output string buffer */

#define FILETABSIZE 20
#define PATHSIZE 200
#define NOOFSY     3   /* number of channels reserved for system */
                       /* used to translate user file numbers to zero index for filetab */ 
typedef struct {
    FILE  *streamot;
    char  filein[PATHSIZE];
    char  fileot[PATHSIZE];
    int  didout;
    int  didin;
    } filet;  

 filet  filetab[FILETABSIZE];

#endif
