/* EM_RTNS_ARITH.C 910415 */

#include "stdarg.h"
#include "stdio.h"
#include "ctype.h"
#include "string.h"
#include "p2_symtab.h"
#include "em_common.h"
#include "p2_common.h"
#include "p2_header.h"

/* the scalar routines, fxscad, fxscsb, fxscmin assume that the
   long fields are stored as follows:
    msb             lsb
	 arg+n...arg+1,arg+0,   where n = size/HOST_SIZE   */


/* fixed parallel add - masked, signed */
void fxpaad(int arg1, int arg2, int arg3, int size)
{

  em_stack(arg1);  /* save sign bit of arguments */
  em_stack(arg2);
  fxpaadun(arg1,arg2,arg3,size);
  lgpaxor(emptys+2,emptys+1,emptys,1);
  lgpanot(emptys,emptys,1);
  lgpaand(MREG,emptys,emptys,1);
  papamvmk(emptys+2,arg3,emptys,1);
  inc_pa_perform(-8);
  em_pop(arg2);
  em_pop(arg1);
}

/* fixed parallel add or subtract - masked, signed */
void fxpaadsb(int opcode, int arg1, int arg2, int arg3, int size)
{
int argt;

  argt = arg2;
  if(opcode == T_SUB)
    {argt = get_add("TEMP2");
     fxpamin(arg2, argt, size);
     inc_pa_perform(-size*3);}
  em_stack(arg1);  /* save sign bit of arguments */
  em_stack(argt);
  fxpaadun(arg1,argt,arg3,size);
  lgpaxor(emptys+2,emptys+1,emptys,1);
  lgpanot(emptys,emptys,1);
  lgpaand(MREG,emptys,emptys,1);
  papamvmk(emptys+2,arg3,emptys,1);
  inc_pa_perform(-8);
  em_pop(argt);
  em_pop(arg1);
}

/* extended fixed parallel add - masked, signed */
fxpaadex(arg1,arg2,arg3,size)
int size, arg1, arg2, arg3;
{

  em_stack(arg1);  /* save sign bit of arguments */
  em_stack(arg2);
  fxpaadun(arg1,arg2,arg3,size); /* make room for extended bit */
  lgpaxor(emptys+2,emptys+1,emptys,1);
  lgpanot(emptys,emptys,1);
  lgpaand(MREG,emptys,emptys,1);
  papamvmk(emptys+2,arg3,emptys,1);
  inc_pa_perform(-8);
  em_pop(arg2);
  em_pop(arg1);
}

/* Fixed parallel add - masked, unsigned */
void fxpaadun(int arg1, int arg2, int arg3, int size)
{
      int i,p,x,y,carry[NOOFW],yxorc;

      for(p=0; p<NOOFW; p++) carry[p] = 0;
      for(i=size-1; i>=0; i--)
	for(p=0; p<NOOFW; p++)
	  { x = pa_mem[p][arg1+i];
	    y = pa_mem[p][arg2+i];
	    yxorc = y^carry[p];
	    pa_mem[p][arg3+i] = ((~x&yxorc|x&~yxorc) & pa_mem[p][MREG])
				   | (pa_mem[p][arg3+i] & ~pa_mem[p][MREG]);
	    carry[p] = carry[p]&(x|y)|x&y;
	  };

}

/* fixed common right add - masked */
/* fxcrad(pa,sc,pa,size) */
void fxcrad(int arg1, int arg2, int arg3, int size)
{
      int argt;

      argt = em_stack_temp(size);
      scpamvun(arg2,argt,size);
      fxpaad(arg1,argt,arg3,size);
      inc_pa_perform(-size*2);
      em_pop_temp(size);
}

#define HMASK (1<<SIZE_HOST)-1   /*mask corresponding to SIZE_HOST */

/* fixed scalar add or subtrace */
void fxscadsb(int opcode, int arg1, int arg2, int arg3, int size)
{int i, carry, del;
 long int temp;

		carry = 0;
      for(i=0; i<=(size-1)/SIZE_HOST; i++)
			 {temp  = sc_mem[arg1+i];
			  if(opcode == T_ADD)
			    {temp += sc_mem[arg2+i]; del = 2; }
			  else temp -= sc_mem[arg2+i];
			    {temp += carry; del = 3; }
			  sc_mem[arg3+i] = (unsigned int) temp&HMASK;
			  carry = (temp>>SIZE_HOST)&HMASK;
			  }
      inc_sc_perform(del*((size-1)/SIZE_HOST+1));
      return;
}

/* fixed scalar add */
fxscad(arg1, arg2, arg3, size)
int size, arg1, arg2, arg3;
{int i, carry;
 long int temp;

		carry = 0;
      for(i=0; i<=(size-1)/SIZE_HOST; i++)
			 {temp  = sc_mem[arg1+i];
			  temp += sc_mem[arg2+i];
			  temp += carry;
			  sc_mem[arg3+i] = temp&HMASK;
			  carry = (temp>>SIZE_HOST)&HMASK;
			  }
      inc_sc_perform(2*((size-1)/SIZE_HOST+1));
}


/* long int scalar add or subtract*/
void liscadsb(int opcode, int arg1, int arg2, int arg3)
{ long lw1,lw2,lw3;

 if(get_size(carg1) <= 16)
	lw1 = 0;
 else
   {lw1 = (unsigned) sc_mem[arg1+1];
    lw1 <<= 16;}
 lw1 += (unsigned) sc_mem[arg1];
 if(get_size(carg2) <= 16)
	lw2 = 0;
 else
   {lw2 = (unsigned) sc_mem[arg2+1];
    lw2 <<= 16;}
 lw2 += (unsigned) sc_mem[arg2];
 if(opcode == T_ADD) lw3 = lw1 + lw2;
 else if (opcode == T_SUB) lw3 = lw1 - lw2;
 sc_mem[arg3+1] = lw3>>16;
 lw3<<=16;
 sc_mem[arg3] = lw3>>16;
 return;
}


/* SUBTRACT ROUTINES */

/* fixed parallel subtract - masked */
/* arg3 = arg1-arg2 */
fxpasb(arg1, arg2, arg3, size)
int size, arg1, arg2, arg3;
{
  int argt;

  argt = get_add("TEMP2");
  fxpamin(arg2, argt, size);
  inc_pa_perform(-size*3);
  fxpaad(arg1, argt, arg3, size);
}

/* fixed common left subtract - masked */
/* fxclsb(sc,pa,pa,size) */
/* arg3 = arg1 - arg2 */
fxclsb(arg1, arg2, arg3, size)
int size, arg1, arg2, arg3;
{
  int argt;

  argt = get_add("TEMP2");
  fxpamin(arg2, argt, size);
  inc_pa_perform(-size*3);
  fxcrad(argt,arg1,arg3,size);
}

/* fixed common right subtract - masked */
/* fxcrsb(pa,sc,pa,size) */
/* arg3 = arg1-arg2 */
void fxcrsb(int arg1, int arg2, int arg3, int size)
{
  int argt;

  if((argt = get_add("SC_TEMP")) == FAIL)
    printf("ERROR - Can't find SC_TEMP.\n");
  fxscmin(arg2, argt, size);
  inc_pa_perform(-size*3);
  fxcrad(arg1,argt,arg3,size);
}

/* fixed scalar subtract */
/* arg3 = arg1-arg2 */
fxscsb(arg1, arg2, arg3, size)
int size, arg1, arg2, arg3;
{int i,carry;
 long int temp;

		carry = 0;
      for(i=0; i<=(size-1)/SIZE_HOST; i++)
			 {temp  = sc_mem[arg1+i];
			  temp -= sc_mem[arg2+i];
			  temp += carry;
			  sc_mem[arg3+i] = (unsigned int) temp&HMASK;
			  carry = (temp>>SIZE_HOST)&HMASK;
	   }
      inc_sc_perform(3*((size-1)/SIZE_HOST+1));
}

/* long int scalar sub */
liscsb(arg1, arg2, arg3)
int arg1, arg2, arg3;
{ long lw1,lw2,lw3;

 if(get_size(carg1) <= 16)
	lw1 = 0;
 else
   {lw1 = (unsigned) sc_mem[arg1+1];
    lw1 <<= 16;}
 lw1 += (unsigned) sc_mem[arg1];
 if(get_size(carg2) <= 16)
	lw2 = 0;
 else
   {lw2 = (unsigned) sc_mem[arg2+1];
    lw2 <<= 16;}
 lw2 += (unsigned) sc_mem[arg2];
 lw3 = lw1 - lw2;
 sc_mem[arg3+1] = lw3>>16;
 lw3 = lw3<<16;
 sc_mem[arg3] = lw3>>16;
}

/* MULTIPLY ROUTINES */

/* fixed parallel multiply - masked */
/* size of argt must be twice the size of arg1 and arg2 */
void fxpaml(int arg1, int arg2, int arg3, int size)
{
      int i,shift,argt;

  argt = arg3;
  paclrun(argt,size*2);
  em_stack(MREG);
  /* complement negative values  and save flag */
  em_stack(arg1);
  papamvun(arg1,MREG,1);
  fxpamin(arg1,arg1,size);
  em_stack(arg2);
  papamvun(arg2,MREG,1);
  fxpamin(arg2,arg2,size);
      shift = 0;
      for(i=size-1; i>=0; i--)
	{lgpaand(fillds+2,arg2+i,MREG,1);
	 if(em_responders(MREG))
		fxpaadun(arg1,argt+size-shift,argt+size-shift,size);
	      shift++;
	     }
  /* generate sign of product */
  lgpaxor(fillds,fillds+1,MREG,1);
  paclrun(emptys,1);
  for(i=0; i<size*2; i++) lgpaor(argt+i,emptys,emptys,1);
  lgpaand(emptys,MREG,MREG,1); /* mask out zero products */
  fxpamin(argt,argt,size*2);
  /* restore arg1 and arg2 */
  em_pop(MREG);
  fxpamin(arg2,arg2,size);
  em_pop(MREG);
  fxpamin(arg1,arg1,size);
  em_pop(MREG);  /* restore mask */
}

/* fixed common right  multiply - masked */
/* fxcrml(pa,sc,pa) */
void fxcrml(int arg1, int arg2, int arg3, int size)
{
      int argt;

      argt = get_add("TEMP2");
      scpamvun(arg2,argt,size);
      inc_pa_perform(-size);
      fxpaml(arg1,argt,arg3,size);
}

/* fixed scalar multiply */
void fxscml(int arg1, int arg2, int arg3, int size)
{
      if(size > SIZE_HOST)
			printf("ERROR - multiplication of scalars length over %d bits not supported.\n",
					  SIZE_HOST);
      sc_mem[arg3] = sc_mem[arg1]*sc_mem[arg2];
      inc_sc_perform(3);
}

/* long int scalar mul */
void liscml(int arg1, int arg2, int arg3)
{ long lw1,lw2,lw3;

 if(get_size(carg1) <= 16)
	lw1 = 0;
 else
   {lw1 = (unsigned) sc_mem[arg1+1];
    lw1 <<= 16;}
 lw1 += (unsigned) sc_mem[arg1];
 if(get_size(carg2) <= 16)
	lw2 = 0;
 else
   {lw2 = (unsigned) sc_mem[arg2+1];
    lw2 <<= 16;}
 lw2 += (unsigned) sc_mem[arg2];
 lw3 = lw1 * lw2;
 sc_mem[arg3+1] = lw3>>16;
 lw3 = lw3<<16;
 sc_mem[arg3] = lw3>>16;
}

/* DIVISION ROUTINES */

/* fixed parallel divide  */
/* arg1/arg2 -> arg3      */
/* convrtpapa() in em_rtns_io.c assumes fxpadv uses LITMP$+0 and leaves
   the remainder in its low order bits.        */
void fxpadv(int arg1, int arg2, int arg3, int size)
{ int shift,argl,argt2,args,sizel,one,i;

  /* set up long dividend field */
  argl = get_add("LITMP$");
  sizel = size*2;
  one = get_add("TRUE$");
  argt2 = get_add("TEMP2");
  em_stack(MREG); /* save mask */
  /* complement negative values  and save flag */
  em_stack(arg1);
  papamvun(arg1,MREG,1);
  fxpamin(arg1,arg1,size);
  em_stack(arg2);
  papamvun(arg2,MREG,1);
  fxpamin(arg2,arg2,size);
  paclrun(argl,sizel);
  papamvun(arg1,argl+size,size);
  /* set up minus field */
  papamvun(fillds+2,MREG,1);
  fxpamin(arg2,argt2,size);
  papamvun(fillds+2,MREG,1); /* restore mask */
  paclrmk(arg3,MREG,size);   /* clear result field */
  for(i=0; i<size; i++)
    {lgpaand(MREG,argl+size+i,emptys,1);
     if(em_responders(emptys)) break; }
  /* flag nonzero quotients in emty$ */
  paclrun(emptys,1);
  for(shift=i; shift<=size; shift++)
       {args = argl+shift;
	    papamvun(fillds+2,MREG,1); /* restore mask */
	    fxpaadun(args,argt2,args,size); /* subtract arg2 from arg1 */
	    papamvun(args,MREG,1); /* if neg, add back in */
	    fxpaadun(args,arg2,args,size);
	    lgpanot(MREG,MREG,1); /* is positive, set quotient bit */
	    papamvmk(one,arg3+shift-1,MREG,1);
	    lgpaor(MREG,emptys,emptys,1);
       }
  /* generate sign of quotient for non zero quotient */
  lgpaxor(fillds,fillds+1,MREG,1);
  lgpaand(emptys,MREG,MREG,1); /* mask out zero quotients */
  fxpamin(arg3,arg3,size);
  /* restore arg1 and arg2 */
  em_pop(MREG);
  fxpamin(arg2,arg2,size);
  em_pop(MREG);
  fxpamin(arg1,arg1,size);
  em_pop(MREG);  /* restore mask */
}


/* fixed common right divide */
void fxcrdv(int arg1, int arg2, int arg3, int size)
/* fxcrdv(pa,sc,pa,size) */
{ int argt;

  argt = get_add("TEMP3");
  scpamvun(arg2,argt,size);
  inc_pa_perform(-size*2);
  fxpadv(arg1,argt,arg3,size);
}

/* fixed common left divide */
fxcldv(arg1,arg2,arg3,size)
/* fxcldv(sc,pa,pa,size) */
int size, arg1, arg2, arg3;
{ int argt;

  argt = get_add("TEMP3");
  scpamvun(arg1,argt,size);
  inc_pa_perform(-size*2);
  fxpadv(argt,arg2,arg3,size);
}

/* fixed scalar division */
void fxscdv(int arg1, int arg2, int arg3, int size)
{
  if(size > SIZE_HOST)
	 printf("ERROR - division of scalars over %d bits not supported.\n",
			   SIZE_HOST);
  sc_mem[arg3] = sc_mem[arg1]/sc_mem[arg2];
  inc_sc_perform(3);
}


/* long int scalar div */
void liscdv(int arg1, int arg2, int arg3)
{ long lw1,lw2,lw3;

 if(get_size(carg1) <= 16)
	lw1 = 0;
 else
   {lw1 = (unsigned) sc_mem[arg1+1];
    lw1 <<= 16;}
 lw1 += (unsigned) sc_mem[arg1];
 if(get_size(carg2) <= 16)
	lw2 = 0;
 else
   {lw2 = (unsigned) sc_mem[arg2+1];
    lw2 <<= 16;}
 lw2 += (unsigned) sc_mem[arg2];
 lw3 = lw1 / lw2;
 sc_mem[arg3+1] = lw3>>16;
 lw3 = lw3<<16;
 sc_mem[arg3] = lw3>>16;
}

/* long int scalar modulo */
void liscmd(int arg1, int arg2, int arg3)
{ long lw1,lw2,lw3;

 if(get_size(carg1) <= 16)
	lw1 = 0;
 else
   {lw1 = (unsigned) sc_mem[arg1+1];
    lw1 <<= 16;}
 lw1 += (unsigned) sc_mem[arg1];
 if(get_size(carg2) <= 16)
	lw2 = 0;
 else
   {lw2 = (unsigned) sc_mem[arg2+1];
    lw2 <<= 16;}
 lw2 += (unsigned) sc_mem[arg2];
 lw3 = lw1 % lw2;
 sc_mem[arg3+1] = lw3>>16;
 lw3 = lw3<<16;
 sc_mem[arg3] = lw3>>16;
}

/* ARITHMETIC UNARY ROUTINES */
/* fixed  parallel minus - masked */
void fxpamin(int arg1, int arg2, int size)
{
    int p,i,sargt,pargt2;

   for (i=0; i<size; i++)
    {for (p=0; p<NOOFW; p++)
		 pa_mem[p][arg2+i] =
	   (~pa_mem[p][arg1+i]&pa_mem[p][MREG]) |
			  (pa_mem[p][arg1+i]&~pa_mem[p][MREG]);
     }
   inc_pa_perform(size*3);
   sargt = get_add("SC_TEMP"); /* add 1 for two's compliment */
   sc_mem[sargt] = 1;          /* SC_TEMP is at sc_mem[0] */
   pargt2 = em_stack_temp(size);
   scpamvmk(sargt,pargt2+size-SIZE_HOST,get_add("TRUE$"),SIZE_HOST,SIZE_HOST);
   pasnexmk(pargt2+size-SIZE_HOST,pargt2,get_add("TRUE$"),size-SIZE_HOST);
   fxpaad(arg2,pargt2,arg2,size);
   inc_pa_perform(-size*5);
   em_pop_temp(size);

   /* get rid of -0's */
   paclrun(emptys,1);
   for(i=1; i<size; i++) lgpaor(arg2+i,emptys,emptys,1);
   lgpanot(emptys,emptys,1);
   paclrmk(arg2,emptys,1);
}

/* fixed scalar minus */
void fxscmin(int arg1, int arg2, int size)
{int i,j,k;
  
  k=(size-1)/SIZE_HOST;
  for(i=0; (i<=k) & (sc_mem[arg1+i]==0); i++)
     sc_mem[arg2+i] = 0;
  if(i<=k)
     sc_mem[arg2+i] = (-sc_mem[arg1+i]) & 0xFFFF;
  for(j=i+1; j<=k; j++)
	  sc_mem[arg2+j] = (~sc_mem[arg1+j]) & 0xFFFF;

  inc_sc_perform(2*(k+1));
  return;
}
