/* EM_RTNS_LOG.C 910415 */

#include "stdarg.h"
#include "stdio.h"
#include "ctype.h"
#include "string.h"

#include "p2_symtab.h"
#include "em_common.h"
#include "p2_common.h"
#include "p2_header.h"

/* LOGICAL FUNCTIONS */

/* the scalar routines, lgscor, lgscand, lgscnot, lgscxor assume that the
   long fields are stored as follows:
    msb             lsb
	 arg+n...arg+1,arg+0,   where n = size/HOST_SIZE   */

/* logical parallel or */
void lgpaor(int arg1, int arg2, int arg3, int size)
{
    int p,i;

   for (i=0; i<size; i++)
     for (p=0; p<NOOFW; p++)
        pa_mem[p][arg3+i] = pa_mem[p][arg1+i]| pa_mem[p][arg2+i];
   inc_pa_perform(size*4);
  }


/* logical common right or */
/* lgcror(pa,sc,pa,size) */
void lgcror(int arg1, int arg2, int arg3, int size)
{
    int argt;

    argt = get_add("TEMP");
    scpamvun(arg2,argt,size);
    inc_pa_perform(-size*2);
    lgpaor(arg1,argt,arg3,size);
    inc_pa_perform(-size*2);
 }

/* logical scalar or */
void lgscor(int arg1, int arg2, int arg3, int size)
{int i;

    for(i=0; i<=(size-1)/SIZE_HOST; i++)
      sc_mem[arg3+i] = sc_mem[arg1+i] || sc_mem[arg2+i];
    inc_sc_perform(3);
  }

/* logical parallel and */
void lgpaand(int arg1, int arg2, int arg3, int size)
{
    int p,i;

   for (i=0; i<size; i++)
     for (p=0; p<NOOFW; p++)
        pa_mem[p][arg3+i] = pa_mem[p][arg1+i] & pa_mem[p][arg2+i];
}


/* logical common right and */
/* lgcrand(pa,sc,pa,size) */
void lgcrand(int arg1, int arg2, int arg3, int size)
{
    int argt;

    argt = get_add("TEMP");
    scpamvun(arg2,argt,size);
    inc_pa_perform(-size*2);
    lgpaand(arg1,argt,arg3,size);
    inc_pa_perform(-size);
 }

/* logical scalar and */
void lgscand(int arg1, int arg2, int arg3, int size)
{int i;

  for(i=0; i<=(size-1)/SIZE_HOST; i++)
    sc_mem[arg3+i] = sc_mem[arg1+i] && sc_mem[arg2+i];
  inc_sc_perform(3);
  }

/* logical parallel xor */
void lgpaxor(int arg1, int arg2, int arg3, int size)
{
    int p,i;

   for (i=0; i<size; i++)
     for (p=0; p<NOOFW; p++)
        pa_mem[p][arg3+i] = pa_mem[p][arg1+i] ^ pa_mem[p][arg2+i];
   inc_pa_perform(size*4);
}


/* logical common right xor */
/* lgcrxor(pa,sc,pa,size) */
void lgcrxor(int arg1, int arg2, int arg3, int size)
{
    int argt;

    argt = get_add("TEMP");
    scpamvun(arg2,argt,size);
    inc_pa_perform(-size*2);
    lgpaxor(arg1,argt,arg3,size);
    inc_pa_perform(-size);
}

/* logical scalar xor */
void lgscxor(int arg1, int arg2, int arg3, int size)
{int i;

  for(i=0; i<=(size-1)/SIZE_HOST; i++)
    sc_mem[arg3+i] = (sc_mem[arg1+i] & !sc_mem[arg2+i]) |
                    (!sc_mem[arg1+i] &  sc_mem[arg2+i]);
    inc_sc_perform(3);
}

/* logical parallel not */
void lgpanot(int arg1, int arg2, int size)
{
    int p,i;

   for (i=0; i<size; i++)
     for (p=0; p<NOOFW; p++)
       pa_mem[p][arg2+i] = ~pa_mem[p][arg1+i];

   inc_pa_perform(size*2);
}

/* logical scalar not */
void lgscnot(int arg1, int arg2, int size)
{int i;

  for(i=0; i<=(size-1)/SIZE_HOST; i++)
    sc_mem[arg2+i] = !sc_mem[arg1+i];
  inc_sc_perform(2);
}

