/* EM_RTNS_MOV.C 910415 */

#include "stdarg.h"
#include "stdio.h"
#include "ctype.h"
#include "string.h"
#include "p2_symtab.h"
#include "em_common.h"
#include "p2_common.h"
#include "p2_header.h"

/* MOVE ROUTINES */

/* SCALARS ARE ALWAYS RIGHT JUSTIFIED */

/* scalar to scalar move */
void scscmv(int arg1, int arg2, int size)
{int i,j;

  j=(size-1)/SIZE_HOST;
  sc_mem[arg2+j] = (sc_mem[arg1+j]&(~0>>(SIZE_HOST-size%SIZE_HOST))) & 0xFFFF;
  if(j >= 1)
    for(i=j-1; i>=0; i--)
 	   sc_mem[arg2+i] = sc_mem[arg1+i] & 0xFFFF;
}

/* scalar to parallel move, masked */
/* smaller scalar field is right justified in larger parallel field
   with 0 fill */
/* scpamvmk(sc,pa,mask,size1,size2) */
void scpamvmk(int arg1, int arg2, int arg3, int size1, int size2)
{
  int i, j, size;
  unsigned int word;


  j = (size2-1)/SIZE_HOST;
  size = size2-j*SIZE_HOST;
  word = 0;
  if(j*SIZE_HOST < size1) word = (unsigned int) sc_mem[arg1+j];
  wordmvmk(word,arg2,arg3,size,size);
  for (i =j-1; i>=0; i--)
   {word = 0;
    if(i*SIZE_HOST < size1) word = (unsigned int) sc_mem[arg1+i];
    wordmvmk(word,arg2+size2-SIZE_HOST*(i+1),
				  arg3,SIZE_HOST,SIZE_HOST);}
}

void wordmvmk(unsigned int word, int arg2, int arg3, int size1, int size2)
{
  int mask, p, i;

/*printf("In wordmvmk, word=%d, arg2=%d, arg3=%d, size1=%d, size2=%d\n",
        word,arg2,arg3,size1,size2);
*/
  mask=1;
  if(size1 == 0) return;
  for (i = size2-1; i>=0; i--)
    {if(word & mask)
       for(p=0; p<NOOFW; p++)
           pa_mem[p][arg2+i] =
	       pa_mem[p][arg2+i]|((~pa_mem[p][arg2+i]) & pa_mem[p][arg3]);
     else
	   for(p=0; p<NOOFW; p++)
	     pa_mem[p][arg2+i] =
	       pa_mem[p][arg2+i] & ~pa_mem[p][arg3];
     word >>= 1;}

  inc_sc_perform(1);
  inc_pa_perform(2*size2);
}


/* scalar to parallel move, unmasked */
/* scpamvun(sc,pa,size) */
/* scalar is right justified in word */
void scpamvun(int arg1, int arg2, int size)
{
  int i, j;

//printf("In scpamvun\n");
  j = (size-1)/SIZE_HOST;
  wordmvun((unsigned int)sc_mem[arg1+j],arg2,size-j*SIZE_HOST);
//printf("sc1 fillds=%d\n",fillds);
  for (i =j-1; i>=0; i--)
  { wordmvun((unsigned int)sc_mem[arg1+i],arg2+size-SIZE_HOST*(i+1),SIZE_HOST);
//printf("sc2 fillds=%d\n",fillds);
  }
  }

/* scalar to parallel move, unmasked */
/* scpamvun(sc,pa,size) */
/* scalar is right justified in word */
/* 16 bits per word even in 32 bit machines */
void wordmvun(unsigned int word, int arg2, int size)
{
  int mask,i, p, bit;

//printf("In wordmvun, word=%x, arg2=%x(hex), size=%d\n",word,arg2,size);

  mask=1;
  for (i = size-1; i>=0; i--)
    {bit = (word&mask) ? -1 : 0 ;
//  printf("mv1 fillds=%d\n",fillds);
  for(p=0; p<NOOFW; p++) pa_mem[p][arg2+i] = bit;
//  printf("mv2 fillds=%d\n",fillds);
  word >>= 1;
    }
  inc_pa_perform(2*size);
}

void em_mvpa(int arg1, int arg2, int arg3, int size2, int size3)
/*em_mvpa(mask,pa,pa)
   em_mvpa is a generalized parallel move routine, it
   will move smaller arg2 field into larger arg3 with sign extend.  It
   will move larger arg2 field into smaller arg3 by modulo and copy sign */
{ int nsize,narg3,narg2;

   if(size3 >= size2)
     {nsize = size3-size2;
      narg3 = arg3 + nsize;
      papamvmk(arg2,narg3,arg1,size2);
      pasnexmk(arg2,arg3,arg1,nsize);
     }
   else
     {nsize = size2-size3;
      narg2 = arg2 + nsize;
      papamvmk(narg2,arg3,arg1,size3);
      if (size3 > 1) papamvmk(arg2,arg3,arg1,1);
     }
}

/* parallel to parallel move - masked */
/* papamvmk(pa,pa,mask,size) */
void papamvmk(int arg1, int arg2, int arg3, int size)
{
  int i, p;
//printf("in papamvmk, from %d to %d for %d bits\n",arg1,arg2,size);
  for (i = size-1; i>=0; i--)
    for(p=0; p<NOOFW; p++)
	   pa_mem[p][arg2+i] = (pa_mem[p][arg1+i]&pa_mem[p][arg3])|
                          (pa_mem[p][arg2+i]&~pa_mem[p][arg3]);
  inc_pa_perform(2*size);
}

/* parallel to parallel move - masked, high to low */
/* papamvmkhl(pa,pa,mask,size) */
void papamvmkhl(int arg1, int arg2, int arg3, int size)
{
  int i, p;

  for (i = 0; i<size; i++)
    for(p=0; p<NOOFW; p++)
	   pa_mem[p][arg2+i] = (pa_mem[p][arg1+i]&pa_mem[p][arg3])|
                          (pa_mem[p][arg2+i]&~pa_mem[p][arg3]);
  inc_pa_perform(2*size);
}

/* parallel to parallel move - unmasked */
void papamvun(int arg1, int arg2, int size)
{
  int i, p;

  for (i = size-1; i>=0; i--)
    for(p=0; p<NOOFW; p++)
       pa_mem[p][arg2+i] =  pa_mem[p][arg1+i];
  inc_pa_perform(2*size);
}


/* parallel to scalar (indexed) move */
/* pascmv(pa,sc,index,size) */
/* index is saved */
void pascmv(int arg1, int arg2, int arg3, int size)
{
   int i,index,argt,j,rsize;

if(em_debug == 8) printf("pascmv, line=%d, size=%d.  arg3(mask)=\n", linenum,size);
  argt = get_add("TEMP");
  papamvun(arg3,argt,1);
  index = first(arg3);
  j = (size-1)/SIZE_HOST;
  rsize = size-j*SIZE_HOST;
  wdpascmv(arg1,arg2+j,arg3,index,rsize);
  for (i =j-1; i>=0; i--)
    {wdpascmv(arg1+size-SIZE_HOST*(i+1),arg2+i,arg3,index,SIZE_HOST);
    }
  inc_pa_perform(2*size);
  papamvun(argt,arg3,1);

}

/* parallel to scalar (indexed) move */
/* wdpascmv(pa,sc,index,mask,size) */
/* index is saved */
void wdpascmv(int arg1, int arg2, int arg3, int index, int size)
{
   int i;
   unsigned int word;

   word = 0;
   for(i=0;i<size;i++)
     {word  <<= 1;
      if((pa_mem[index][arg1+i]&pa_mem[index][arg3]) != 0) word = word|1;
    };
   sc_mem[arg2] = word;
}

/* parallel to scalar (indexed) move with sign extend to scalar 32 bits */
/* pascmvsx(pa,sc,index,size) */
/* index is saved */
void pascmvsx(int arg1, int arg2, int arg3, int size)
{
   int i,index,argt,j,rsize;

  argt = get_add("TEMP");
  papamvun(arg3,argt,1);
  index = first(arg3);
  j = (size-1)/SIZE_HOST;
  rsize = size-j*SIZE_HOST;
  wdpascsx(arg1,arg2+j,arg3,index,rsize);
  for (i =j-1; i>=0; i--)
    wdpascsx(arg1+size-SIZE_HOST*(i+1),arg2+i,arg3,index,SIZE_HOST);
  inc_pa_perform(2*size);
  papamvun(argt,arg3,1);
}

/* parallel to scalar (indexed) move */
/* wdpascmv(pa,sc,index,size) */
/* index is saved */
void wdpascsx(int arg1, int arg2, int arg3, int index, int size)
{
   int i;
   unsigned int word;

   if(pa_mem[index][arg1]&pa_mem[index][arg3]) word = -1; else word = 0;
   for(i=0;i<size;i++)
     {word  <<= 1;
      if(pa_mem[index][arg1+i]&pa_mem[index][arg3]) word = word|1;
    };
   sc_mem[arg2]=   word;
}



/* parallel clear - masked */
/* paclrmk(pa,mask,size) */
void paclrmk(int arg1, int arg2, int size)
{
   int i,p;

  for (i = size-1; i>=0; i--)
    for(p=0; p<NOOFW; p++)
     pa_mem[p][arg1+i] = pa_mem[p][arg1+i]&~pa_mem[p][arg2];
  inc_pa_perform(1+size);
 }


/* parallel clear - unmasked */
/* paclrun(pa,size) */
void paclrun(int arg1, int size)
{
   int i,p;

  for (i = size-1; i>=0; i--)
    for(p=0; p<NOOFW; p++)
	   pa_mem[p][arg1+i] = 0;
  inc_pa_perform(1+size);
}

/* parallel sign extend - masked */
/* pasnexmk(sign,pa,mask,size) */
void pasnexmk(int arg1, int arg2, int arg3, int size)
{
   int i,p;

  for(i = size-1; i>=0; i--)
    for(p=0; p<NOOFW; p++)
	   pa_mem[p][arg2+i] = (pa_mem[p][arg1]&pa_mem[p][arg3]) |
	                       (pa_mem[p][arg2+i]&~pa_mem[p][arg3]);
  inc_pa_perform(size);
}


/* parallel to parallel move, shifted- unmasked */
/* the parameters rotate and spinc determine behaviour
   1) rotate = 1 means to end around shift
   2) spinc is the number of 32 bit pieces to shift  */

void papamvsh(int arg1, int arg2, int size, int shiftdx, int moddx, int rotatedx)
{
  int i, p, shift, cshift,spinc,sp,mod,rotate;
  int pupper;
  unsigned carry;
   /* when host word size is > 16, shift will be 65535 instead of -1 */
   /* convert to -1 by subtracting 65535 if shift is greater than 16k */
   /* this will effectively limit the number of PEs to 16k or less  */
   /* That is, the max shift is 16k -1 */
   shift = sc_mem[shiftdx] & 0xFFFF;
   if (shift > 16384)
	   shift = shift - 65536;
   mod = sc_mem[moddx];
   rotate = sc_mem[rotatedx];
  spinc = mod/32;
  if(spinc <= 0) spinc = 1;
  if (shift >0)
    {cshift = sizeof(unsigned)*8-shift;
    for (i = size-1; i>=0; i--)
      {for(sp=0; sp<NOOFW; sp += spinc)
        {carry = 0;
         pupper = sp+spinc;
         if (pupper > NOOFW) pupper=NOOFW;
         for(p=sp; p<pupper; p++)
          {pa_mem[p][arg2+i] = (unsigned) pa_mem[p][arg1+i]>>shift | carry;
//printf("pa_mem[%d][%d]=%x<=[%d][%d]=%x\n",
//	   p,arg2+i,pa_mem[p][arg2+i],p,arg1+i,pa_mem[p][arg1+i]);
		 carry = (unsigned)pa_mem[p][arg1+i]<<cshift;}
       if(rotate)
          pa_mem[sp][arg2+i] |= carry;
    } }}
  else if (shift <0)
    {shift = -shift;
     cshift = sizeof(unsigned)*8-shift;
     for (i = size-1; i>=0; i--)
       {for(p=0; p<NOOFW-1; p++)
        {carry = (unsigned)pa_mem[p+1][arg1+i]>>cshift;
         pa_mem[p][arg2+i] =  (unsigned)pa_mem[p][arg1+i]<<shift | carry;}
       if(rotate) carry = (unsigned)pa_mem[0][arg1+i] >>cshift;
       else carry = 0;
       pa_mem[NOOFW-1][arg2+i] = ((unsigned)pa_mem[p][arg1+i] << shift) | carry;}
     }
   else papamvun(arg1,arg2,size);
  inc_pa_perform(size*3);
}
