#ifndef P2_COMMON
#define P2_COMMON

/* ==========================================================================

   This file contains preprocessor definitions and global data structure
   declarations for use with 'genicode' intermediate code generator.

============================================================================ */

/* TOKEN TYPES RETURNED BY THE SCANNER AND USED BY THE PARSER */

/*---end of file---*/
#define T_EOF           0

/* mode declarations */
/* mode is encoded  mode1,mode2 each has 16 values (4 bits) */
/* mode2 must be right most because some routines in pass2 test
   entire mode (all 8 bits) for T_PARALLEL or T_SCALAR  */

#define SIZE_MODE     256
#define SIZE_MODE2     16

#define T_NONE          0       /* no mode for quadruple */

/*  mode 1 */
#define T_INT           1       /* T_INT is decimal */
#define T_REAL          2
#define T_LOGICAL       3
/*      T_INDEX converted to T_LOGICAL in pass1    */
#define T_LONGINT       4
#define T_OCT     5       /* T_OCT is octal integer */
#define T_CHAR    6
#define T_CARD    7       /* cardinal - unsigned integers */
#define T_BIN     8
#define T_HEX     9
#define T_IFNANY 10       /* mode of T_ELSENANY if in T_IF */
#define T_CSTR   11
#define T_PRONOUN 12
#define T_DOL    13
#define T_COMDEL       14       /* use comma as a delimeter in T_PRINT_DEL */

/* mode 2 */
#define T_SCALAR              1
#define T_PARALLEL         2
#define T_COMMON_LEFT   3       /* not really token types, but are needed */
#define T_COMMON_RIGHT  4       /* to generate the op code portion of an  */
#define T_CNST          5       /* CONSTANT by definition is scalar */
#define T_MASKEDMOVE    6       /* pa[xx,...] = ...                  */
#define T_REG_SCALAR    7
#define T_REG_LEFT      8
#define T_REG_RIGHT     9
#define T_SCALAR_LEFT  10
#define T_SCALAR_RIGHT 11
#define T_THE          12  /* mode2 for THE allocate */
#define T_SILENT       13  /* prevents INPUT from printing header */
#define T_ALLDIMP      14  /* allocate in dimp - no terminating endalloc for emulator */


/*---type declaration keywords---*/
/*---arithmetic operators---*/
#define T_ADD           0x100
#define T_SUB           0x200
#define T_MULT             0x300
#define T_DIV           0x400

/*---relational operators---*/
#define T_LT            0x500
#define T_LE               0x600
#define T_EQ               0x700
#define T_NE               0x800
#define T_GT            0x900
#define T_GE               0xa00

/*---logical and bitwise operators---*/
#define T_OR               0xb00
#define T_AND              0xc00
#define T_NOT              0xd00
#define T_XOR              0xe00

/*---unary plus and minus operators---*/
#define T_MINUS         0xf00
#define T_PLUS          0x1000


/*---arithmetic functions----*/
#define T_COUNT      0x7f00
#define T_SUM        0x9d00

/*---assignment operator---*/
#define T_ASSIGN           0x1100

/* TOKEN TYPE STATUS VALUES - must be unique types  */
#define T_TRUE       0x1200
#define T_FALSE      0x1300

/* DEFINE OPERATORS */
#define T_DEFLOG     0x1500
#define T_DEFVAR     0x1600
#define T_DEFINE           0x1700
#define T_SHAPE      0xa100

/* procedure type */
#define T_MAIN             0x1800
#define T_SUBROUTINE    0x1900

/* include and assembly types */
#define T_INCLUDE          0x1a00
#define T_IOSPEC         0xc100
#define T_ASSEMBLE      0x1b00

/* Beginning of statement */
#define T_BOS        0x1c00

/* "Find" quadruples */
#define T_RESFST     0x1d00

/* Error quadruples */
#define T_MSG        0x1e00

/*---token types for control statement keywords---*/
#define T_BEGIF      0x2000
#define T_ANY              0x2100
#define T_DURING           0x2200
#define T_END              0x2300
#define T_ENDANY           0x2400
#define T_ENDFOR           0x2500
#define T_ENDGET           0x2600
#define T_ENDIF         0x2700
#define T_ENDIFNANY     0x2800
#define T_ENDLOOP          0x2900
#define T_ENDNEXT          0x2a00
#define T_ENDWHILE      0x2b00
#define T_ELSE             0x2c00
#define T_ELSENANY      0x2d00
#define T_FOR              0x2e00
#define T_GET              0x2f00
#define T_IF               0x3000
#define T_NANY        0x3100
#define T_LOOP             0x3200
#define T_NEXT             0x3300
#define T_THEN             0x3400
#define T_THENANY          0x3500
#define T_UNTIL         0x3600
#define T_WHILE         0x3700
#define T_BEGWHILE   0x3900
#define T_ALLOCATE      0x3a00
#define T_ASSOCIATE     0x3b00
#define T_FIRST         0x3d00
#define T_IN               0x3e00
#define T_RELEASE    0x3f00
#define T_ENDALLOCATE 0x4000
#define T_WITH             0x4300
#define T_CALL       0x4500
#define T_RETURN     0x4600

/* Subscript intermediate code */
#define T_MVSC       0x4700
#define T_MVPA       0x4800
#define T_MVPAID     0x4900
#define T_MVID       0x4a00
#define T_MVCB       0x4b00 /* added 880221 jlp */

/* added for ANDCONS   */
#define T_ANDFOR     0x4d00
#define T_ENDANDFOR  0x4e00
#define T_ANDIF      0x4f00
#define T_ATHEN      0x5000
#define T_ENDANDIF   0x5100
#define T_ANDCONS    0x5200
#define T_ENDANDCONS 0x5300

/* I/O KEYWORDS  */
   /* Print statement quadruples */
#define T_FSTOUT     0x6a00     /* NOTE: value out of sequence */
#define T_OUTPT      0x5400
#define T_COLAPSE    0x5500
#define T_BEGPRINT   0x5600
#define T_PRINT      0x5700
#define T_NL         0x5800
#define T_ENDPRINT   0x5900
#define T_PRINT_DEL  0x8a00    /*out of sequence value */
/* Read statement quadruples */
#define T_BEGREAD    0x5a00
#define T_READLN     0x5b00
#define T_READ       0x5c00
#define T_IOALLOC    0x5d00
#define T_EXPAND     0x5e00
#define T_INPUT      0x5f00
#define T_ENDREAD    0x6000
#define T_REREAD     0x9e00
#define T_COREAD     0x9f00
#define T_CONTIG     0xa000

#define T_DECLARE          0x6100
#define T_FIX              0x6200
#define T_FLOAT         0x6300
#define T_DEF_CONST     0x6400
#define T_IDENT         0x6500
#define T_CLRARRAY      0x6600
#define T_INITSTACK     0x6700
#define T_LABEL         0x6800
#define T_SETSCOPE      0x6900
/* #define T_FSTOUT: 0X6a00   DEFINED ABOVE  */
#define T_TEMP       0X6b00   /* added 880707 jlp */

#define T_STOP             0x6c00
#define T_SHORT         0x6d00
#define T_LONG             0x6e00
#define T_ENTRY         0x6f00
#define T_EXTERN           0x7000
#define T_STACK_SCOPE 0x7100
#define T_POP_SCOPE     0x7200
#define T_ENDSETSCOPE 0x7300
#define T_ENDMAIN          0x7400
#define T_ENDSUB           0x7500

#define T_MINDEX           0x7600
#define T_MINVAL           0x7700
#define T_MAXDEX           0x7800
#define T_MAXVAL           0x7900
#define T_NTHDEX     0xa900
#define T_NTHVAL     0xaa00

#define T_ADDRESS          0x7a00 /* indirect address token type */
#define T_OPEN       0x7b00
#define T_CLOSE      0x7c00
#define T_SCIN       0x7d00
#define T_SCOT       0x7e00
/*#define T_COUNT    0x7f00 */

#define T_BEGSTACK   0x8000 /* STACK FOR AND WHILE KEYWORDS */
#define T_STACK      0x8100
#define T_RECURSE    0x8200
#define T_ENDRECURSE 0x8300
#define T_POP        0x8400
#define T_STACKFOR   0x8500
#define T_ENDSTACKFOR 0x8600
#define T_DUMP       0x8700
#define T_STACKWHILE 0x8800
#define T_ENDSTACKWHILE 0x8900
/* fine T_PRINT_DEL  0x8a00      see print quadruples */
#define T_SCINL      0x8b00
#define T_SCINP      0x8c00
#define T_SCOTL      0x8d00
#define T_SCOTP      0x8e00

#define T_FSTCD      0x9000
#define T_NXTCD      0x9100
#define T_PREVCD     0x9200
#define T_TRNCD      0x9300
#define T_TRNACD     0x9400
#define T_SCIN8      0x9500
#define T_SCOT8      0x9600
#define T_FSTCD8     0x9700
#define T_NXTCD8     0x9800
#define T_PREVCD8    0x9900
#define T_TRNCD8     0x9a00
#define T_TRNACD8    0x9b00
#define T_SIBDEX     0x9c00
/*efine T_SUM        0x9d00
#define T_REREAD     0x9e00
#define T_COREAD     0x9f00
#define T_CONTIG     0xa000
#define T_SHAPE      0xa100 */
#define T_REGSCA     0xa200
#define T_SCAREG     0xa300
#define T_PREFIX     0xa400
#define T_PREFIX8    0Xa500
#define T_POPRC      0xa600
#define T_CHRINT     0xa700
#define T_INTCHR     0xa800
/*#define T_NTHDEX     0xa900
#define T_NTHVAL     0xaa00*/
#define T_COPY       0xab00
#define T_DECLAR     0xac00
#define T_GETADD     0xad00
#define T_DASH       0xae00
#define T_OUTPUT     0xaf00
#define T_DINPUT     0xb000
#define T_SILENCE    0Xb100
#define T_LEX        0xb200
#define T_USING      0xb300
#define T_ADJ        0xb400
#define T_NOUN       0xb500
#define T_PREP       0xb600
#define T_DIMP       0xb700
#define T_ENCODE     0xb800
#define T_DECODE     0xb900
#define T_EXP        0xba00
#define T_SQRT       0xbb00
#define T_POWER      0xbc00
#define T_DUMPMSG    0xbd00
#define T_ELSEWHERE  0xbe00
#define T_WHEREVER   0xbf00
#define T_DO         0xc000
#define T_IOSPEC     0xc100
#define T_COS        0xc200
#define T_SIN        0xc300
#define T_MOD        0xc400
#define T_RAND       0xc500
#define T_TKNCMP     0xc600
#define T_HIT        0xc700
#define T_CLEAR      0xc800
#define T_LOAD       0xc900
#define T_STORE      0xca00

#define T_TO         T_LT
#define T_FROM       T_GT

/* DEFINITIONS FOR FUNCTION RETURN VALUES */

#define TRUE            1
#define FALSE           0
#define FAIL            -1
#define SUCCESS         -2


/* DEFINITIONS FOR SCANNER AND PARSER CONSTANTS */
#define TOKEN_SIZE      80              /* maximum size of a token */
#define MAX_ID_SIZE     6               /* maximum size of an identifier */
#define MAX_NO_OUT  24     /* maximum number of output columns */
#define LINE_LENGTH     256              /* maximum size of input buffer for
					   assembly code */


/* EXTERNAL VARIABLE DECLARATIONS */

 int     debug,                   /* flag for debugging */
	       token_size,              /* size of current token */
	       token_type;              /* type of current token */

 char     token_buf[TOKEN_SIZE];  /* buffer for single token */

 FILE     *source_fp,             /* source file descriptor */
	*list_fp,               /* listing file descriptor */
	*object_fp;             /* object file descriptor */


/* DEFVAR global variables */
#define MAX_DEFVAR 50     /* max number of defvar variables */

struct equival { char  *ida; /* pointer identifer for a */
		 int       opa; /* operatore type for op a, i.e. T_ADD */
		 int       cna; /* constant assoc with a, ie. a+cna */
		 char  *idb; /* pointer identifer for b */
		 int       opb; /* operatore type for op b, i.e. T_ADD */
		 int       cnb; /* constant assoc with b, ie. b+cnb */
	       };

 struct equival  defvar_table[MAX_DEFVAR];

 int defvar_cnt;
 int linenum;  /* set by bos for error messages, etc */

/* FUNCTION DECLARATIONS */

#include "p2_prototypes.h"

#endif


